﻿using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;

using System.Linq;
using System.Text;


namespace QUADG
{
    public unsafe sealed class CMark
    {
        public int        xSW, ySW;

        public int        xImage = -1, yImage = -1;   //mark search result
        public double     xHat   = -1, yHat   = -1;   //predicted from lonTrue, latTrue
        public double     xHatCV = -1, yHatCV = -1;   //cross-v prediction
        public double     xGeoRef = -1, yGeoRef = -1; //found from ouput file search

        public double     lonTrue = -1, latTrue = -1; //as given
        public double     lonHat  = -1, latHat  = -1; //predicted from xImage,yImage

        private MarkType  type = MarkType.None;

        private int       xmin, xmax, ymin, ymax;

        private PTrgb[]   pts;      

        private int       npts   = 0;
        private double    angle  = 0.0;

        //default dimensions for standard mark shapes: tee, corner, cross
        //all these are in search window pixels
        private int       legLength = 25; //corner length and 1/2 length of cross-member of Tee
        private int       ticLength = 25; //length of descender of Tee

        //edge pen, used to guess distance white area is from edge
        private int       lineThickness = 2;

        public double     GeoTiffError,LSFitError;

        public double PixelError
        {
            get
            {
                if (xHat == -1 && yHat == -1) return 0;
                else return                   
                    Math.Sqrt((xImage - xHat) * (xImage - xHat) +
                              (yImage - yHat) * (yImage - yHat));
                    
            }
        }

        public double PixelErrorCV
        {
            get
            {
                if (xHatCV == -1 && yHatCV == -1) return 0;
                else return
                 Math.Sqrt((xImage - xHatCV) * (xImage - xHatCV) +
                           (yImage - yHatCV) * (yImage - yHatCV));
                
            }
        }
        public void ResetError()
        {
            this.xHat = this.yHat = this.xHatCV = this.yHatCV = -1;
            this.xImage = this.yImage = -1;
        }
        public double EarthError
        {
            //quick and dirty spherical approx
            get
            {
                if (xHat == -1 && yHat == -1) return 0;
                else
                    //     earth R in meters
                    return 6371.0 * 1000.0 * 
                    Math.Acos(
                    Math.Sin(latHat * Math.PI / 180.0) * Math.Sin(latTrue * Math.PI / 180.0) +
                    Math.Cos(latHat * Math.PI / 180.0) * Math.Cos(latTrue * Math.PI / 180.0) *
                    Math.Cos((lonHat-lonTrue) * Math.PI / 180.0));
                
            }
        }


        public int LineThickness
        { 
            set {lineThickness = value;}
            get {return lineThickness;}
        }
        public  double    Angle
        {
            set { angle = value; SetPoints(); }
            get { return angle;}
        }
        public MarkType   Type
        {
            set { type = value; SetPoints(); }
            get { return type; }
        }
        public int        LegLength
        {
            set { legLength = value; SetMarkSize(legLength, ticLength); }
            get { return legLength; }
        }
        public int        TicLength
        {
            set { ticLength = value; SetMarkSize(legLength, ticLength); }
            get { return ticLength; }
        }

        public bool SetMarkSize(int LegLength, int TicLength)
        {
            if (LegLength > 0 && TicLength > 0)
            {
                legLength = LegLength;
                ticLength = TicLength;
                SetPoints();
                return true;
            }
            else
            {
                return false;
            }
        }

        public int        bitmapWidthBytes,bitmapWidth,bitmapHeight;
        public Byte*      pBase = null;

        public Bitmap     bitmapProcess = new Bitmap(1, 1);
        public BitmapData bitmapData = null;

        public struct Pixel
        {
            public byte blue;
            public byte green;
            public byte red;
        }

        public Pixel* PixelAt(int x, int y)
        {
            return (Pixel*)(pBase + y * bitmapWidthBytes + x*sizeof(Pixel));
        }


        public void LockBitmap(Bitmap bmp)
        {
            GraphicsUnit unit  = GraphicsUnit.Pixel;
            RectangleF   boundsF = bmp.GetBounds(ref unit);
            Rectangle    bounds   = new Rectangle((int)boundsF.X, (int)boundsF.Y,
                                      (int)boundsF.Width, (int)boundsF.Height);

            bitmapWidthBytes   = (int)boundsF.Width * sizeof(Pixel);

            bitmapWidth  = (int)boundsF.Width;
            bitmapHeight = (int)boundsF.Height;

            if (bitmapWidthBytes % 4 != 0)
            {
                bitmapWidthBytes = 4 * (bitmapWidthBytes / 4 + 1);
            }

            bitmapData    = bmp.LockBits(bounds, ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            pBase = (Byte*)bitmapData.Scan0.ToPointer();
            
        }

        public void UnlockBitmap(Bitmap bmp)
        {
            bmp.UnlockBits(bitmapData);
            bitmapData = null;
            pBase = null;

        }

 
        public double MatchPattern(Bitmap bmp, string StatsFile)
        {

            if (type == MarkType.BotEdge || type == MarkType.TopEdge ||
               type == MarkType.LeftEdge || type == MarkType.RightEdge)
            {
                return FindInsideEdge(bmp, StatsFile);
            }
            else return MarkSearch(bmp, StatsFile);
        }

        public List<PT> FindPossEdges(Bitmap bmp, string StatsFile)
        {
            //return list of possible edge locations within Rect
            //a possible edge is a local minimum of error values. Any local minimum with an error less than tol
            //will be reported. We start with a small value for TOL and increase until a point meet the criteria.

            double        tol    = 0.4;
            bool          noPointsFound = true;
            int           x,y,xstart, xend, ystart, yend;
            PTxyval       xyv;
            List<PTxyval> xyvAll  = new List<PTxyval>();

            List<PT>      xyPoss  = new List<PT>();

            string s;
            System.IO.StreamWriter file = null;

            LockBitmap(bmp);

            if (StatsFile != "")
            {
                bmp.Save(StatsFile + ".png", ImageFormat.Png);
                file = new System.IO.StreamWriter(StatsFile + ".txt");

            }

            //edge patterns depend on bitmap size
            SetEdgePts();

            xstart = -xmin;
            xend = bitmapWidth - xmax - 1;

            ystart = -ymin;
            yend = bitmapHeight - ymax - 1;

            x = xstart;
            y = ystart;

            if (type == MarkType.TopEdge || type == MarkType.BotEdge)
            {
                for (y = ystart; y <= yend; y++)
                {

                    xyv.x = x;
                    xyv.y = y;
                    xyv.value = CheckLocation(x, y);

                    xyvAll.Add(xyv);
                    
                    if (false && file != null)
                    {
                        s = String.Format("{0:d5} {1:d5} {2:f6}", x, y, xyv.value);
                        file.WriteLine(s);
                    }
                }
            }
            else
            {
                for (x = xstart; x <= xend; x++)
                {
                    xyv.x = x;
                    xyv.y = y;
                    xyv.value = CheckLocation(x, y);

                    xyvAll.Add(xyv);
                    if (file != null)
                    {
                        s = String.Format("{0:d5} {1:d5} {2:f6}", x, y, xyv.value);
                        file.WriteLine(s);
                    }
                }

            }
            UnlockBitmap(bmp);

            while(noPointsFound && tol < 0.51)
            {
                for (int i = 1; i < xyvAll.Count - 1; i++)
                {
                    if (xyvAll[i].value < tol &&
                        xyvAll[i].value < xyvAll[i - 1].value &&
                        xyvAll[i].value < xyvAll[i + 1].value)
                    {

                        xyPoss.Add(new PT(xyvAll[i].x,xyvAll[i].y));

                        /*
                        if (file != null)
                        {
                            s = String.Format("possible edge at ({0:d5},{1:d5}) ({2:d5},{3:d5}) error is {4:f6}",
                                xyvAll[i].x,xyvAll[i].y,
                                xyv.x, xyv.y, 
                                xyv.value);
                            file.WriteLine(s);
                        }
                        */
                        noPointsFound = false;
                    }
                }
                tol += 0.1;
               
            }

            if (file != null) file.Close();


            return xyPoss;
        }
        
        public double FindInsideEdge(Bitmap bmp, string StatsFile)
        {
            //find inner-most edge

            double err, errMin;
            int xstart, xend, ystart, yend;

            List<double> errorVals = new List<double>();
            List<int> xy = new List<int>();

            string s;
            System.IO.StreamWriter file = null;

            LockBitmap(bmp);

            if (StatsFile != "")
            {
                bmp.Save(StatsFile+ ".png", ImageFormat.Png);
                file = new System.IO.StreamWriter(StatsFile + ".txt");

            }

            //edge patterns depend on bitmap size
            SetEdgePts();

            errMin = 1e20;
            xSW = ySW = 0;

            xstart = -xmin;
            xend = bitmapWidth - xmax - 1;

            ystart = -ymin;
            yend = bitmapHeight - ymax - 1;

            //searching for an edge, so sweep pattern in just one direction
            if (xend - xstart < 25) xend = xstart;
            if (yend - ystart < 25) yend = ystart;

            for (int x = xstart; x <= xend; x++)
            {
                for (int y = ystart; y <= yend; y++)
                {
                    err = CheckLocation(x, y);
                    errorVals.Add(err);

                    if (type == MarkType.BotEdge || type == MarkType.TopEdge)
                        xy.Add(y);

                    else xy.Add(x);

                    if (false && file != null)
                    {
                        s = String.Format("{0:d5} {1:d5} {2:f6}", x, y, err);
                        file.WriteLine(s);
                    }
                }
            }


            double mean = errorVals.Average();
            if (type == MarkType.TopEdge || type == MarkType.LeftEdge)
            {
                errorVals.Reverse();
                xy.Reverse();
            }
           

            for (int i = 1; i < errorVals.Count - 1; i++)
            {
                if (errorVals[i] < mean / 2.0 &&
                    errorVals[i] < errorVals[i - 1] &&
                    errorVals[i] < errorVals[i + 1])
                {
                    errMin = errorVals[i];
                    xSW = xy[i];
                    ySW = xy[i];

                    if (file != null)
                    {
                        s = String.Format("possible edge at {0:d5} error is {1:f6}", i, errMin);
                        file.WriteLine(s);
                    }
                    break;
                }
            }

            if (file != null) file.Close();

            UnlockBitmap(bmp);

            return errMin;
        }
        public double MarkSearch(Bitmap bmp, string StatsFile)
        {

            double err, errMin;
            int xstart, xend, ystart, yend;
           
            string s;
            System.IO.StreamWriter file = null;

            LockBitmap(bmp);

            if (StatsFile != "")
            {
                bmp.Save(StatsFile+".png", ImageFormat.Png);
                file = new System.IO.StreamWriter(StatsFile+".txt");

            }

            errMin = 1e20;
            xSW = ySW = 0;

            xstart = -xmin;
            xend = bitmapWidth - xmax - 1;

            ystart = -ymin;
            yend = bitmapHeight - ymax - 1;

            //if searching for an edge, sweep pattern in just one direction
            if (xend - xstart < 25) xend = xstart;
            if (yend - ystart < 25) yend = ystart;

            for (int x = xstart; x <= xend; x++)
            {
                for (int y = ystart; y <= yend; y++)
                {
                    err = CheckLocation(x, y);

                    if (file != null)
                    {
                        s = String.Format("{0:d5} {1:d5} {2:f6}", x, y, err);
                        file.WriteLine(s);
                    }
                    if (err < errMin)
                    {
                        xSW = x;
                        ySW = y;
                        errMin = err;
                    }
                }
            }


            if (file != null) file.Close();

            UnlockBitmap(bmp);

            return errMin;
        }
        public double CheckLocation(Bitmap bmp, int x, int y)
        {
            
                LockBitmap(bmp);
                double err = CheckLocation(x, y);
                UnlockBitmap(bmp);

                return err;
        
        }

        private double CheckLocation(int x, int y)
        {
            int    i,j;
            double sum = 0;
            Pixel  *pix;
           
            for(int n=0; n<npts; n++)
            {
                i = x+pts[n].x;
                j = y+pts[n].y;
                if (i >= 0 && i < bitmapWidth && j >= 0 && j < bitmapHeight)
                {

                    pix = PixelAt(i, j);

                    sum += (pix->red   - pts[n].r) * (pix->red   - pts[n].r) +
                           (pix->green - pts[n].g) * (pix->green - pts[n].g) +
                           (pix->blue  - pts[n].b) * (pix->blue  - pts[n].b);
                }
            }
            return Math.Sqrt((sum / (255 * 255)) / npts)/Math.Sqrt(3.0);

        }
        private void SetPoints()
        {
            npts = 0;
            if (type == MarkType.None) return;

            // edge points are set once bitmap limits are known
            else if (type == MarkType.RightEdge || type == MarkType.LeftEdge ||
                     type == MarkType.BotEdge   || type == MarkType.TopEdge)
                
                return;
        

            else if (type == MarkType.Custom)
                SetCustomPts();

            else SetStdPoints();
            
            //if(Angle != 0) RotatePoints();

        }
        void RotatePoints()
        {
            RotatePoints(Angle);
        }

        void RotatePoints(double angD)
        {
            int    xtemp;
            double ct,st;

            if (npts < 1 || angD == 0.0) return;

            ct = Math.Cos(angD * Math.PI / 180.0);
            st = Math.Sin(angD * Math.PI / 180.0);

            for(int i=0; i<npts; i++)
            {
                xtemp    = (int) Math.Round(pts[i].x*ct - pts[i].y*st);
                pts[i].y = (int) Math.Round(pts[i].x*st + pts[i].y*ct);
                pts[i].x = xtemp;
            }
            return;
        }
        private void SetCustomPts()
        {
            //to be implemented
            xmax = 0; 
        }

        private void SetEdgePts()
        {
            /*
             * an edge pattern consists of two vertical lines (left and right edges)
             * or two horizontal lines (top and bottom edges). The lines are black (the
             * edge itself, and white (the collar area around the map). For a right edge, 
             * the left line is black with a white line some penAllowance pixels to the right. 
             * For a left edge the white line is penAllowance pixels to the left. Similarly 
             * for top and bottom edges.
             */

            int  penAllowance = 2 + lineThickness;
            int  size;

            //a horizontal edge is as wide as the bitmap. Vertical edge is as tall as the bitmap
            if (type == MarkType.TopEdge || type == MarkType.BotEdge) size = bitmapWidth;
            else                                                      size = bitmapHeight;

            pts = new PTrgb[2 * size];

            npts = 0;
            for (int i = -size / 2 + 2; i < size / 2 - 2; i++)
            {
                pts[npts].x = 0;
                pts[npts].y = i;
                pts[npts].r = pts[npts].g = pts[npts].b = 0;
                npts++;

                pts[npts].x = -penAllowance;
                pts[npts].y = i;
                pts[npts].r = pts[npts].g = pts[npts].b = 255;
                npts++;
                
            }
            
            if (type == MarkType.TopEdge)        RotatePoints(90);
            else if (type == MarkType.RightEdge) RotatePoints(180);
            else if( type == MarkType.BotEdge)   RotatePoints(-90);

            if (Angle != 0) RotatePoints();
            
            xmax = ymax = -10000;
            xmin = ymin =  10000;
            for (int i = 0; i < npts; i++)
            {
                xmin = Math.Min(xmin, pts[i].x);
                ymin = Math.Min(ymin, pts[i].y);
                xmax = Math.Max(xmax, pts[i].x);
                ymax = Math.Max(ymax, pts[i].y);
            }

        }

        private void SetStdPoints()
        {
            /*  
             
             * The standard marks are either a corner, a tee or a plus. All mark patterns  
             * are centered at 0,0. Corners and tees have black pixels representing the
             * neatline and white pixels located penAllowance outside the neatline. 
            
             */

            int penAllowance = 2 + lineThickness;

            npts = 0;

            if (type == MarkType.LeftTee || type == MarkType.RightTee ||
                type == MarkType.TopTee  || type == MarkType.BotTee
                )
            {
                //mark is a tee---set up as if on left edge
                pts = new PTrgb[4 * legLength + ticLength];

                for (int i = 0; i < 2 * legLength; i++)
                {
                    pts[npts].x = 0;
                    pts[npts].y = i - legLength;
                    pts[npts].r = pts[npts].g = pts[npts].b = 0;
                    npts++;
                    
                    pts[npts].x = -penAllowance;
                    pts[npts].y = i - legLength;
                    pts[npts].r = pts[npts].g = pts[npts].b = 255;
                    npts++;
                     
                }

                for (int i = 0; i < ticLength; i++)
                {
                    pts[npts].x = i;
                    pts[npts].y = 0;
                    pts[npts].r = pts[npts].g = pts[npts].b = 0;
                    npts++;
                }

                //rotate clockwise as appropriate for actual tee type
                if (type == MarkType.RightTee)    RotatePoints(180);
                else if (type == MarkType.TopTee) RotatePoints(90);
                else if (type == MarkType.BotTee) RotatePoints(-90);


            }


            else if (type == MarkType.BotLeftCorn  || type == MarkType.TopLeftCorn ||
                     type == MarkType.BotRightCorn || type == MarkType.TopRightCorn)
            {
                //mark is a corner---set up for lower,left corner
                pts = new PTrgb[4 * legLength];
                for (int i = 0; i < legLength; i++)
                {
                    pts[npts].x = 0;
                    pts[npts].y = i - legLength;
                    pts[npts].r = pts[npts].g = pts[npts].b = 0;
                    npts++;
                    pts[npts].x = -penAllowance;
                    pts[npts].y = i - legLength;
                    pts[npts].r = pts[npts].g = pts[npts].b = 255;
                    npts++;
                }
                for (int i = 0; i < legLength; i++)
                {
                    pts[npts].x = i;
                    pts[npts].y = 0;
                    pts[npts].r = pts[npts].g = pts[npts].b = 0;
                    npts++;
                    pts[npts].x = i;
                    pts[npts].y = penAllowance;
                    pts[npts].r = pts[npts].g = pts[npts].b = 255;
                    npts++;
                }

                //rotate clockwisee as appropriate for actual corner type
                if (type == MarkType.TopLeftCorn)       RotatePoints(90);
                else if (type == MarkType.TopRightCorn) RotatePoints(180);
                else if (type == MarkType.BotRightCorn) RotatePoints(-90);

            }

            //interior plus mark, all black pixels
            else
            {
                pts = new PTrgb[4 * ticLength];
                for (int i = 0; i < 2 * ticLength; i++)
                {
                    pts[npts].x = -ticLength + i;
                    pts[npts].y = 0;
                    pts[npts].r = pts[npts].g = pts[npts].b = 0;
                    npts++;
                }
                for (int i = 0; i < 2 * ticLength; i++)
                {
                    pts[npts].x = 0;
                    pts[npts].y = -ticLength + i;
                    pts[npts].r = pts[npts].g = pts[npts].b = 0;
                    npts++;
                }
            }

            RotatePoints();
            xmax = ymax = -10000;
            xmin = ymin =  10000;
            for (int i = 0; i < npts; i++)
            {
                xmin = Math.Min(xmin, pts[i].x);
                ymin = Math.Min(ymin, pts[i].y);
                xmax = Math.Max(xmax, pts[i].x);
                ymax = Math.Max(ymax, pts[i].y);
            }
        }
    }   
}