﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;


namespace QUADG
{
    static class GDALtools
    {
        static public string Dir;
        static public string GDAL_Translate;
        static public string GDAL_Warp;

        static GDALtools()
        {
            if (Dir == null)
            {
                FindGDALtools();
            }
        }
        static public bool AddBinToPath()
        {
            if (GDALtools.Dir == null) return false;

            else
            {

                string newPath = Environment.GetEnvironmentVariable("PATH");

                if (!newPath.EndsWith(";")) newPath += ";";
                    

               newPath += GDALtools.Dir+";";

                Environment.SetEnvironmentVariable("PATH", newPath, EnvironmentVariableTarget.Process);

                return true;

            }

        }
        static public bool FindGDALtools()
        {
            bool found = false;

            for (int i = Convert.ToInt16('C'); i <= Convert.ToInt16('Z') && !found; i++)
            {
                string drive = Convert.ToChar(i) + ":";
                if (CheckDriveForGDAL(drive))
                {
                    found = true;
                    break;
                }
            }

            if (!found)
            {
                string msg = "Can't find GDAL utilities in standard locations on any of your drives.";
                msg += "\nShould be \\osgeo4w or \\osgeo4w64 or \\Program Files (x86)\\FWTools2.4.7 or similar.";
                msg += "\n\nSee www.gdal.org";

                MessageBox.Show(msg, "GDAL is Required");
                Dir = "";
                System.Environment.Exit(-1);

                return false;
            }

            else return true;
        }
        static private bool CheckDriveForGDAL(string Drive)
        {
            string[] subdirectoryEntries;
            string dir;

            for (int i = 0; i < 2; i++ )
            {
                dir = Drive + "\\osgeo4w";
                if (i == 0) dir = Drive + "\\osgeo4w";
                else        dir = Drive + "\\osgeo4w64";
                
                if (Directory.Exists(dir))
                {
                    Dir = dir;
                    if (File.Exists(Dir + "\\bin\\gdal_translate.exe"))
                        GDAL_Translate = Dir + "\\bin\\gdal_translate.exe";

                    if (File.Exists(Dir + "\\bin\\gdalwarp.exe"))
                        GDAL_Warp = Dir + "\\gdalwarp.exe";
                    if (GDAL_Translate == "" || GDAL_Warp == "")
                        return false;

                    else return true;
                }
            }

            for (int i = 0; i < 2; i++)
            {
                if (i == 0) dir = Drive + @"\Program Files (x86)";
                else        dir = Drive + @"\Program Files";
                
                if (Directory.Exists(dir))
                {
                    subdirectoryEntries = Directory.GetDirectories(dir);

                    foreach (string subdirectory in subdirectoryEntries)
                    {
                        string s = subdirectory.ToUpper();

                        if (s.Contains("FWTOOLS") )
                            
                        {
                            Dir = subdirectory;
                            if(File.Exists(Dir+"\\gdal_translate.exe"))
                                GDAL_Translate = Dir+"\\gdal_translate.exe";

                            else if (File.Exists(Dir + "\\bin\\gdal_translate.exe"))
                                GDAL_Translate = Dir + "\\bin\\gdal_translate.exe";


                            if (File.Exists(Dir + "\\gdalwarp.exe"))
                                GDAL_Warp = Dir + "\\gdalwarp.exe";

                            else if (File.Exists(Dir + "\\bin\\gdalwarp.exe"))
                                GDAL_Warp = Dir + "\\bin\\gdalwarp.exe";

                            return true;
                        }

                    }
                }
            }

            //didn't find FWTOOLS (no Program Files on this drive, or no FWTOOLS below that)
            return false;

        }




    }
}
