﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.IO;
using System.Windows.Forms;
using System.Diagnostics;
using System.Threading;
using System.Runtime.InteropServices;


namespace QUADG
{
    public unsafe class ImageFile
    {

        private FRM_MainForm thisParent;
        
        //bool               debug = false;
        private FileStream file;
        public string      fileName;

        private int        scanImageHeight, scanImageWidth;  //height,width of image file
        private long[]     scanLineOffsets;

        private int        bitmapWidthBytes, bitmapHeightBytes;
        private int        bitmapWidthPixels, bitmapHeightPixels;

        private Bitmap     bitmapProcess = new Bitmap(1, 1);
        private BitmapData bitmapData = null;
        private Byte*      pBase = null;

        public bool        needRotation = false;
        public string      tempPpmName;

        public ImageFile(FRM_MainForm xThisParent)
        {
            thisParent = xThisParent;
        }


        public struct Pixel
        {
            public byte blue;
            public byte green;
            public byte red;
        }


        public int Height
        {
            get { return scanImageHeight; }
            set { }
        }
        public int Width
        {
            get { return scanImageWidth; }
            set { }
        }
        public bool CloseFiles()
        {
            if (file != null) 
            { 
                file.Close();
                if (File.Exists(tempPpmName)) File.Delete(tempPpmName);
                return true;
            }
            else return false;
        }
        public bool LoadFile(string FileName, string ConverterProc)
        {
            fileName    = FileName;
            tempPpmName = "";

            if (fileName.ToLower().Contains(".tif")) return LoadTifFile(fileName);

            else if (fileName.ToLower().Contains(".ppm")) return LoadPPMFile(fileName);

            else return ConvertToPPM(fileName, ConverterProc);
        }

        public bool ConvertToPPM(string fileName, string converterProc)
        {

            string cmd, args;

            tempPpmName = Path.GetTempFileName();
            tempPpmName += ".ppm";

            FileInfo fileInfo = new FileInfo(fileName);

            long fileSize = fileInfo.Length;

            string dName = Directory.GetCurrentDirectory();

            cmd = converterProc;

            //args for irfanview
            args = "\""+ fileName + " /convert=\"" + tempPpmName + "\"";

            //args for gdal_translate
            args = "\"" + fileName + "\" -of pnm \"" + tempPpmName + "\"";

            thisParent.TBX_Output.AppendText(
                    "Converting: " + fileName + Environment.NewLine + 
                    "        to: " + tempPpmName + Environment.NewLine);

            Cursor.Current = Cursors.WaitCursor;

            using (Process proc = Process.Start(cmd, args))
            {
                while (!proc.HasExited)
                {
                    Thread.Sleep(1000);

                    if (converterProc.ToLower().Contains("i_view"))
                    {
                        FileInfo ppmInfo = new FileInfo(tempPpmName);
                        if (ppmInfo.Exists)
                        {
                            long ppmSize = ppmInfo.Length;
                            float fraction = (float)ppmSize / fileSize;
                            string pcs = String.Format(
                                         "{0:P} converted", fraction
                                         );


                            //textBox1.AppendText(pcs+"\n");
                            //textBox1.Update();
                        }
                    }
                }
                Cursor.Current = Cursors.Arrow;
                if (proc.ExitCode != 0) return false;

            }

            Application.DoEvents();

            return LoadPPMFile(tempPpmName);
        }


        public bool LoadPPMFile(string fileName)
        {
            string s;
            string[] sList;
            long headerLength;

            try
            {

                StreamReader sr = new StreamReader(fileName);

                s = sr.ReadLine();
                headerLength = s.Length + 1;

                if (s != "P6")
                {
                    MessageBox.Show(fileName + "not P6 (binary rgb)");
                    return false;
                }
                else
                {
                    s = sr.ReadLine();
                    headerLength += s.Length + 1;
                    while (s.StartsWith("#"))
                    {
                        s = sr.ReadLine();
                        headerLength += s.Length + 1;
                    }

                    sList = s.Split(' ');
                    scanImageWidth = int.Parse(sList[0]);
                    scanImageHeight = int.Parse(sList[1]);

                    s = sr.ReadLine();
                    headerLength += s.Length + 1;

                    sr.Close();

                    file = File.OpenRead(fileName);

                    scanLineOffsets = new long[scanImageHeight];
                    for (int i = 0; i < scanImageHeight; i++)
                    {
                        scanLineOffsets[i] = headerLength + i * 3 * scanImageWidth;
                    }

                    return true;
                }

            }
            catch
            {
                MessageBox.Show("Couldn't open\n" + fileName);
                return false;
            }
        }

        public bool LoadTifFile(string fileName)
        {
            string tagDataType, tagCode;

            int      answer, nTags, tagDataTypeI, tagCodeI;
            int      rowsPerStrip;
            long     ifdOffset, tagCount, nBytes;

            byte[]   bTagVals;
            long[]   lTagVals, stripOffsets;
            double[] dTagVals;


            Dictionary<int, string> tifCodes =
              new Dictionary<int, string>()
            {
               {254,"NewSubFileType"},
               {255,"SubFileType"},
               {256,"ImageWidth"},
               {257,"ImageLength"},
               {258,"BitsPerSample"},
               {259,"Compression"},
               {262,"Photometric"},
               {273,"StripOffsets"},
               {274,"Orientation"},
               {277,"SamplesPerPixel"},
               {278,"RowsPerStrip"},
               {279,"StripByteCounts"},
               {282,"XResolution"},
               {283,"YResolution"},
               {284,"PlanarConfig"},
               {296,"ResolutionUnit"},
               {305,"Software"},
               {306,"DateTime"},
               {339,"SampleFormat"},
               {700,"XMP"},
               {33550,"ModelPixelScale"},
               {33922,"ModelTiePoint"},
               {34377,"Photoshop"},
               {34665,"ExifIFD"},
               {34735,"GeoKeyDirectory"},
               {34736,"GeoDoubleParamsTag"},
               {34737,"GeoAsciiParams"}

            };

            //6 and onward are TIFF 6.0 types
            Dictionary<int, string> tagDataTypes =
              new Dictionary<int, string>() 
            {
               {1,"Byte"},  
               {2,"String"}, 
               {3,"Short"},  
               {4,"Long"},   
               {5,"Rational"}, 
               {6,"Signed Byte"}, 
               {7,"8-bit chunk"}, 
               {8,"Signed Short"},
               {9,"Signed Long"},
               {10,"Signed Rational"},
               {11,"Float"},
               {12,"Double"},
            };

            //admissible values
            int tifRGB = 2;
            int tifNoCompression = 1;
            int tifSamplesPerPixel = 3;

            rowsPerStrip = 1;
            stripOffsets = new long[1];
            scanLineOffsets = new long[1];

            bTagVals = null; lTagVals = null; dTagVals = null; 

            try
            {
                file = File.OpenRead(fileName);

                BinaryReader bReader = new BinaryReader(file);

                bReader.ReadInt16();
                answer = bReader.ReadInt16();
                if (answer != 42 && answer != 12)
                {
                    if (!thisParent.RBN_Automatic.Checked)
                    {
                        MessageBox.Show("Invalid Tiff indicator (not 42)");
                    }
                    else
                    {
                        thisParent.TBX_Errors.AppendText(fileName + " : Invalid Tiff indicator (not 42)\n");
                    }
                    return false;
                }
                ifdOffset = bReader.ReadInt32();
                file.Seek(ifdOffset, SeekOrigin.Begin);
                nTags = bReader.ReadInt16();
                for (int i = 0; i < nTags; i++)
                {
                    
                    file.Seek(ifdOffset + 2 + 12 * i, SeekOrigin.Begin);

                    tagCodeI = bReader.ReadUInt16();
                    tagDataTypeI = bReader.ReadUInt16();
                    tagCount = bReader.ReadInt32();

                    if (tifCodes.ContainsKey(tagCodeI))
                        tagCode = tifCodes[tagCodeI];
                    else tagCode = "Unknown";


                    if (!tagDataTypes.ContainsKey(tagDataTypeI))
                    {
                        /* use this to note unknown data types
         
                        thisParent.TBX_Errors.AppendText(fileName + " : Unknown data type in tag: "
                            + tagDataTypeI.ToString() + Environment.NewLine);
                        */

                        tagDataType = "unknown";

                    }
                    else tagDataType = tagDataTypes[tagDataTypeI];


                    if(tagDataType == "Byte")                   {nBytes = 1 * tagCount; bTagVals = new byte[tagCount];}

                    else if(tagDataType == "String")            {nBytes = 1 * tagCount; bTagVals = new byte[tagCount];}
                    else if(tagDataType == "Short")             {nBytes = 2 * tagCount; lTagVals = new long[tagCount];}
                    else if(tagDataType == "Long")              {nBytes = 4 * tagCount; lTagVals = new long[tagCount];}
                    else if(tagDataType == "Rational")          {nBytes = 2 * tagCount; lTagVals = new long[tagCount];}
                    else if(tagDataType == "Signed Byte")       {nBytes = 1 * tagCount; bTagVals = new byte[tagCount];}
                    else if(tagDataType == "8-bit chunk Bytes") {nBytes = 8 * tagCount; bTagVals = new byte[tagCount];}
                    else if(tagDataType == "Signed Short")      {nBytes = 2 * tagCount; lTagVals = new long[tagCount];}
                    else if(tagDataType == "Signed Long")       {nBytes = 4 * tagCount; lTagVals = new long[tagCount];}
                    else if(tagDataType == "Signed Rational")   {nBytes = 2 * tagCount; lTagVals = new long[tagCount];}
                    else if(tagDataType == "Float")             {nBytes = 4 * tagCount; dTagVals = new double[tagCount];}
                    else if(tagDataType == "Double")            {nBytes = 8 * tagCount; dTagVals = new double[tagCount];}
                    else                                        {nBytes = 1 * tagCount; lTagVals = new long[tagCount];}


                    if (nBytes > 4)
                    {
                        long tagOffset = bReader.ReadInt32();
                        file.Seek(tagOffset, SeekOrigin.Begin);
                    }
                    for (int j = 0; j < tagCount; j++)
                   {
                        if(tagDataType.Contains("Byte"))           bTagVals[j] = bReader.ReadByte();
                        else if (tagDataType.Contains("String"))   bTagVals[j] = bReader.ReadByte();
                        else if (tagDataType.Contains("Long"))     lTagVals[j] = bReader.ReadInt32();
                        else if (tagDataType.Contains("Rational")) lTagVals[j] = bReader.ReadInt32();
                        else if (tagDataType.Contains("Short"))    lTagVals[j] = bReader.ReadInt16();

                        else if (tagDataType.Contains("Float"))    dTagVals[j] = bReader.ReadSingle();
                        else if (tagDataType.Contains("Double"))   dTagVals[j] = bReader.ReadDouble();

                        else lTagVals[j] = bReader.ReadInt32();


                    }

                    
                    /*
                    if (fileName.ToUpper().Contains("GEO.TIF"))
                    {
                        using (StreamWriter sw = File.AppendText("TagDump.txt"))
                        {
                           sw.WriteLine(string.Format("Tag Code {0:d} {1}, {2}:",tagCodeI,tagCode,tagDataType));
                           for (int j = 0; j < Math.Min(25,tagCount); j++)
                           {
                              if(tagDataType.Contains("Byte") || tagDataType.Contains("String")) 
                                sw.Write(string.Format("{0}",Convert.ToChar( bTagVals[j])));
                                

                              else if (tagDataType.Contains("Float") || tagDataType.Contains("Double"))
                                sw.Write(string.Format("{0} ", dTagVals[j]));
                             
                              else sw.Write(string.Format("{0} ", lTagVals[j]));
                           }
                           sw.WriteLine(" ");

                        }
                    }
                    */


                    //all values for this tag are in
                    if (tagCode == "ImageWidth") scanImageWidth = (int)lTagVals[0];
                    else if (tagCode == "ImageLength")
                    {
                        scanImageHeight = (int)lTagVals[0];
                        scanLineOffsets = new long[scanImageHeight];
                    }
                    else if (tagCode == "StripOffsets")
                    {
                        stripOffsets = new long[tagCount];
                        for (int j = 0; j < tagCount; j++)
                        {
                            stripOffsets[j] = lTagVals[j];
                        }
                    }

                    else if (tagCode == "RowsPerStrip") rowsPerStrip = (int)lTagVals[0];

                    /* some scanned files don't code this correctly
                     * just do our best however the image is labeled
                    else if ((tagCode == "SubFileType"    && tagVals[0] > 2 )
                        ||   (tagCode == "NewSubFileType" && tagVals[0] > 1))
                    {
                        MessageBox.Show("Can only process single page tifs");
                        return false;
                    }
                    */

                    else if (tagCode == "Photometric" && lTagVals[0] != tifRGB)
                    {
                        if (!thisParent.RBN_Automatic.Checked)
                        {
                            MessageBox.Show("Can only process rgb tif files");
                        }
                        else
                        {
                            thisParent.TBX_Errors.AppendText(fileName + " : Can only process rgb tif files\n");
                        }
                        return false;
                    }
                    else if (tagCode == "Compression" && lTagVals[0] != tifNoCompression)
                    {
                        if (!thisParent.RBN_Automatic.Checked)
                        {
                            MessageBox.Show("Can only process uncompressed tif files");
                        }
                        else
                        {
                            thisParent.TBX_Errors.AppendText(fileName + " : Can only process uncompressed tif files\n");
                        }
                        return false;
                    }

                    else if (tagCode == "SamplesPerPixel" && lTagVals[0] != tifSamplesPerPixel)
                    {
                        if (!thisParent.RBN_Automatic.Checked)
                        {
                            MessageBox.Show("Can only process tif files with 3 bytes/pixels");
                        }
                        else
                        {
                            thisParent.TBX_Errors.AppendText(fileName + " : Can only process tif files with 3 bytes/pixels\n");
                        }
                        return false;
                    }

                }

                for (int row = 0; row < scanImageHeight; row++)
                {
                    int strip = row / rowsPerStrip;

                    scanLineOffsets[row] = stripOffsets[strip];

                    int rem = row - strip * rowsPerStrip;

                    scanLineOffsets[row] += rem * scanImageWidth * 3;
                }
                return true;
            }
            catch
            {
                if (!thisParent.RBN_Automatic.Checked)
                {
                    MessageBox.Show("Failed opening or reading tif\n" + fileName);
                }
                else
                {
                    thisParent.TBX_Errors.AppendText(fileName + " : Failed opening or reading tif\n");
                }
                return false;
            }
        }


        public Bitmap ExtractRectangle(Rectangle RectIn)
        {
            long      b0;
            Pixel*    pPixel;
            Rectangle rect;
            Bitmap    bmp;

            if (!needRotation) rect = RectIn;
            else
            {
                rect = new Rectangle(this.Width - (RectIn.Top+RectIn.Height) , RectIn.Left, RectIn.Height, RectIn.Width);
            }

            try
            {
                bmp = new Bitmap(rect.Width, rect.Height);

                byte[] pixelRow = new byte[rect.Width * 3];

                LockBitmap(bmp);

                for (int row = 0; row < rect.Height; row++)
                {
                    b0 = scanLineOffsets[rect.Top + row] + 3 * rect.Left;

                    file.Seek(b0, SeekOrigin.Begin);
                    file.Read(pixelRow, 0, rect.Width * 3);
                   
                    
                    for (int b = 0; b < 3*rect.Width; b+=3)
                    {

                        //bmp.SetPixel(b, row, Color.FromArgb(pixelRow[b*3],pixelRow[b*3+1],pixelRow[b*3+2]));
                        
                        //pPixel = PixelAt(col, row);

                        pPixel = (Pixel*)(pBase + row * bitmapWidthBytes + b);
                        
                        pPixel->red   = pixelRow[b];
                        pPixel->green = pixelRow[b + 1];
                        pPixel->blue  = pixelRow[b + 2];
                        
                    }
                    
                }

                UnlockBitmap(bmp, bitmapData);
      
                if(needRotation) bmp.RotateFlip(RotateFlipType.Rotate270FlipNone);

                return bmp;
            }
            catch
            {
                string s = string.Format(
                    "Problem extracting rectangle (left, top, right ,bottom ,width, height):"
                    + Environment.NewLine + "{0:d5} {1:d5} {2:d5} {3:d5} {4:d5} {5:d5}",
                    rect.Left, rect.Top, rect.Right, rect.Bottom, rect.Width, rect.Height) + Environment.NewLine;

                if (!thisParent.RBN_Automatic.Checked)
                    MessageBox.Show(s);
                else
                {
                        thisParent.TBX_Errors.AppendText(s);
                }
           
                return null;
            }
        }
        public FRM_MainForm.Pixel[,] ExtractSwath(int SwathTop, int SwathHeight)
        {

           int   x0,row,col,width;

           FRM_MainForm.Pixel[,] swath;
           byte[]                pixelRow;


           row = col = -999;
           try{
              if (!needRotation)
              {
                 if (SwathTop < 0 || SwathTop + SwathHeight > this.Height)
                 {
                    MessageBox.Show(string.Format(
                      "Bad swath arguments. SwathTop, SwathHeight are: {0:d5} {1:d5}", SwathTop, SwathHeight));
                    return null;
                 }

                 swath    = new FRM_MainForm.Pixel[SwathHeight, this.Width];
                 pixelRow = new byte[this.Width * 3];

                 for (row = 0; row < SwathHeight; row++)
                 {

                     file.Seek(scanLineOffsets[SwathTop + row], SeekOrigin.Begin);
                     file.Read(pixelRow, 0, this.Width * 3);
                     for (col = 0; col < this.Width; col++)
                     {
                        swath[row, col].red   = pixelRow[col * 3];
                        swath[row, col].green = pixelRow[col * 3 + 1];
                        swath[row, col].blue  = pixelRow[col * 3 + 2];
                     }
                 }
                 return swath;
              }//end of normal orientation

              else //rotated scan
              {
                 x0    = this.Width - SwathTop - SwathHeight;
                 width = SwathHeight;
                 if (x0 < 0 || x0 + width > this.Width)
                 {
                     MessageBox.Show(string.Format(
                       "Bad swath arguments. SwathTop, SwathHeight are: {0:d5} {1:d5}", SwathTop, SwathHeight));
                     return null;
                 }

                 swath    = new FRM_MainForm.Pixel[SwathHeight, this.Height];
                 pixelRow = new byte[width * 3];

                 for(row = 0; row < this.Height; row++)
                 {
                     file.Seek(3L*x0 + scanLineOffsets[row], SeekOrigin.Begin);
                     file.Read(pixelRow, 0, width * 3);
                     for (col = 0; col < width; col++)
                     {
                        swath[SwathHeight-1-col, row].red   = pixelRow[col * 3];
                        swath[SwathHeight-1-col, row].green = pixelRow[col * 3 + 1];
                        swath[SwathHeight-1-col, row].blue  = pixelRow[col * 3 + 2];
                     }
                 }
                 return swath;
              }      
           }

           catch
           {
                string s = string.Format(
                    "Problem extracting swath, top, height, row, col:" + Environment.NewLine 
                    + "{0:d5} {1:d5} {2:d5} {3:d5}" + Environment.NewLine,
                    SwathTop, SwathHeight,row,col);

                if (!thisParent.RBN_Automatic.Checked)
                    MessageBox.Show(s);
                else
                {
                    thisParent.TBX_Errors.AppendText(s);
                }
                return null;
           }
        }

        public void LockBitmap(Bitmap bmp)
        {
            GraphicsUnit unit = GraphicsUnit.Pixel;
            RectangleF boundsF = bmp.GetBounds(ref unit);
            Rectangle bounds = new Rectangle((int)boundsF.X, (int)boundsF.Y,
                                      (int)boundsF.Width, (int)boundsF.Height);

            bitmapWidthBytes = (int)boundsF.Width * sizeof(Pixel);
            bitmapHeightBytes = (int)boundsF.Height * sizeof(Pixel);

            bitmapWidthPixels = (int)boundsF.Width;
            bitmapHeightPixels = (int)boundsF.Height;

            if (bitmapWidthBytes % 4 != 0)
            {
                bitmapWidthBytes = 4 * (bitmapWidthBytes / 4 + 1);
            }


            //bitmapProcess = bmp;
            bitmapData = bmp.LockBits(bounds, ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            pBase = (Byte*)bitmapData.Scan0.ToPointer();



        }

        public void UnlockBitmap(Bitmap Bmp, BitmapData BitmapData)
        {
            Bmp.UnlockBits(BitmapData);
            bitmapData = null;
            pBase = null;

        }
        public Pixel* PixelAt(int x, int y)
        {
            return (Pixel*)(pBase + y * bitmapWidthBytes + x * sizeof(Pixel));
        }
    }

}
