﻿using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Media;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Diagnostics;
using System.Threading;
using System.Collections;
using Mapack;
using System.Security.Principal;


//check DemoOnly and debug flags

namespace QUADG
{
    public enum MarkType
    {
        TopTee, BotTee, LeftTee, RightTee, Plus,
        BotLeftCorn, TopLeftCorn, TopRightCorn, BotRightCorn,
        LeftEdge, RightEdge, TopEdge, BotEdge,
        Custom, None
    }
    public enum ThumbCaptionType
    {
        Nothing, SearchError, TransformError, GeoTiffError
    }
    public struct PT
    {
        public int x, y;
        public PT(int X, int Y) { x = X; y = Y; }
    }
    public struct PTrgb
    {
        public int x, y, r, g, b;
        public PTrgb(int X, int Y, int R, int G, int B)
        {  x = X; y = Y; r = R; g = G; b = B;}
            
    }
    public struct PTxyval
    {
        public int x, y;
        public double value;
        public PTxyval(int X, int Y, double V)
        {   x = X; y = Y; value = V; }
    }

    public struct EdgeSeg
    {
        //segments oriented left (p1) to right (p2) or top (p1) to bottom (p2)

        public PT     p1, p2, pMid; 
        public double angle;         //cc angle in degrees, will be near 0 or near 90
        public bool   vertical;

        public EdgeSeg(PT P1, PT P2)
        {
            //load points so p2 is larger in x or y
            if (Math.Abs(P1.x - P2.x) > Math.Abs(P1.y - P2.y))
            {
                vertical = false;
                if (P1.x < P2.x) { p1 = P1; p2 = P2; }
                else             { p2 = P1; p1 = P2; }
            }
            else
            {
                vertical = true;
                if (P1.y < P2.y) { p1 = P2; p2 = P1; }
                else             { p1 = P1; p2 = P2; }
            }
            if(Math.Abs(P1.x - P2.x) > 1.0e-10)
            {
                angle = Math.Atan((double)(p2.y - p1.y) / (p1.x - p2.x)) * 180.0 / Math.PI;
                if (angle < -45) angle += 180;
            }
            else
            {
                if (Math.Abs(p1.y - p2.y) > 1.0e-10) angle = 90;
                else angle = 0;
            }

            pMid.x = (int) Math.Round(.5*(p1.x+p2.x));
            pMid.y = (int) Math.Round(.5*(p1.y+p2.y));
        }
    }


    public struct SearchInfo
    {
        public MarkType  markType;
        public Rectangle imageRect;   //rectangle in original image being searched
        public double    searchRatio; //ratio of search rectangle size to that of imageRect
        public PT        ptFound;     //location of best match in original image coordinates
    }

    public struct XmlScanNode
    {
        public string fileName, mapName, controlMarkLayout, datum, outputGeographicFileName;
        public string controlMarkSpacing, latitude, longitude, mapScale, controlMarkSize, resolution;
        public string controMarkLonSpacing, controMarkLatSpacing;
        public string perimeterMarksOnly;
    }

    public unsafe sealed partial class FRM_MainForm : Form
    {
        //use to suppress output image files
        public bool       DemoOnly = false;


        public Bitmap     bitmapGray = new Bitmap(300, 300);
        public Bitmap     bitmapProcess = new Bitmap(1, 1);
        public Bitmap     bitmapForThumbBox = new Bitmap(1, 1);
        public int        bitmapWidth;
        public BitmapData bitmapData = null;
        public Byte*      pBase = null;

        private ThumbBox  currentSelectedThumbBox;
        public int        totalThumbColumns = 4;
        public int        totalThumbRows = 4;
        public int        totalThumbBoxes = 16;


        public ThumbBox[,] thumbBoxArray;
        public bool       thumbBoxesAreColor = true;
        private Boolean   thumbBoxesCreated = false;
        //public Boolean    fitButtonPressed = false;
        //public bool       showThumbInfo = false;

        private String[,] fileList;
        private int       totalFilesInList = 0;
        private String    currentXMLFileName = "";

        private FRM_Preferences preferencesDialog;
        private ManualInput     manualInputDialog;

        public int        prefSizeThumbWidth = 160;
        public int        prefSizeThumbHeight = 160;

        public int        sizeThumbWidth,sizeThumbHeight;

        public double     threshold1 = 2.0;  //error thresholds for thumbnail caption colors (pixels)
        public double     threshold2 = 3;
        public double     threshold3 = 4;
        public Color      crosshairColor  = Color.Green;
        public Color      predictionColor = Color.Red;
        public Color      threshold1Color = Color.FromArgb(255,255,255);
        public Color      threshold2Color = Color.FromArgb(180, 180, 180);
        public Color      threshold3Color = Color.FromArgb(255, 255, 0);
        public Color      threshold4Color = Color.FromArgb(255, 0, 0);
        public string     transformationXMLDirectory = Directory.GetCurrentDirectory() + "\\OutputXML";
        public string     transformedImageDirectory = Directory.GetCurrentDirectory() + "\\OutputImages";
        public string     errorDirectory = Directory.GetCurrentDirectory() + "\\OutputXMLErrors";
        public string     sourceImageDirectory = Directory.GetCurrentDirectory() + "\\SourceImages";
        public string     preferencesXMLFile = "preferences.xml";
        public int        maxLoadFileSize = 250;  //largest file szie (MB) to load as bitmap
        public bool       wantCVErrors;

        public double     metersPerPixel;


        private ArrayList   scanListArray = new ArrayList();
        private XmlScanNode xmlScanNode = new XmlScanNode();
        private int         currentFileSelected = 0;
        private System.Windows.Forms.Timer automaticTimer;

        public string     preTransformedFileName = "";
        private string    errorFileNameSuffix = "";
        private bool      runningAutoMode = false;
        private bool      stopProcessing = false;

        private string    verifyCSVFilename = "Verify.csv";
        private string    defaultDatum = "NAD27";


        public List<Rectangle> Neatline;

        double            quadAspect;             //based on lat,lon extent
        double            quadHeight, quadWidth;  //size in pixels based on quad extent, map scale and resolution
        int               imageHeight,imageWidth; //pixel sizes in normal aspect (north at top), not nec. same as image


        //*///////////////////////////////////////////////////////////////////////////////////////////////
        public bool debug = false;
        public bool noRotation = false;
        public bool noQA       = false;
        public bool noGeoref   = false;
        public bool linearOnly = false;
        public bool perimeterMarksOnly = false;

        //Items related to map series, mostly provided by input file

        //image dots per inch and scale (e.g. 24000 means 1:24,000)
        public int    imageDPI;  
        public double mapScale; 

        //neatline pen width (inches)
        public double neatLineThicknessIN = 0.01;

        //Datum variable, allowable values are NAD27, NAD83, WGS72, WGS84
        public string datum = "";

        //neatlines guaranteed to be w/in these fractions of scan 
        //if not, more search time will be needed
        public double MaxTopMargin   = 0.2;
        public double MaxLeftMargin  = 0.2;
        public double MaxBotMargin   = 0.20;
        public double MaxRightMargin = 0.25;

        public int maxRotation = 5;  //assumed maximum rotation in scanner (degrees)

        //end of map information

        public Affine scaleLonLat = new Affine();
        public Affine scaleuv = new Affine();
        public int    degreeFit = 2;
        Mapack.Matrix polyFit;


        public CMark[,] controlMark;               //used for control mark searches

        public Bitmap searchBmp, subBmp;

        public int    bitmapWidthBytes, bitmapHeightBytes;
        public int    bitmapWidthPixels, bitmapHeightPixels;

        public int    searchActualSize = 250; //pixels in bitmap we search
        public int    searchDisplaySize = 250; //pixels in screen display of search window

        public PT      corn1PT, corn2PT, corn3PT, corn4PT; 

        public ImageFile imageFile;

        public string preferenceFolder;  //directory to hold preferences file

        //An external image converter (e.g, GDAL, Irfanview, Imagic)is used for
        //unsupported file types. The full path to that procedure is converterProc.
        public string converterProc;

        public string newLine = Environment.NewLine;

        public int    neatLineThickness;

        public double maxImageRotation = 2;   //maximum departure from vertical/horizontal for edges
        public double maxEdgeChange    = 1; //acceptable difference between edge segments

        //*//////////////////////////////////////////////////////////////////////////////////////////////////

        public FRM_MainForm()
        {
            InitializeComponent();

            imageFile = new ImageFile(this);
            Neatline  = new List<Rectangle>();

            sizeThumbWidth  = prefSizeThumbWidth;
            sizeThumbHeight = prefSizeThumbHeight;

            //Gdal.AllRegister();

            PNL_ThumbArea.SizeChanged += new EventHandler(PNL_ThumbArea_SizeChanged);

            DisableButtons();

            //default location
            preferenceFolder = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData)
                                + "\\QUAD-G";
            
            if(!Directory.Exists(preferenceFolder))
            {
                DirectoryInfo di = Directory.CreateDirectory(preferenceFolder);
            }

            preferencesXMLFile = preferenceFolder + "\\Preferences.xml";

            ReadPreferences();

            preferencesDialog = new FRM_Preferences(this);

            for (int x = 0; x < bitmapGray.Height; x++)
                for (int y = 0; y < bitmapGray.Width; y++)
                    bitmapGray.SetPixel(x, y, Color.Gray);

            neatLineThickness = (int)Math.Round(imageDPI * neatLineThicknessIN);

            converterProc = GDALtools.GDAL_Translate;

        }
        public double SegAngle(PT P1, PT P2)
        {

            if (Math.Abs(P1.x - P2.x) < 1.0e-10)
            {
                if (Math.Abs(P1.y - P2.y) > 1.0e-10) return 90;
                else                                 return 0;
            }
            else
            {
                double a = Math.Atan((double)(P2.y - P1.y) / (P1.x - P2.x)) * 180.0 / Math.PI;
                if (a < -45) a += 180;
                return a;
            }
        }
        
        public void ReadPreferences()
        {

            if (!File.Exists(preferencesXMLFile))
            {
                MessageBox.Show("Preferences.xml does not exist.  A new preferences XML file will be created.");
                preferencesDialog = new FRM_Preferences(this);
                preferencesDialog.WritePreferences();
                return;
            } 

            XmlDocument xmlDoc = new XmlDocument();

            xmlDoc.Load(preferencesXMLFile);

            XmlNodeList preferencesList = xmlDoc.GetElementsByTagName("PreferencesList");

            foreach (XmlNode node in preferencesList)
            {
                XmlElement preferencesElement = (XmlElement)node;
                prefSizeThumbWidth = Convert.ToInt32(preferencesElement.GetElementsByTagName("ThumbWidth")[0].InnerText);
                prefSizeThumbHeight = Convert.ToInt32(preferencesElement.GetElementsByTagName("ThumbHeight")[0].InnerText);

                if(preferencesElement.InnerXml.Contains("SearchWindowSize"))
                 searchActualSize = Convert.ToInt32(preferencesElement.GetElementsByTagName("SearchWindowSize")[0].InnerText);
                

                string parseColor = preferencesElement.GetElementsByTagName("CrosshairColor")[0].InnerText;
                crosshairColor = Color.FromArgb(ParseRBGValues(parseColor, 0),
                                                ParseRBGValues(parseColor, 1),
                                                ParseRBGValues(parseColor, 2));

                if (preferencesElement.InnerXml.Contains("PredictionColor"))
                {
                    parseColor = preferencesElement.GetElementsByTagName("PredictionColor")[0].InnerText;
                    predictionColor = Color.FromArgb(ParseRBGValues(parseColor, 0),
                                                    ParseRBGValues(parseColor, 1),
                                                    ParseRBGValues(parseColor, 2));
                }


                threshold1 = Convert.ToDouble(preferencesElement.GetElementsByTagName("ThresholdLevel1")[0].InnerText);
                threshold2 = Convert.ToDouble(preferencesElement.GetElementsByTagName("ThresholdLevel2")[0].InnerText);
                threshold3 = Convert.ToDouble(preferencesElement.GetElementsByTagName("ThresholdLevel3")[0].InnerText);

                string parseThreshold1Color = preferencesElement.GetElementsByTagName("ThresholdColor1")[0].InnerText;
                threshold1Color = Color.FromArgb(ParseRBGValues(parseThreshold1Color, 0),
                                                ParseRBGValues(parseThreshold1Color, 1),
                                                ParseRBGValues(parseThreshold1Color, 2));
                string parseThreshold2Color = preferencesElement.GetElementsByTagName("ThresholdColor2")[0].InnerText;
                threshold2Color = Color.FromArgb(ParseRBGValues(parseThreshold2Color, 0),
                                                ParseRBGValues(parseThreshold2Color, 1),
                                                ParseRBGValues(parseThreshold2Color, 2));
                string parseThreshold3Color = preferencesElement.GetElementsByTagName("ThresholdColor3")[0].InnerText;
                threshold3Color = Color.FromArgb(ParseRBGValues(parseThreshold3Color, 0),
                                                ParseRBGValues(parseThreshold3Color, 1),
                                                ParseRBGValues(parseThreshold3Color, 2));
                string parseThreshold4Color = preferencesElement.GetElementsByTagName("ThresholdColor4")[0].InnerText;
                threshold4Color = Color.FromArgb(ParseRBGValues(parseThreshold4Color, 0),
                                                ParseRBGValues(parseThreshold4Color, 1),
                                                ParseRBGValues(parseThreshold4Color, 2));
                if (preferencesElement.GetElementsByTagName("XMLDirectory")[0].InnerText != "")
                {
                    transformationXMLDirectory = preferencesElement.GetElementsByTagName("XMLDirectory")[0].InnerText;
                }
                if (preferencesElement.GetElementsByTagName("ImageDirectory")[0].InnerText != "")
                {
                    transformedImageDirectory = preferencesElement.GetElementsByTagName("ImageDirectory")[0].InnerText;
                }
                if (preferencesElement.GetElementsByTagName("ErrorDirectory")[0].InnerText != "")
                {
                    errorDirectory = preferencesElement.GetElementsByTagName("ErrorDirectory")[0].InnerText;
                }
                if (preferencesElement.GetElementsByTagName("SourceImageDirectory")[0].InnerText != "")
                {
                    sourceImageDirectory = preferencesElement.GetElementsByTagName("SourceImageDirectory")[0].InnerText;
                }

                if(preferencesElement.InnerXml.Contains("NoRotation"))
                {
                    string s = preferencesElement.GetElementsByTagName("NoRotation")[0].InnerText;
                    if(s.ToUpper().Contains("TRUE")) noRotation = true;
                    else                             noRotation = false;
                }
                if (preferencesElement.InnerXml.Contains("LinearOnly"))
                {
                    string s = preferencesElement.GetElementsByTagName("LinearOnly")[0].InnerText;
                    if (s.ToUpper().Contains("TRUE")) linearOnly = true;
                    else linearOnly = false;
                }

                if (preferencesElement.InnerXml.Contains("NoGeoref"))
                {
                    string s = preferencesElement.GetElementsByTagName("NoGeoref")[0].InnerText;
                    if (s.ToUpper().Contains("TRUE")) noGeoref = true;
                    else noGeoref = false;
                }

                if (preferencesElement.InnerXml.Contains("NoQA") )
                {
                    string s = preferencesElement.GetElementsByTagName("NoQA")[0].InnerText;
                    if (s.ToUpper().Contains("TRUE")) noQA = true;
                    else noQA = false;
                }

                if (noGeoref) noQA = true;

            }
            
            LBL_TransformationDirectory.Text = transformationXMLDirectory;
            LBL_TransformedImageDirectory.Text = transformedImageDirectory;
            LBL_ErrorDirectory.Text = errorDirectory;
        }
        
        private int ParseRBGValues(string colorField, int colorIndex)
        {
            string[] fieldArray = colorField.Split(new char[] { ','});
            return Convert.ToInt32(fieldArray[colorIndex]);
        }

        private int CenterThumbBoxesHorizontal(int xRowNumber)
        {

            return (((totalThumbRows - 1) - xRowNumber) * sizeThumbHeight) + ((PNL_ThumbArea.Height - (totalThumbRows * sizeThumbHeight)) / 2);
        }

        private int CenterThumbBoxesVertical(int xColumnNumber)
        {
            
            return (xColumnNumber * sizeThumbWidth) + ((PNL_ThumbArea.Width - (totalThumbColumns * sizeThumbWidth)) / 2);
        }

        private void RedrawThumbBoxes()
        {
            if (thumbBoxArray == null) return;

            for (int row = 0; row < totalThumbRows; row++)
            {
                for (int col = 0; col < totalThumbColumns; col++)
                {
                    thumbBoxArray[row, col].Redraw();
                }
            }
        }

        public void DeselectLastThumbBox(ThumbBox xSelectedThumbBox)
        {
            if (currentSelectedThumbBox != null && currentSelectedThumbBox != xSelectedThumbBox)
            {
                currentSelectedThumbBox.thisHasFocus = false;
                currentSelectedThumbBox.Redraw();
            }
            currentSelectedThumbBox = xSelectedThumbBox;
        }

        private void TSM_OpenFile_Click(object sender, EventArgs e)
        {
            currentXMLFileName = "";
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "XML Files|*.xml";
            openFileDialog.Title = "Select an XML File";
            
            if(openFileDialog.FileName == "") openFileDialog.FileName = transformationXMLDirectory;

            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {

                BTN_Process.Enabled = false;
                DisableButtons();
                BTN_Stop.Enabled = false;

                currentXMLFileName = openFileDialog.FileName;
                XmlDocument xmlDoc = new XmlDocument();
                xmlDoc.Load(currentXMLFileName);

                if (xmlDoc.GetElementsByTagName("FileName")[0] == null)
                {
                    MessageBox.Show("Invalid batch XML file.");
                }
                else
                {
                    ClearList(scanListArray);
                    XmlNodeList scanList = xmlDoc.GetElementsByTagName("Scan");
                    totalFilesInList = scanList.Count;
                    int fileCount = 0;
                    fileList = new String[totalFilesInList, 3];

                    foreach (XmlNode node in scanList)
                    {
                        XmlElement scanElement = (XmlElement)node;
                        xmlScanNode.fileName = scanElement.GetElementsByTagName("FileName")[0].InnerText;
                        
                        xmlScanNode.controlMarkLayout = scanElement.GetElementsByTagName("ControlMarkLayout")[0].InnerText;

                        xmlScanNode.longitude = scanElement.GetElementsByTagName("Longitude")[0].InnerText;
                        xmlScanNode.latitude = scanElement.GetElementsByTagName("Latitude")[0].InnerText;

                        xmlScanNode.resolution = scanElement.GetElementsByTagName("Resolution")[0].InnerText;
                        xmlScanNode.controlMarkSize = scanElement.GetElementsByTagName("ControlMarkSize")[0].InnerText;


                        if (scanElement.GetElementsByTagName("ControlMarkSpacing")[0] != null)
                        {
                            xmlScanNode.controlMarkSpacing = scanElement.GetElementsByTagName("ControlMarkSpacing")[0].InnerText;
                            xmlScanNode.controMarkLatSpacing = xmlScanNode.controlMarkSpacing;
                            xmlScanNode.controMarkLonSpacing = xmlScanNode.controlMarkSpacing;
                        }
                        if (scanElement.GetElementsByTagName("ControlMarkLonSpacing")[0] != null)
                        {
                            xmlScanNode.controMarkLonSpacing = scanElement.GetElementsByTagName("ControlMarkLonSpacing")[0].InnerText;
                        }
                        if (scanElement.GetElementsByTagName("ControlMarkLatSpacing")[0] != null)
                        {
                            xmlScanNode.controMarkLatSpacing = scanElement.GetElementsByTagName("ControlMarkLatSpacing")[0].InnerText;
                        }

                        if (scanElement.GetElementsByTagName("PerimeterMarksOnly")[0] != null)
                        {
                            xmlScanNode.perimeterMarksOnly = scanElement.GetElementsByTagName("PerimeterMarksOnly")[0].InnerText;
                        }
                        else xmlScanNode.perimeterMarksOnly = "False";

                        if (scanElement.GetElementsByTagName("MapName")[0] != null)
                        {
                            xmlScanNode.mapName = scanElement.GetElementsByTagName("MapName")[0].InnerText;
                        }
                        else xmlScanNode.mapName = null;

                        if (scanElement.GetElementsByTagName("Datum")[0] != null)
                        {
                            xmlScanNode.datum = scanElement.GetElementsByTagName("Datum")[0].InnerText;
                        }
                        else xmlScanNode.datum = null;



                        if (scanElement.GetElementsByTagName("OutputGeographicFileName")[0] != null)
                        {
                            xmlScanNode.outputGeographicFileName = scanElement.GetElementsByTagName("OutputGeographicFileName")[0].InnerText;
                        }
                        else
                        {
                            xmlScanNode.outputGeographicFileName = null;
                        }
                        if (scanElement.GetElementsByTagName("Scale")[0] != null)
                        {
                            xmlScanNode.mapScale = scanElement.GetElementsByTagName("Scale")[0].InnerText;
                        }
                        else
                        {
                            xmlScanNode.mapScale = null;
                        }

                        scanListArray.Add(xmlScanNode);
                        fileCount++;
                    }
                    FillFileListBox();
                    BTN_Process.Enabled = true;

                }
            }
        }
        public double ParseDegreeString(string Dstring)
        {

            double value = 0, one = 1;

            string[] fields = Dstring.Split(new char[] {' '});

            if (Dstring.Contains('-')) one = -1;


            if (fields.Length > 0) value  = one * Math.Abs(Convert.ToDouble(fields[0]));
            if (fields.Length > 1) value += one * Math.Abs(Convert.ToDouble(fields[1])) / 60.0;
            if (fields.Length > 2) value += one * Math.Abs(Convert.ToDouble(fields[2])) / 3600.0;

            return value;

        }
        public double ParseControlMarkSize(string SizeString)
        {
            int len = SizeString.ToLower().IndexOf("i");
            
            if(len < 1) len = SizeString.ToLower().IndexOf("p");
            if(len < 1) len = SizeString.ToLower().IndexOf("f");
            if(len < 1) len = SizeString.Length;

            return Convert.ToDouble(SizeString.Substring(0,len));
        }

        public static void ClearList(IList iList)
        {
            iList.Clear();
        }

        private void FillFileListBox()
        {
            LBX_ListOfFiles.Items.Clear();

            for (int i = 0; i < totalFilesInList; i++)
            {
                LBX_ListOfFiles.Items.AddRange(new object[] { ((XmlScanNode)scanListArray[i]).fileName });
            }

            LBX_ListOfFiles.ClearSelected();
            LBX_ListOfFiles.SelectedIndex = 0;
            LBX_ListOfFiles.Refresh();
        }

        private void LBX_ListOfFiles_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (LBX_ListOfFiles.SelectedIndex > -1)
            {

                DisableButtons();

                xmlScanNode = (XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex];

                LBL_FileNameText.Text = xmlScanNode.fileName.ToString();

                LBL_LonControlMarkSpacingText.Text = xmlScanNode.controMarkLonSpacing;
                LBL_LatControlMarkSpacingText.Text = xmlScanNode.controMarkLatSpacing;
                LBL_PerimeterMarksOnly.Text             = xmlScanNode.perimeterMarksOnly;
                LBL_ControlMarkLayoutText.Text = xmlScanNode.controlMarkLayout.ToString();
                LBL_LongitudeText.Text = xmlScanNode.longitude; 
                LBL_LatitudeText.Text = xmlScanNode.latitude;
                LBL_ResolutionText.Text = xmlScanNode.resolution;
                LBL_ControlMarkSizeText.Text = xmlScanNode.controlMarkSize;
                LBL_Datum.Text = xmlScanNode.datum;

                imageDPI = Convert.ToInt32(xmlScanNode.resolution);

                perimeterMarksOnly = xmlScanNode.perimeterMarksOnly.ToUpper().Contains("T");

                currentFileSelected = LBX_ListOfFiles.SelectedIndex;

                LBL_MapNameText.Text = xmlScanNode.mapName == null ? "" : xmlScanNode.mapName;
                LBL_Scale.Text = xmlScanNode.mapScale == null ? "" : xmlScanNode.mapScale;
                LBL_OutputFile.Text = xmlScanNode.outputGeographicFileName == null ? "" : xmlScanNode.outputGeographicFileName;
              

                if (runningAutoMode)
                {
                    BTN_Process.Enabled = false;
                    BTN_Stop.Enabled = true;
                    PBR_ProgressTransform.Value = Convert.ToInt32(((Convert.ToDouble(LBX_ListOfFiles.SelectedIndex) / Convert.ToDouble(LBX_ListOfFiles.Items.Count)) * 100));
                }
                else
                {
                    BTN_Process.Enabled = true;
                    BTN_Stop.Enabled = false;
                }
            }
        }

        void PNL_ThumbArea_SizeChanged(object sender, EventArgs e)
        {

            if (thumbBoxesCreated)
            {
                AdjustThumbSizes();
                for (int row = 0; row < totalThumbRows; row++)
                {
                    for (int col = 0; col < totalThumbColumns; col++)
                    {
                        if(thumbBoxArray[row,col] != null)
                        {
                            thumbBoxArray[row, col].Location = new System.Drawing.Point(CenterThumbBoxesVertical(col), CenterThumbBoxesHorizontal(row));
                        }
                        
                    }
                }
            }
            int desiredWidth = PNL_BottomLeft.Width - 20;
            TCL_Feedback.Width = desiredWidth;
            TBP_Output.Width = desiredWidth;
            TBP_Error.Width = desiredWidth;
            TBX_Output.Width = desiredWidth - 20;
            TBX_Errors.Width = desiredWidth - 20;
        }

        private void RBN_Color_CheckedChanged(object sender, EventArgs e)
        {
            if (RBN_Color.Checked)
            {
                thumbBoxesAreColor = true;
            }
            else
            {
                thumbBoxesAreColor = false;
            }
            RedrawThumbBoxes();
        }

        private void preferencesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            preferencesDialog.Text = "Preferences File: " + Path.GetFileName(preferencesXMLFile);
            preferencesDialog.ShowDialog();
        }

        public void AdjustThumbSizes()
        {
            if (prefSizeThumbHeight * totalThumbRows <= 10 + PNL_ThumbArea.Height &&
               prefSizeThumbWidth * totalThumbColumns <= 10 + PNL_ThumbArea.Width)
            {
                sizeThumbWidth  = prefSizeThumbWidth;
                sizeThumbHeight = prefSizeThumbHeight;
            }
            else
            {
                double aspect = (double)prefSizeThumbWidth / prefSizeThumbHeight;
                if (prefSizeThumbHeight * totalThumbRows > 10 + PNL_ThumbArea.Height)
                {
                    sizeThumbHeight = (int) ((double) PNL_ThumbArea.Height-10.0)/totalThumbRows;
                    sizeThumbWidth =  (int) Math.Round(sizeThumbHeight*aspect);
                }
                if(prefSizeThumbWidth * totalThumbColumns > 10 + PNL_ThumbArea.Width)
                {
                    sizeThumbWidth  = (int)((double)PNL_ThumbArea.Width - 10.0) / totalThumbColumns;
                    sizeThumbHeight = (int) Math.Round(sizeThumbWidth/aspect);
                }
            }
        }


        public void UpdatePreferences()
        {

            prefSizeThumbWidth  = preferencesDialog.sizeThumbWidth;
            prefSizeThumbHeight = preferencesDialog.sizeThumbHeight;

            searchActualSize    = preferencesDialog.sizeSearchWindow;


            AdjustThumbSizes();
            if(thumbBoxArray != null)
            {
                for (int row = 0; row < totalThumbRows; row++)
                {
                    for (int col = 0; col < totalThumbColumns; col++)
                    {
                        thumbBoxArray[row, col].thisWidth = sizeThumbWidth;
                        thumbBoxArray[row, col].thisHeight = sizeThumbHeight;
                        thumbBoxArray[row, col].thisXPos = CenterThumbBoxesVertical(col);
                        thumbBoxArray[row, col].thisYPos = CenterThumbBoxesHorizontal(row);
                        thumbBoxArray[row, col].EstablishProperties();
                        //thumbBoxArray[row, col].Location = new System.Drawing.Point(CenterThumbBoxesVertical(col), CenterThumbBoxesHorizontal(row));
                    }
                }
            }

            DisableButtons();


            crosshairColor = preferencesDialog.crosshairColor;
            predictionColor = preferencesDialog.predictionColor;

            threshold1 = preferencesDialog.threshold1;
            threshold2 = preferencesDialog.threshold2;
            threshold3 = preferencesDialog.threshold3;
            threshold1Color = preferencesDialog.threshold1Color;
            threshold2Color = preferencesDialog.threshold2Color;
            threshold3Color = preferencesDialog.threshold3Color;
            threshold4Color = preferencesDialog.threshold4Color;
            LBL_TransformationDirectory.Text = preferencesDialog.transformationXMLDirectory;
            LBL_TransformedImageDirectory.Text = preferencesDialog.transformedImageDirectory;
            LBL_ErrorDirectory.Text = preferencesDialog.errorDirectory;
            transformationXMLDirectory = preferencesDialog.transformationXMLDirectory;
            transformedImageDirectory = preferencesDialog.transformedImageDirectory;
            errorDirectory = preferencesDialog.errorDirectory;
            sourceImageDirectory = preferencesDialog.sourceImageDirectory;


            noRotation = preferencesDialog.noRotation;
            noQA       = preferencesDialog.noQA;
            noGeoref   = preferencesDialog.noGeoref;
            linearOnly = preferencesDialog.linearOnly;

            /* don't disable...need to way to write ouput info
            if (noGeoref)
            {
                BTN_ExportImage.Enabled = false;
            }
            */
           

            RedrawThumbBoxes();
        }

        //*///////////////////////////////////////////////////////////////////////////////////////////////
        public void MakeControlMarks()
        {
            //fill control mark array with standard marks
            int ticLength;

            controlMark = new CMark[totalThumbRows, totalThumbColumns];

            double dtic = ParseControlMarkSize(xmlScanNode.controlMarkSize);

            //mark size might be supplied as a fraction of the image height or width
            if ((xmlScanNode.controlMarkSize.ToLower()).Contains("f"))
            {
                ticLength = (int) Math.Round(dtic * imageWidth);
            }

            //or pixels
            else if((xmlScanNode.controlMarkSize.ToLower()).Contains("p"))
            {
                ticLength = (int)Math.Round(dtic);

            }
            
            //or it might be given in inches (again, we want pixels)
            else
            {
                ticLength     = (int) Math.Round(dtic * imageDPI);
            }



            //make sure mark will fit comfortably in search window
            ticLength = Math.Min(ticLength, (int)Math.Round((double) searchActualSize/6.0));

            totalThumbBoxes = 0;

            for (int row = 0; row < totalThumbRows; row++)
            {
                for (int col = 0; col < totalThumbColumns; col++)
                {
                    controlMark[row, col] = new CMark();

                    totalThumbBoxes++;

                    if (row == 0 && col == 0)
                        controlMark[row, col].Type = MarkType.BotLeftCorn;

                    else if (row == 0 && col == totalThumbColumns - 1)
                        controlMark[row, col].Type = MarkType.BotRightCorn;

                    else if (row == totalThumbRows - 1 && col == 0)
                        controlMark[row, col].Type = MarkType.TopLeftCorn;

                    else if (row == totalThumbRows - 1 && col == totalThumbColumns - 1)
                        controlMark[row, col].Type = MarkType.TopRightCorn;

                    else if (row == 0)
                        controlMark[row, col].Type = MarkType.BotTee;

                    else if (row == totalThumbRows - 1)
                        controlMark[row, col].Type = MarkType.TopTee;

                    else if (col == 0)
                        controlMark[row, col].Type = MarkType.LeftTee;

                    else if (col == totalThumbColumns - 1)
                        controlMark[row, col].Type = MarkType.RightTee;


                    else if (perimeterMarksOnly == false)
                        controlMark[row, col].Type = MarkType.Plus;

                    else
                    {
                        controlMark[row, col].Type = MarkType.None;
                        totalThumbBoxes--;
                    }


                    double lonSpacing = ParseDegreeString(xmlScanNode.controMarkLonSpacing);
                    double latSpacing = ParseDegreeString(xmlScanNode.controMarkLatSpacing);
                    double lat     = ParseDegreeString(xmlScanNode.latitude);
                    double lon     = ParseDegreeString(xmlScanNode.longitude);

                    controlMark[row, col].SetMarkSize(2*ticLength,ticLength);
                    controlMark[row, col].xImage  = 0;
                    controlMark[row, col].yImage  = 0;
                    controlMark[row, col].xSW     = 0;
                    controlMark[row, col].ySW     = 0;
                    controlMark[row, col].latTrue = lat + row * latSpacing;
                    controlMark[row, col].lonTrue = lon + (col - (totalThumbColumns - 1)) * lonSpacing;
                }
            }
        }

        private bool CheckDirectories()
        {
            if (Directory.Exists(transformationXMLDirectory) &&
                Directory.Exists(transformedImageDirectory) &&
                Directory.Exists(errorDirectory) &&
                Directory.Exists(sourceImageDirectory))
            {
                return true;
            }
            else
            {
                return false;
            }

        }

        private void BTN_Process_Click(object sender, EventArgs e)
        {
            string s;


            imageFile.CloseFiles();


            if (!CheckDirectories())
            {
                MessageBox.Show("At least one of the output directories does not exist.\nPlease check paths in preferences.");
                return;
            }
            if (RBN_Automatic.Checked == true)
            {
                runningAutoMode = true;
                BTN_Process.Enabled = false;

                DisableButtons();

                BTN_Stop.Enabled = true;
                LBX_ListOfFiles.Enabled = false;
                GBX_Mode.Enabled = false;
                GBX_Extent.Enabled = false;
            }
            else
            {
                runningAutoMode = false;
                DisableButtons();
            }

            if (sender is Button)
            {
                DateTime timeNow = DateTime.Now;

                errorFileNameSuffix =
                    string.Format("{0:d4}_{1:d2}_{2:d2}_{3:d2}_{4:d2}_{5:d2}",
                     timeNow.Year,timeNow.Month,timeNow.Day,timeNow.Hour,timeNow.Minute,timeNow.Second);
                
            }

            xmlScanNode = (XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex];

            string[] findLayout = xmlScanNode.controlMarkLayout.ToLower().Split('x');
            totalThumbRows = Convert.ToInt32(findLayout[0]);
            totalThumbColumns = Convert.ToInt32(findLayout[1]);

            
           

            mapScale = Convert.ToDouble(xmlScanNode.mapScale);



            s = preferencesDialog.sourceImageDirectory + "/" + xmlScanNode.fileName;

            if (!File.Exists(s))
            {
                if (!RBN_Automatic.Checked)
                {
                    MessageBox.Show(s + ": Image file doesn't exist." + newLine);
                    return;
                }
                else
                {
                    TBX_Errors.AppendText(s + " : Image file doesn't exist" + newLine);
                    TBX_Output.AppendText(s + " : Image file doesn't exist" + newLine);

                    AppendErrorFile();
                    automaticTimer = new System.Windows.Forms.Timer();
                    automaticTimer.Interval = 2000;
                    automaticTimer.Tick += new System.EventHandler(AutomaticTimerTick);
                    automaticTimer.Enabled = true;
                    return;
                }
            }

            MakeControlMarks();

            if (totalThumbBoxes < 7 || linearOnly) degreeFit = 1;
            else                                   degreeFit = 2;

            try
            {

                if (!imageFile.LoadFile(s, converterProc))
                {
                    if (!RBN_Automatic.Checked)
                    {
                        MessageBox.Show("Couldn't open image file\n" + s);
                        return;
                    }
                    else
                    {
                        TBX_Errors.AppendText(s + " : Couldn't open image file\n");
                        AppendErrorFile();
                        automaticTimer = new System.Windows.Forms.Timer();
                        automaticTimer.Interval = 2000;
                        automaticTimer.Tick += new System.EventHandler(AutomaticTimerTick);
                        automaticTimer.Enabled = true;
                        return;
                    }
                }
                ComputeQuadInfo();

            }
            catch
            {
                if (!RBN_Automatic.Checked)
                {
                    MessageBox.Show("Exception when opening image file\n" + s);
                    return;
                }
                else
                {
                    TBX_Errors.AppendText(s + " : Exception when opening image file\n");
                    AppendErrorFile();
                    automaticTimer = new System.Windows.Forms.Timer();
                    automaticTimer.Interval = 2000;
                    automaticTimer.Tick += new System.EventHandler(AutomaticTimerTick);
                    automaticTimer.Enabled = true;
                    return;
                }
            }

            //edge searches are used to guess corner locations
            //the edgeInfo structures are needed for edge searches
            //for (int i = 0; i < 4; i++) SetEdgeInfo4(i, out edgeInfo[i]);

            bitmapWidthPixels = imageWidth;
            bitmapHeightPixels = imageHeight;

            bitmapWidthBytes = imageWidth * 3;
            bitmapHeightBytes = imageHeight * 3;

            TBX_Output.AppendText(newLine+newLine+"Starting " + xmlScanNode.fileName+newLine);
            if(imageFile.needRotation)
               TBX_Output.AppendText("Map appears to be scanned in transverse orientation...rotating by 90 degrees" + newLine);

            //search for edges and extrapolate corner locations from same
            if (FindCorners4Way())
            {
                //clear thumb boxes and redraw
                AdjustThumbSizes();
                CreateThumbBoxes();


                if (RBN_Manual.Checked)
                {

                    FindWindow(0, 0);
                    FindWindow(0, totalThumbColumns - 1);
                    FindWindow(totalThumbRows - 1, 0);
                    FindWindow(totalThumbRows - 1, totalThumbColumns - 1);
                    //showThumbInfo = true;
                    

                    BTN_Process.Enabled = false;
                    BTN_FindWindows.Enabled = true;
                    
                }
                else
                {
                    for (int row = 0; row < totalThumbRows; row++)
                    {
                        for (int col = 0; col < totalThumbColumns; col++)
                            FindWindow(row, col);
                    }
                   
                    SearchAllWindows();
                    
                    PerformLSFit();



                    if (CheckBelowThreshold() && (converterProc.Length > 10 && !noGeoref || DemoOnly))
                    {
                        TransformImage();
                    }

                    WriteOutputFile();

                    automaticTimer = new System.Windows.Forms.Timer();
                    automaticTimer.Interval = 2000;
                    automaticTimer.Tick += new System.EventHandler(AutomaticTimerTick);
                    automaticTimer.Enabled = true;
                    
                }
            }


        }
        public void ComputeQuadInfo()
        {
            double imageAspect;

            double lonSize = Math.Abs(controlMark[0, totalThumbColumns - 1].lonTrue - controlMark[0, 0].lonTrue);
            double latSize = Math.Abs(controlMark[totalThumbRows - 1, 0].latTrue    - controlMark[0, 0].latTrue);

            double latMid = (controlMark[totalThumbRows - 1, 0].latTrue + controlMark[0, 0].latTrue) / 2.0;

            double dlat, dlon;
            
            DegreeLength(latMid, out dlon, out dlat);

            quadAspect = (lonSize * dlon) / (latSize*dlat);

            if (mapScale < 100) quadHeight = quadWidth = 1e80;

            else
            {
                //             dots per inch *  meters    *   in/m    / (earth in per map inch)
                quadHeight = imageDPI * (dlat * latSize) * 39.3700787 / mapScale;

                quadWidth = imageDPI * (dlon * lonSize) * 39.3700787 / mapScale;

            }
            imageAspect = (double) imageFile.Width / imageFile.Height;

            if (!noRotation && totalThumbColumns > totalThumbRows &&
                (quadAspect - imageAspect) / quadAspect > .3)
            {
                imageFile.needRotation = true;
                imageWidth = imageFile.Height;
                imageHeight = imageFile.Width;
            }
            else
            {
                imageFile.needRotation = false;
                imageWidth = imageFile.Width;
                imageHeight = imageFile.Height;
            }

            

        }
        public void HandleImageOpenErrors(string errorMessage)
        {
            automaticTimer = new System.Windows.Forms.Timer();
            automaticTimer.Interval = 2000;
            automaticTimer.Tick += new System.EventHandler(AutomaticTimerTick);
            automaticTimer.Enabled = true;
        }

        private void Original2SearchWindow(PT pt, SearchInfo SW, out PT ptSW)
        {
            //convert from original image coordinates to search window
            ptSW.x = (int)Math.Round((double)(pt.x - SW.imageRect.Left) * SW.searchRatio);
            ptSW.y = (int)Math.Round((double)(pt.y - SW.imageRect.Top) * SW.searchRatio);

        }

        private void SearchWindow2Original(PT ptSW, SearchInfo SW, out PT pt)
        {
            //convert from search window coordinates to scanned image coordinates
            pt.x = SW.imageRect.Left + (int)Math.Round((double)ptSW.x / SW.searchRatio);
            pt.y = SW.imageRect.Top + (int)Math.Round((double)ptSW.y / SW.searchRatio);

        }

        private void Original2DisplaySearchWindow(PT pt, SearchInfo SW, out PT dw)
        {
            //convert from original image coordinates to display of search window
            PT ptSW;

            Original2SearchWindow(pt, SW, out ptSW);

            dw.x = (int)Math.Round((double)ptSW.x * DisplayRectangle.Width /
                      (SW.imageRect.Width * SW.searchRatio));

            dw.y = (int)Math.Round((double)ptSW.y * DisplayRectangle.Height /
                      (SW.imageRect.Height * SW.searchRatio));

        }

        public void DrawThumbPanelImageBox()
        {
            using (Graphics g = PNL_ThumbArea.CreateGraphics())
            {
                g.FillRectangle(Brushes.White,
                    new Rectangle(Image2PNL(0, 0).X, Image2PNL(0, 0).Y,
                                  Image2PNL(imageWidth - 1, 0).X  - Image2PNL(0, 0).X,
                                  Image2PNL(0, imageHeight - 1).Y - Image2PNL(0, 0).Y));
            }
        }
        public void ClearThumbPanel(bool ShowImageBox)
        {
            PNL_ThumbArea.Controls.Clear();
            using (Graphics g = PNL_ThumbArea.CreateGraphics())
            {
                g.FillRectangle(new SolidBrush(PNL_ThumbArea.BackColor),
                        new Rectangle(0, 0, ClientSize.Width, ClientSize.Height));

                if(ShowImageBox)
                {
                    g.FillRectangle(Brushes.White,
                        new Rectangle(Image2PNL(0, 0).X, Image2PNL(0, 0).Y,
                                      Image2PNL(imageWidth - 1, 0).X - Image2PNL(0, 0).X,
                                      Image2PNL(0, imageHeight - 1).Y - Image2PNL(0, 0).Y));
                }

            }

        }
        public Point Image2PNL(PT P)
        {
            return Image2PNL(P.x, P.y);
        }
        
        public Point Image2PNL(int X, int Y)
        {

            double scaleXY = Math.Min((double)PNL_ThumbArea.ClientRectangle.Width / imageWidth,
                                      (double)PNL_ThumbArea.ClientRectangle.Height / imageHeight);

            int qWidth = (int)(imageWidth * scaleXY);
            int qHeight = (int)(imageHeight * scaleXY);
            int x0 = (PNL_ThumbArea.ClientRectangle.Width - qWidth) / 2;
            int y0 = (PNL_ThumbArea.ClientRectangle.Height - qHeight) / 2;

            return new Point((int)(x0 + X * scaleXY), (int)(y0 + Y * scaleXY));
        }

        public List<PT> FindPossEdges(CMark Mark, Rectangle Rect, string LogFileName)
        {

            List<PT> ptsAbsolute = new List<PT>();

            Bitmap bmp = new Bitmap(1,1);

            int xSize = Image2PNL(Rect.Right, Rect.Top).X   -
                        Image2PNL(Rect.Left,  Rect.Top).X;

            int ySize = Image2PNL(Rect.Left, Rect.Bottom).Y -
                        Image2PNL(Rect.Left, Rect.Top).Y;

            int stripSize = 50;

            bmp = imageFile.ExtractRectangle(Rect);

            if (bmp == null) return ptsAbsolute;

            using (Graphics g = PNL_ThumbArea.CreateGraphics())
            {
                if (Mark.Type == MarkType.LeftEdge || Mark.Type == MarkType.RightEdge)
                {
                    g.DrawImage(bmp, new Rectangle(
                         Image2PNL(Rect.Left, Rect.Top), new Size(xSize, stripSize)));
                }

                else if (Mark.Type == MarkType.TopEdge || Mark.Type == MarkType.BotEdge)
                {
                    g.DrawImage(bmp, new Rectangle(
                         Image2PNL(Rect.Left, Rect.Top), new Size(stripSize, ySize)));
                }
            }

            Application.DoEvents();


            List<PT> ptsRelative = Mark.FindPossEdges(bmp, LogFileName);


            foreach (PT pt in ptsRelative)
            {
                if (Mark.Type == MarkType.LeftEdge || Mark.Type == MarkType.RightEdge)
                    ptsAbsolute.Add(new PT(pt.x + Rect.Left,         (Rect.Top + Rect.Bottom) / 2));
                
                else
                    ptsAbsolute.Add(new PT((Rect.Left+Rect.Right)/2, pt.y + Rect.Top));
            }
           
            return ptsAbsolute;

        }
        
        [System.Runtime.InteropServices.DllImportAttribute("gdi32.dll")]
        private static extern int BitBlt(
               IntPtr hdcDest,     // handle to destination DC (device context)
               int nXDest,         // x-coord of destination upper-left corner
               int nYDest,         // y-coord of destination upper-left corner
               int nWidth,         // width of destination rectangle
               int nHeight,        // height of destination rectangle
               IntPtr hdcSrc,      // handle to source DC
               int nXSrc,          // x-coordinate of source upper-left corner
               int nYSrc,          // y-coordinate of source upper-left corner
               System.Int32 dwRop  // raster operation code
        );
        public void SaveAsPng(Control control, string fileName)
        {

            const int SRCCOPY = 0xcc0020;
            Bitmap bmp;

            
            if (fileName.Length < 1) return;

            if(!fileName.EndsWith(".png")) fileName += ".png";


            using (Graphics g1 = control.CreateGraphics())
            {
                bmp = new Bitmap(control.Width, control.Height, g1);
                using (Graphics g2 = Graphics.FromImage(bmp))
                {
                    IntPtr dc1 = g1.GetHdc();
                    IntPtr dc2 = g2.GetHdc();

                    BitBlt(dc2, 0, 0, control.Width, control.Height, dc1, 0, 0, SRCCOPY);
                    g1.ReleaseHdc(dc1);
                    g2.ReleaseHdc(dc2);
                }
                bmp.Save(fileName, ImageFormat.Png);
            }


        }
        public bool FindCorners4Way()
        {
            //searches for neatlines and uses their intersections as guess for neatline corners
            
            CMark    edgePattern = new CMark(); //used for edge searches
            
            //neatlines guaranteed to be outside this rectangle
            int innerTop   = (int)(MaxTopMargin           * imageHeight);
            int innerBot   = (int)((1.0 - MaxBotMargin)   * imageHeight);
            int innerLeft  = (int)(MaxLeftMargin          * imageWidth);
            int innerRight = (int)((1.0 - MaxRightMargin) * imageWidth);

            int outterLeft  = 0;
            int outterRight = imageWidth - 1;
            int outterTop   = 0;
            int outterBot   = imageHeight - 1;
            
            Rectangle  rect;

            ClearThumbPanel(true);

            int x0, y0, x1, y1;

            //max pixel departures from horizontal and vertical over 1/2 width or 1/2 height
            int dy = (int) (.5*(innerRight-innerLeft)*Math.Tan(maxRotation * Math.PI / 180.0));
            int dx = (int) (.5*(innerBot - innerTop) *Math.Tan(maxRotation * Math.PI / 180.0));

            List<EdgeSeg> leftEdges  = new List<EdgeSeg>();
            List<EdgeSeg> rightEdges = new List<EdgeSeg>();
            List<EdgeSeg> topEdges   = new List<EdgeSeg>();
            List<EdgeSeg> botEdges   = new List<EdgeSeg>();

            List<PT> xy1 = new List<PT>();
            List<PT> xy2 = new List<PT>();


            List<PT> xyMidLeft  = new List<PT>();
            List<PT> xyMidTop   = new List<PT>();
            List<PT> xyMidRight = new List<PT>();
            List<PT> xyMidBot   = new List<PT>();


            //first find possible edge locations for the 4 edges
            //look at "center" of map for possible edges

            edgePattern.Type = MarkType.LeftEdge;
            xyMidLeft = FindPossEdges(edgePattern, 
                     new Rectangle(outterLeft,              (innerTop+innerBot)/2,
                                   innerLeft-outterLeft-1,  200),  "");//"LeftEdgeCenter");

            TBX_Output.AppendText("Finished edge...");
            TBX_Output.Refresh();
            Application.DoEvents();


            //sections below need generalizing
            if (xyMidLeft.Count < 1)
            {
                TBX_Errors.AppendText("Doubling left edge search area for" + LBL_FileNameText.Text + newLine);

                int m = innerLeft - outterLeft;
                innerLeft += m;
                outterLeft += m;
                xyMidLeft = FindPossEdges(edgePattern,
                     new Rectangle(outterLeft, (innerTop + innerBot) / 2,
                                   innerLeft - outterLeft - 1, 200), "");// "LeftEdgeCenter");


                if (xyMidLeft.Count < 1)
                {
                    xyMidLeft.Add(new PT(innerLeft, (innerTop + innerBot) / 2));

                    TBX_Errors.AppendText(
                        String.Format("Left edge apparently not w/in {0:d} pixels of image boundary",
                        innerLeft)
                        + newLine);
                }
            }
            edgePattern.Type = MarkType.RightEdge;
            xyMidRight = FindPossEdges(edgePattern,
                     new Rectangle(innerRight,               (innerTop + innerBot) / 2, 
                                   outterRight-innerRight-1,  200),  "");// "RightEdgeCenter");

            if (xyMidRight.Count < 1)
            {
                TBX_Errors.AppendText("Doubling right edge search area for " + LBL_FileNameText.Text + newLine);
                int m = outterRight - innerRight;
                innerRight -= m;
                outterRight -= m;
                xyMidRight = FindPossEdges(edgePattern,
                     new Rectangle(innerRight, (innerTop + innerBot) / 2,
                                   outterRight - innerRight - 1, 200), "");// "RightEdgeCenter");

                if (xyMidRight.Count < 1)
                {
                    xyMidRight.Add(new PT(innerRight, (innerTop + innerBot) / 2));
                    TBX_Errors.AppendText(
                        String.Format("Right edge apparently not w/in {0:d} pixels of image boundary",
                        imageWidth - innerRight)
                        + newLine);
                }
            }

            edgePattern.Type = MarkType.TopEdge;
            xyMidTop = FindPossEdges(edgePattern,
                     new Rectangle((innerLeft+innerRight)/2, outterTop, 
                                    200,                     innerTop-outterTop), "");//"TopEdgeCenter");

            if (xyMidTop.Count < 1)
            {
                TBX_Errors.AppendText("Doubling top edge search area for " + LBL_FileNameText.Text + newLine);
                int m = innerTop - outterTop;
                innerTop += m;
                outterTop += m;
                xyMidTop = FindPossEdges(edgePattern,
                     new Rectangle((innerLeft + innerRight) / 2, outterTop,
                                    200, innerTop - outterTop), "");//"TopEdgeCenter");

                if (xyMidTop.Count < 1)
                {
                    xyMidTop.Add(new PT((innerLeft + innerRight) / 2, innerTop));
                    TBX_Errors.AppendText(
                        String.Format("Top edge apparently not w/in {0:d} pixels of image boundary",
                        innerTop)
                        + newLine);
                }
            }


            edgePattern.Type = MarkType.BotEdge;
            xyMidBot = FindPossEdges(edgePattern,
                       new Rectangle((innerLeft + innerRight) / 2, innerBot,
                                      200,                         outterBot-innerBot-1), "");//"BotEdgeCenter");

            if (xyMidBot.Count < 1)
            {
                TBX_Errors.AppendText("Doubling bottom search area for " + LBL_FileNameText.Text + newLine);
                int m = outterBot - innerBot;
                innerBot -= m;
                outterBot -= m;

                xyMidBot = FindPossEdges(edgePattern,
                       new Rectangle((innerLeft + innerRight) / 2, innerBot,
                                      200, outterBot - innerBot - 1),  "");//"BotEdgeCenter");
                if (xyMidBot.Count < 1)
                {
                    xyMidBot.Add(new PT((innerLeft + innerRight) / 2, innerBot));
                    TBX_Errors.AppendText(
                        String.Format("Bottom edge apparently not w/in {0:d} pixels of image boundary",
                        imageHeight - innerBot)
                        + newLine);
                }
            }


            outterLeft = Math.Max(outterLeft-dx, xyMidLeft[0].x - dx);
            outterLeft = Math.Max(0, outterLeft);

            innerLeft  = Math.Min(innerLeft+dx,  xyMidLeft.Last().x + dx);


            outterRight = Math.Min(outterRight+dx, xyMidRight.Last().x + dx);
            outterRight = Math.Min(imageWidth - 1, outterRight);

            innerRight  = Math.Max(innerRight-dx, xyMidRight[0].x - dx);


            outterTop = Math.Max(outterTop-dy, xyMidTop[0].y - dy);
            outterTop = Math.Max(0, outterTop);

            innerTop  = Math.Min(innerTop+dy,   xyMidTop.Last().y + dy);


            outterBot = Math.Min(outterBot+dy, xyMidBot.Last().y + dy);
            outterBot = Math.Min(imageHeight - 1, outterBot);

            innerBot  = Math.Max(innerBot-dy,  xyMidBot[0].y - dy);


            //cycle through all 4 edge searches
            for (int i = 0; i < 4; i++)
            {
                int iMain = -100 * (i + 1);

                xy1.Clear();
                xy2.Clear();
               
                if (i == 0) 
                {
                    edgePattern.Type = MarkType.LeftEdge;

                    x0 = outterLeft;
                    x1 = innerLeft;

                    y0 = innerBot-200;
                    y1 = y0+200;

                    rect = new Rectangle(x0,y0,x1-x0+1,y1-y0+1);
                    xy1 = FindPossEdges(edgePattern, rect, "");//LeftEdgeLower");

                    if (xy1.Count > 0)
                    {
                        x0 = Math.Max(0, xy1[0].x - dx);
                        x1 = Math.Min(innerLeft, xy1.Last().x + dx);

                        y0 = innerTop;
                        y1 = y0 + 200;

                        rect = new Rectangle(x0, y0, x1 - x0 + 1, y1 - y0 + 1);
                        xy2 = FindPossEdges(edgePattern, rect, "");//"LeftEdgeUpper");
                    }
                    
                    if (xy1.Count > 0 && xy2.Count > 0) 
                        leftEdges = IdentifyEdges(xy1, xyMidLeft, xy2);

                    if (leftEdges.Count < 1)
                    {
                        TBX_Errors.AppendText("Missed left edge for" + LBL_FileNameText.Text + newLine);
                        leftEdges.Add(new EdgeSeg(new PT(innerLeft, 0), new PT(innerLeft, 1000)));
                    }
                    DrawEdges(leftEdges);

                }
                else if (i == 1)
                {
                    edgePattern.Type = MarkType.BotEdge;

                    x0 = innerLeft;
                    x1 = x0 + 200;

                    y0 = innerBot;
                    y1 = outterBot;

                    rect = new Rectangle(x0,y0,x1-x0+1,y1-y0+1);

                    xy1 = FindPossEdges(edgePattern, rect, "");//"BotEdgeLeft");

                    if (xy1.Count > 0)
                    {
                        x0 = innerRight - 200;
                        x1 = x0 + 200;

                        y0 = Math.Max(innerBot,  xy1[0].y - dy );
                        y1 = Math.Min(outterBot, xy1.Last().y + dy );

                        rect = new Rectangle(x0, y0, x1 - x0 + 1, y1 - y0 + 1);
                        xy2 = FindPossEdges(edgePattern, rect, "");//"BotEdgeRight");

                    }

                    if (xy1.Count > 0 && xy2.Count > 0)
                    {
                        botEdges = IdentifyEdges(xy1, xyMidBot, xy2);
                    }
                    if (botEdges.Count < 1)
                    {
                        TBX_Errors.AppendText("Missed bottom edge for "  + LBL_FileNameText.Text + newLine);
                        botEdges.Add(new EdgeSeg(new PT(0,innerBot), new PT(1000,innerBot)));
                    }
                    botEdges.Reverse();  //outside to inside order
                    DrawEdges(botEdges);
                }
                
                else if (i == 2)
                {
                    edgePattern.Type = MarkType.TopEdge;
                    
                    x0 = innerLeft;
                    x1 = x0 + 200;

                    y0 = outterTop;
                    y1 = innerTop;

                    rect = new Rectangle(x0,y0,x1-x0+1,y1-y0+1);

                    xy1 = FindPossEdges(edgePattern, rect, "");//"TopEdgeLeft");

                    if (xy1.Count > 0)
                    {
                        x0 = innerRight - 200;
                        x1 = x0 + 200;
                        y0 = Math.Max(outterTop, xy1[0].y - dy);
                        y1 = Math.Min(innerTop,  xy1.Last().y + dy);

                        rect = new Rectangle(x0, y0, x1 - x0 + 1, y1 - y0 + 1);
                        xy2 = FindPossEdges(edgePattern, rect, "");//"TopEdgeRight");
                    }

                    if (xy1.Count > 0 && xy2.Count > 0)
                    {
                        topEdges = IdentifyEdges(xy1, xyMidTop, xy2);
                    }
                    if (topEdges.Count < 1)
                    {
                        TBX_Errors.AppendText("Missed top edge for " + LBL_FileNameText.Text + newLine);
                        topEdges.Add(new EdgeSeg(new PT(0,innerTop), new PT(1000,innerTop)));
                    }
                    DrawEdges(topEdges);
                }
                else
                {
                    edgePattern.Type = MarkType.RightEdge;

                    x0 = innerRight;
                    x1 = outterRight;

                    y0 = innerBot-200;
                    y1 = y0+200;

                    rect = new Rectangle(x0, y0, x1-x0+1, y1-y0+1);

                    xy1 = FindPossEdges(edgePattern, rect, "");//"RightEdgeBot");

                    if (xy1.Count > 0)
                    {
                        x0 = Math.Max(innerRight,  xy1[0].x - dx);
                        x1 = Math.Min(outterRight, xy1.Last().x + dx);

                        y0 = innerTop;
                        y1 = y0 + 200;

                        rect = new Rectangle(x0, y0, x1 - x0 + 1, y1 - y0 + 1);

                        xy2 = FindPossEdges(edgePattern, rect, "");// "RightEdgeTop");

                    }
                    if (xy1.Count > 0 && xy2.Count > 0)
                    {
                         rightEdges = IdentifyEdges(xy1, xyMidRight, xy2);
                    }
                    if (rightEdges.Count < 1)
                    {
                        TBX_Errors.AppendText("Missed right edge for " + LBL_FileNameText.Text + newLine);
                        rightEdges.Add(new EdgeSeg(new PT(innerRight,0),new PT(innerRight,1000)));
                    }
                    rightEdges.Reverse(); //outside to inside order
                    DrawEdges(rightEdges);

                }
                TBX_Output.AppendText(String.Format("{0}...", i + 1));



                TBX_Output.Refresh();
                Application.DoEvents();

            }





            FindQuadCorners(leftEdges, rightEdges, topEdges, botEdges);
            
            TBX_Output.AppendText(newLine +
                String.Format("Left neatline:\tbest match at {0:d5}, border width of {1,4:f1}%.\t Neatline angle {2,7:f2} degrees",
                (corn1PT.x+corn2PT.x)/2, 100.0 * (corn1PT.x+corn2PT.x)/2 / imageWidth, SegAngle(corn1PT, corn2PT))
                +newLine);
            TBX_Output.AppendText(
                String.Format("Right neatline:\tbest match at {0:d5}, border width of {1,4:f1}%.\t Neatline angle {2,7:f2} degrees",
                   (corn3PT.x+corn4PT.x)/2,  100 * (1.0 - (float)(corn3PT.x+corn4PT.x)/2 / imageWidth), SegAngle(corn4PT, corn3PT))
                   + newLine);
            TBX_Output.AppendText(
                String.Format("Top neatline:\tbest match at {0:d5}, border width of {1,4:f1}%.\t Neatline angle {2,7:f2} degrees",
                (corn2PT.y+corn3PT.y)/2, 100.0 * (corn2PT.y+corn3PT.y)/2 / imageHeight, SegAngle(corn2PT, corn3PT))
                + newLine);

            TBX_Output.AppendText(
                String.Format("Bottom neatline:\tbest match at {0:d5}, border width of {1,4:f1}%.\t Neatline angle {2,7:f2} degrees",
                (corn4PT.y+corn1PT.y)/2, 100 * (1.0 - (float)(corn4PT.y+corn1PT.y)/2 / imageHeight), SegAngle(corn1PT, corn4PT))
                    + newLine);



            using (Graphics g = PNL_ThumbArea.CreateGraphics())
            {
                Point p1, p2, p3, p4;
                Pen pen = new Pen(Brushes.DarkCyan);
                pen.Width = 5;
                p1 = Image2PNL(corn1PT.x,corn1PT.y);
                p2 = Image2PNL(corn2PT.x, corn2PT.y);
                p3 = Image2PNL(corn3PT.x, corn3PT.y);
                p4 = Image2PNL(corn4PT.x, corn4PT.y);

                g.DrawLine(pen, p1, p2);
                g.DrawLine(pen, p2, p3);
                g.DrawLine(pen, p3, p4);
                g.DrawLine(pen, p4, p1);

            }
            


            Thread.Sleep(1000);

            ClearThumbPanel(false);

            Application.DoEvents();

            return true;
        }
        public void DrawQuads(List<EdgeSeg> Left, List<EdgeSeg> Top, List<EdgeSeg> Right, List<EdgeSeg> Bot)
        {
            using (Graphics g = PNL_ThumbArea.CreateGraphics())
            {
                Pen pen = new Pen(Brushes.DarkCyan);
                foreach (EdgeSeg left in Left)
                {
                    foreach (EdgeSeg top in Top)
                    {
                        foreach (EdgeSeg right in Right)
                        {
                            foreach (EdgeSeg bot in Bot)
                            {
                                Point p1 = Image2PNL(IntersectEdges(left,   bot));
                                Point p2 = Image2PNL(IntersectEdges(left,   top));
                                Point p3 = Image2PNL(IntersectEdges(top,  right));
                                Point p4 = Image2PNL(IntersectEdges(right,  bot));

                                g.DrawLine(pen, p1, p2);
                                g.DrawLine(pen, p2, p3);
                                g.DrawLine(pen, p3, p4);
                                g.DrawLine(pen, p4, p1);
                            }
                        }
                    }
                }

            }
        }
        public void DrawEdges(List<EdgeSeg> Edges)
        {
            using (Graphics g = PNL_ThumbArea.CreateGraphics())
            {
                Pen pen = new Pen(Brushes.DarkCyan);
                pen.Width = 1;

                foreach (EdgeSeg edge in Edges)
                {
                    PT pabs1, pabs2;
                    if (edge.vertical)
                    {
                        pabs1 = IntersectEdges(edge, new EdgeSeg(new PT(0, 0),                new PT(imageFile.Width, 0)));
                        pabs2 = IntersectEdges(edge, new EdgeSeg(new PT(0, imageFile.Height), new PT(imageFile.Width, imageFile.Height)));

                    }
                    else
                    {
                        pabs1 = IntersectEdges(edge, new EdgeSeg(new PT(0, 0), new PT(0, imageFile.Height)));
                        pabs2 = IntersectEdges(edge, new EdgeSeg(new PT(imageFile.Width,0), new PT(imageFile.Width, imageFile.Height)));
                    }

                    g.DrawLine(pen, Image2PNL(pabs1), Image2PNL(pabs2));
                }
            }

        }


        public void FindQuadCorners(List<EdgeSeg> Left, List<EdgeSeg> Right,
            List<EdgeSeg> Top, List<EdgeSeg> Bot)
        {
            double width, height, err, errMin;

            double topQuadWidth, botQuadWidth, midQuadHeight,quadArea;

            PT p1, p2, p3, p4;

            double topErr, botErr, hErr;

            double dlat, dlon;
            
            double lonLeft  =  controlMark[0, 0].lonTrue;
            double lonRight = controlMark[0, totalThumbColumns - 1].lonTrue;
            double latBot   = controlMark[0, 0].latTrue;
            double latTop   = controlMark[totalThumbRows - 1, 0].latTrue;

            double lonSize = lonRight - lonLeft;
            double latSize = latTop   - latBot;


            if(mapScale > 1)
            {
                DegreeLength((latTop+latBot)/2, out dlon, out dlat);
                midQuadHeight = imageDPI * (dlat * latSize) * 39.3700787 / mapScale;

                DegreeLength(latTop, out dlon, out dlat);
                topQuadWidth = imageDPI * (dlon * lonSize) * 39.3700787 / mapScale;

                DegreeLength(latBot, out dlon, out dlat);
                botQuadWidth = imageDPI * (dlon * lonSize) * 39.3700787 / mapScale;

                quadArea   = midQuadHeight*(topQuadWidth+botQuadWidth)/2;

                quadAspect = 0;
            }

            else
            {
                DegreeLength((latTop+latBot)/2, out dlon, out dlat);
                quadAspect = (lonSize * dlon) / (latSize*dlat);

                midQuadHeight = topQuadWidth = botQuadWidth = quadArea = 0;
            }

            errMin = 1.0e50;
            foreach(EdgeSeg left in Left)
            {
                foreach(EdgeSeg right in Right)
                {
                    foreach(EdgeSeg top in Top)
                    {
                        foreach(EdgeSeg bot in Bot)
                        {

                            p1 = IntersectEdges(left, bot);
                            p2 = IntersectEdges(left, top);
                            p3 = IntersectEdges(right, top);
                            p4 = IntersectEdges(right, bot);

                            if (botQuadWidth > 10)
                            {
                                topErr = Math.Abs((p3.x - p2.x) - topQuadWidth) / topQuadWidth;
                                botErr = Math.Abs((p4.x - p1.x) - botQuadWidth) / botQuadWidth;
                                hErr =   Math.Abs((p1.y - p2.y) - midQuadHeight) / midQuadHeight;
                                hErr +=  Math.Abs((p4.y - p3.y) - midQuadHeight) / midQuadHeight;

                                hErr /= 2.0;

                                err = Math.Abs((p3.x - p2.x) - topQuadWidth) / topQuadWidth +
                                      Math.Abs((p4.x - p1.x) - botQuadWidth) / botQuadWidth +
                                      Math.Abs((p1.y - p2.y) - midQuadHeight) / midQuadHeight +
                                      Math.Abs((p4.y - p3.y) - midQuadHeight) / midQuadHeight;
                                    //+ Math.Abs(QuadArea(p1, p2, p3, p4) - quadArea) / quadArea;
                            }        
                            else
                            {
                                width  = .5 * ((p3.x + p4.x) - (p1.x + p2.x));
                                height = .5 * ((p1.y + p4.y) - (p2.y + p3.y));
                                err = Math.Abs((width / height - quadAspect) / quadAspect);
                            }

                            if(err < errMin)
                            {
                                corn1PT = p1;
                                corn2PT = p2;
                                corn3PT = p3;
                                corn4PT = p4;

                                errMin  = err;
                            }
                        }
                    }
                }
            }

            //jeb use to test extapolation (Ken's 8/12 image)
            /*
            corn1PT = new PT(689 * 4, 1783 * 4);
            corn2PT = new PT(690 * 4, 873 * 4);
            corn3PT = new PT(3283*4, 875*4);
            corn4PT = new PT(3277*4, 1787*4);
            */
        }
        public double QuadArea(PT P1, PT P2, PT P3, PT P4)
        {
            //area of quadralateral

            double area = P1.x * P2.y - P2.x * P1.y;

            area += P2.x * P3.y - P3.x * P2.y;
            area += P3.x * P4.y - P4.x * P3.y;
            area += P4.x * P1.y - P1.x * P4.y;

            return Math.Abs(area / 2);
        }
 


        public void DegreeLength(double Lat, out double Dlon, out double Dlat)
        {
            double a = 6378206.4;
            double b = 6356583.8;
            double m, n, e2, s2;

            if(datum.ToUpper().Contains("WGS84"))
            {
                a = 6378137;
                b = a * (1.0 - 1.0/298.257223563);
            }
            else if(datum.ToUpper().Contains("WGS72"))
            {
                a = 6378135;
                b = a * (1.0 - 1.0/298.26);
            }
            else if (datum.ToUpper().Contains("NAD83"))
            {
                a = 6378137;
                b = a * (1.0 - 1.0/298.257222101); 
            }
            else //use NAD27
            {
                a = 6378206.4;
                b = 6356583.8;
            }

            e2 = (a * a - b * b) / (a * a);
            s2 = Math.Pow(Math.Sin(Math.PI * Lat / 180.0), 2);
            
            m = (a*(1.0-e2))/Math.Pow(1.0-e2*s2,1.5);
            n = a/Math.Sqrt(1.0-e2*s2);

            /*alternate
            m = Math.Pow(a * b, 2) /
                Math.Pow(
                  Math.Pow(a * Math.Cos(Math.PI * Lat / 180.0), 2) +
                  Math.Pow(b * Math.Sin(Math.PI * Lat / 180.0), 2),
                  1.5);

            n = Math.Pow(a , 2) /
               Math.Pow(
                  Math.Pow(a * Math.Cos(Math.PI * Lat / 180.0), 2) +
                  Math.Pow(b * Math.Sin(Math.PI * Lat / 180.0), 2),
               0.5);
            */


            Dlat = m * Math.PI / 180.0;
            Dlon = n * Math.Cos(Math.PI * Lat / 180.0) * Math.PI / 180.0;
         }
 
        public List<EdgeSeg> IdentifyEdges(List<PT> xyv1, List<PT> xyv2, List<PT> xyv3)
        {
            List<EdgeSeg> edgeList = new List<EdgeSeg>();

            for (int i = 0;  i < xyv1.Count; i++)
            {
                for (int j = 0; j < xyv2.Count; j++)
                {
                    for (int k = 0; k < xyv3.Count; k++)
                    {
                        if (Collinear(xyv1[i], xyv2[j], xyv3[k]))
                        {
                            edgeList.Add(new EdgeSeg(xyv1[i], xyv3[k]));
                        }
                    }
                }
            }

            //sort in increasing x- or y- order based on average position
            edgeList.Sort(EdgeOrder);
            
            return edgeList;
        }
        public int EdgeOrder(EdgeSeg E1, EdgeSeg E2)
        {
            if (E1.vertical)
            {
                if      (E1.pMid.x > E2.pMid.x) return  1;
                else if (E1.pMid.x < E2.pMid.x) return -1;
                else                            return  0;
            }
            else
            {
                if      (E1.pMid.y > E2.pMid.y) return  1;
                else if (E1.pMid.y < E2.pMid.y) return -1;
                else                            return  0;
            }

        }
        public PT IntersectEdges(EdgeSeg E1, EdgeSeg E2)
        {
            //find intersection of infinite lines
            //see http://local.wasp.uwa.edu.au/~pbourke/geometry/lineline2d/

            PT p = new PT();
            int x1, x2, x3, x4, y1, y2, y3, y4;
            x1 = E1.p1.x;
            x2 = E1.p2.x;
            y1 = E1.p1.y;
            y2 = E1.p2.y;

            x3 = E2.p1.x;
            x4 = E2.p2.x;
            y3 = E2.p1.y;
            y4 = E2.p2.y;

            long d = (long)(y4 - y3) * (x2 - x1) - (long)(x4 - x3) * (y2 - y1);
            if (d != 0)
            {
                double t = (double)((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / d;

                p.x = (int)Math.Round(x1 + t * (x2 - x1));
                p.y = (int)Math.Round(y1 + t * (y2 - y1));

                /*could also find by extending edge2
                t = (double)((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) / d;

                if(p.x != (int)Math.Round(x3 + t * (x4 - x3)) ||
                   p.y != (int)Math.Round(y3 + t * (y4 - y3)))
                {
                    MessageBox.Show("Error in intersection");
                }
                */
            }
            else
            {
                //lines are parallel or coincident. (Coincident if numerators for t are both
                //zero.) Return x from edge1, y from edge2
                p.x = x2;
                p.y = y3;
            }

            return p;

        }
        public bool Collinear(PT P1, PT P2, PT P3)
        {
            //check if P1,P2,P3 nearly collinear
            
            double angTotal = new EdgeSeg(P1,P3).angle;

            if(Math.Abs(angTotal) < maxImageRotation)
            {
                if(Math.Abs( (new EdgeSeg(P1,P2).angle - new EdgeSeg(P2,P3).angle)) < maxEdgeChange) return true;
                else                                                                     return false;
            }
            else if(Math.Abs(angTotal-90) < maxImageRotation)
            {
                if(Math.Abs((new EdgeSeg(P1,P2).angle - new EdgeSeg(P2,P3).angle)) < maxEdgeChange) return true;
                else                                                                    return false;
            }
            else return false;
        }
 
        private double Search(Bitmap bmp, CMark mark, out int xSW, out int ySW, string StatsFile)
        {

            double err = mark.MatchPattern(bmp,StatsFile);

            xSW = mark.xSW;
            ySW = mark.ySW;

            Console.WriteLine("x,y: {0},{1}", xSW, ySW);

            return err;
        }

        public struct Pixel
        {
            public byte blue;
            public byte green;
            public byte red;
        }

        public static Bitmap ResizeBmp(Bitmap bmp, int width, int height)
        {
            //from a given bitmap bmp, return a new bitmap scaled to width,height

            Bitmap newBmp = new Bitmap(width, height);
            using (Graphics g = Graphics.FromImage(newBmp))
            {
                g.DrawImage(bmp, 0, 0, width, height);
            }
            return newBmp;
        }

        private void SearchAllWindows()
        {
            int    markNo, row, col;
            double epsilon;
            string s;


                markNo = 1;

                s = "Point\tU\tV\tLongitude\t  Latitude\tDisagreement Index" + newLine;

                TBX_Output.AppendText(s);

                for (row = 0; row < totalThumbRows; row++)
                {
                    for (col = 0; col < totalThumbColumns; col++)
                    {

                        if (controlMark[row, col].Type != MarkType.None)
                        {
                            epsilon = SearchForMark(row, col);
                            s = String.Format(
                                    "   {0:d2}\t{1:d5}\t{2:d5}\t{3:f3}\t {4:f3}\t {5:f3}",
                                    markNo,
                                    controlMark[row, col].xImage,
                                    controlMark[row, col].yImage,
                                    controlMark[row, col].lonTrue,
                                    controlMark[row, col].latTrue,
                                    epsilon);
                            TBX_Output.AppendText(s + newLine);
                            TBX_Output.Refresh();

                            thumbBoxArray[row, col].Redraw();
                            markNo++;
                        }
                    }
                }

            /*
            for (row = 0; row < totalThumbRows; row++)
                for (col = 0; col < totalThumbColumns; col++)
                    thumbBoxArray[row, col].Redraw();
            */
        }

        private void BTN_Fit_Click(object sender, EventArgs e)
        {
            PerformLSFit();

            
            //BTN_SearchWindows.Enabled = false;
            

            //if(converterProc.Length > 10 && !noGeoref) BTN_ExportImage.Enabled = true;
            BTN_ExportImage.Enabled = true;

        }

        private void PerformLSFit()
        {
            int row, col;
            string s;
            double rmse, rmseCV, emax, emin, emaxCV, eminCV;
            //fitButtonPressed = true;

            if (totalThumbBoxes < 7 || linearOnly) degreeFit = 1;
            else                                   degreeFit = 2;


            rmseCV = 0;
            emaxCV = -1.0e10;
            eminCV = 1.0e10;
            for (row = 0; row < totalThumbRows; row++)
            {
                for (col = 0; col < totalThumbColumns; col++)
                {
                    if (controlMark[row, col].Type == MarkType.None)
                    {

                        controlMark[row, col].xHatCV = controlMark[row, col].xHat = -1;
                        controlMark[row, col].yHatCV = controlMark[row, col].yHat = -1;
                        
                    }
                    else
                    {
                        LSFit(row, col);

                        controlMark[row, col].xHatCV = controlMark[row, col].xHat;
                        controlMark[row, col].yHatCV = controlMark[row, col].yHat;

                        rmseCV += controlMark[row, col].PixelErrorCV * controlMark[row, col].PixelErrorCV;
                        emaxCV = Math.Max(emaxCV, controlMark[row, col].PixelErrorCV);
                        eminCV = Math.Min(eminCV, controlMark[row, col].PixelErrorCV);
                    }
                }
            }
            rmseCV = Math.Sqrt(rmseCV / (totalThumbBoxes));
            s = String.Format("Cross-validated RMSE = {0:f2} pixels, max,min: {1:f2}, {2:f2}",
                rmseCV, emaxCV, eminCV);
            TBX_Output.AppendText(s + newLine);

            rmse = LSFit(-1, -1);

            emax = -1.0e10;
            emin = 1.0e10;


            for (row = 0; row < totalThumbRows; row++)
            {
                for (col = 0; col < totalThumbColumns; col++)
                {
                    controlMark[row, col].LSFitError = controlMark[row, col].PixelError;

                    if(controlMark [row,col].Type != MarkType .None )
                    {
                      emax = Math.Max(emax, controlMark[row, col].LSFitError);
                      emin = Math.Min(emin, controlMark[row, col].LSFitError);
                    }
                }
            }

            s = String.Format("        Prediction RMSE = {0:f2} pixels,  max,min: {1:f2}, {2:f2}",
                 rmse, emax, emin);

            TBX_Output.AppendText(s + newLine);

            for (row = 0; row < totalThumbRows; row++)
                for (col = 0; col < totalThumbColumns; col++)
                {
                    thumbBoxArray[row, col].CaptionType = ThumbCaptionType.TransformError;
                    thumbBoxArray[row, col].Redraw();
                }

            if (!RBN_Automatic.Checked)
            {
                if (true && CheckBelowThreshold() && converterProc.Length > 10 && !noGeoref || DemoOnly)
                {
                    BTN_ExportImage.Enabled = true;
                }
                else
                {
                    BTN_ExportImage.Enabled = false;
                }
            }

        }

        private bool CheckBelowThreshold()
        {
           for (int row = 0; row < totalThumbRows; row++)
           {
             for (int col = 0; col < totalThumbColumns; col++) 
                 if (controlMark[row, col].LSFitError > threshold3) return false;
           }

           return true;
        }

        public void DisableButtons()
        {
            
            BTN_Fit.Enabled           = false;
            BTN_FindWindows.Enabled   = false;
            BTN_SearchWindows.Enabled = false;
            BTN_ExportImage.Enabled   = false;
        }

        private void AutomaticTimerTick(object source, EventArgs e)
        {

            if (!stopProcessing && LBX_ListOfFiles.SelectedIndex > -1 && currentFileSelected < LBX_ListOfFiles.Items.Count - 1)
            {
                if (searchBmp != null)
                {
                    searchBmp.Dispose();
                }
                if (subBmp != null)
                {
                    subBmp.Dispose();
                }
                imageFile = new ImageFile(this);
                bitmapProcess = new Bitmap(1, 1);
                bitmapForThumbBox = new Bitmap(1, 1);
                automaticTimer.Enabled = false;

                currentFileSelected++;
                LBX_ListOfFiles.ClearSelected();
                LBX_ListOfFiles.SelectedIndex = currentFileSelected;
                LBX_ListOfFiles.Refresh();

                BTN_Process_Click(new Object(), new EventArgs());
            }
            else
            {
                automaticTimer.Enabled = false;
                PBR_ProgressTransform.Value = PBR_ProgressTransform.Minimum;
                runningAutoMode = false;
                RBN_Manual.Checked = true;
                ResetButtons(stopProcessing,true);
            }
        }

        public void WriteOutputFile()
        {
            XmlTextWriter textWriter;
            bool belowErrorThreshold = CheckBelowThreshold();
            
            if (belowErrorThreshold)
            {
                string fileNameToUse = "";
                if (((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).outputGeographicFileName != null &&
                    ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).outputGeographicFileName.Length > 0)
                {
                    fileNameToUse = transformationXMLDirectory + "\\" + ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).outputGeographicFileName + ".xml";
                }
                else
                {
                    fileNameToUse = transformationXMLDirectory + "\\" + 
                        Path.GetFileName(((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName) + ".xml";
                }
                textWriter = new XmlTextWriter(fileNameToUse, null);
            }
            else
            {
                if (RBN_Manual.Checked)
                {
                    MessageBox.Show("RMSE above upper theshold.  Error file will be created.");
                }
                AppendErrorFile();
                return;
                //textWriter = new XmlTextWriter(LBL_ErrorDirectory.Text + "\\" + ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName.ToString() + ".error.xml", null);
            }

            textWriter.WriteStartDocument();

            textWriter.Formatting = Formatting.Indented;

            textWriter.WriteStartElement("Scan");

            textWriter.WriteStartElement("FileName");
            textWriter.WriteString(((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName);
            textWriter.WriteEndElement();

            if(((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).mapName != null)
            {
              textWriter.WriteStartElement("MapName");
              textWriter.WriteString(((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).mapName);
              textWriter.WriteEndElement();
            }

            textWriter.WriteStartElement("ProcessDate");
            textWriter.WriteString(DateTime.Now.ToString());
            textWriter.WriteEndElement();

            textWriter.WriteStartElement("Comments");
            textWriter.WriteString(TBX_Comments.Text);
            textWriter.WriteEndElement();

            textWriter.WriteStartElement("User");
            textWriter.WriteString(WindowsIdentity.GetCurrent().Name);
            textWriter.WriteEndElement();

            int markNumber = 1;

            for (int row = 0; row < totalThumbRows; row++)
            {
                for (int col = 0; col < totalThumbColumns; col++)
                {
                    if (controlMark[row, col].Type != MarkType.None)
                    {
                        textWriter.WriteStartElement("MarkU" + markNumber.ToString());
                        textWriter.WriteString(controlMark[row, col].xImage.ToString());
                        textWriter.WriteEndElement();

                        textWriter.WriteStartElement("MarkV" + markNumber.ToString());
                        textWriter.WriteString(controlMark[row, col].yImage.ToString());
                        textWriter.WriteEndElement();

                        textWriter.WriteStartElement("MarkLongitude" + markNumber.ToString());
                        textWriter.WriteString(controlMark[row, col].lonTrue.ToString());
                        textWriter.WriteEndElement();

                        textWriter.WriteStartElement("MarkLatitude" + markNumber.ToString());
                        textWriter.WriteString(controlMark[row, col].latTrue.ToString());
                        textWriter.WriteEndElement();

                        textWriter.WriteStartElement("LSFitError" + markNumber.ToString());
                        textWriter.WriteString(controlMark[row, col].LSFitError.ToString() + "pixels");
                        textWriter.WriteEndElement();


                        textWriter.WriteStartElement("MarkGeoTiffError" + markNumber.ToString());
                        textWriter.WriteString(controlMark[row, col].GeoTiffError.ToString() + "meters");
                        textWriter.WriteEndElement();

                        markNumber++;
                    }
                }
            }

            textWriter.WriteEndElement();

            textWriter.WriteEndDocument();
            textWriter.Close();

        }

        private void AppendErrorFile()
        {
            string errorFileName = errorDirectory + "\\Errors_" + errorFileNameSuffix + ".xml";

            if (!File.Exists(errorFileName))
            {
                XmlTextWriter textWriter = new XmlTextWriter(errorFileName, null);
                textWriter.WriteStartDocument();
                textWriter.Formatting = Formatting.Indented;
                textWriter.WriteStartElement("ScanList");
                textWriter.WriteEndElement();
                textWriter.WriteEndDocument();
                textWriter.Close();
            }

            XmlDocument doc = new XmlDocument();
            doc.Load(errorFileName);
            XmlNode node = doc.CreateNode(XmlNodeType.Element, "Scan", null);

            XmlScanNode xmlScanNode = (XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex];

            node.InnerXml += "<FileName>" + xmlScanNode.fileName + "</FileName>\n";
            
            if (xmlScanNode.mapName != null) 
                node.InnerXml += "<MapName>" + xmlScanNode.mapName + "</MapName>\n";
            if(xmlScanNode.datum != null)
                node.InnerXml += "<Datum>" + xmlScanNode.datum + "</Datum>\n";

            node.InnerXml += "<Longitude>" + xmlScanNode.longitude + "</Longitude>\n";
            node.InnerXml += "<Latitude>" + xmlScanNode.latitude + "</Latitude>\n";
            node.InnerXml += "<Resolution>" + xmlScanNode.resolution + "</Resolution>\n";

            if (xmlScanNode.mapScale != null) 
                node.InnerXml += "<Scale>" + xmlScanNode.mapScale + "</Scale>\n";
            
            node.InnerXml += "<ControlMarkLayout>" + xmlScanNode.controlMarkLayout + "</ControlMarkLayout>\n";
            node.InnerXml += "<ControlMarkLonSpacing>" + xmlScanNode.controMarkLonSpacing + "</ControlMarkLonSpacing>\n";
            node.InnerXml += "<ControlMarkLatSpacing>" + xmlScanNode.controMarkLatSpacing + "</ControlMarkLatSpacing>\n";
            node.InnerXml += "<ControlMarkSize>" + xmlScanNode.controlMarkSize + "</ControlMarkSize>\n";
            node.InnerXml += "<PerimeterMarksOnly>" + xmlScanNode.perimeterMarksOnly + "</PerimeterMarksOnly>\n";
            if (xmlScanNode.outputGeographicFileName != null)
              node.InnerXml += "<OutputGeographicFileName>" + xmlScanNode.outputGeographicFileName + "</OutputGeographicFileName>\n";

            doc.DocumentElement.AppendChild(node);
            doc.Save(errorFileName);
        }


        void FindWindow(int row, int col)
        {
            int x, y;//, x0, y0;
            double fx, fy;

            fx = (double)col / (totalThumbColumns - 1);
            fy = (double)row / (totalThumbRows - 1);

            //guess at mark location, this will become search window center
            GuessMarkLocation(fx, fy, out x, out y);

            //MessageBox.Show("ready to extract sub-area...");

            //extract window centered on x,y
            //x0 = x - searchActualSize / 2;
            //y0 = y - searchActualSize / 2;

            controlMark[row, col].xImage = x;
            controlMark[row, col].yImage = y;
            thumbBoxArray[row, col].CaptionType = ThumbCaptionType.Nothing;
            thumbBoxArray[row, col].bitmapLocalFile = CenterThumbImage(row, col);
            thumbBoxArray[row, col].Redraw();
            
        }

        double SearchForMark(int row, int col)
        {
            //search for the mark at row,col in controlMark array

            int x, y, x0, y0, xfound, yfound;
            double misMatch;

            x = controlMark[row, col].xImage;
            y = controlMark[row, col].yImage;

            //extract window centered on x,y
            x0 = x - searchActualSize / 2;
            y0 = y - searchActualSize / 2;


            x0 = Math.Max(0, Math.Min(imageWidth  - searchActualSize - 1, x0));
            y0 = Math.Max(0, Math.Min(imageHeight - searchActualSize - 1, y0));

            searchBmp = imageFile.ExtractRectangle(
                 new Rectangle(x0, y0, searchActualSize, searchActualSize));

            if (searchBmp == null) return 0.0;

            controlMark[row, col].ResetError();
            misMatch = Search(searchBmp, controlMark[row, col], out xfound, out yfound, "");
            
            //update control mark location based on search results 
            controlMark[row, col].xImage = x0 + xfound;
            controlMark[row, col].yImage = y0 + yfound;

            thumbBoxArray[row, col].bitmapLocalFile = CenterThumbImage(row, col);

            thumbBoxArray[row, col].CaptionType  = ThumbCaptionType.SearchError;
            thumbBoxArray[row, col].Redraw();

            return misMatch;


        }


        private Bitmap CenterThumbImage(int xRow, int xCol)
        {
            //center thumbox bitmap around contol mark

            int matchX, matchY, leftEdge, topEdge;
            double f = (double)searchDisplaySize / (double)searchActualSize;
            Rectangle extractedRegion;

            matchX = controlMark[xRow, xCol].xImage;
            matchY = controlMark[xRow, xCol].yImage; 
            leftEdge = matchX - (sizeThumbWidth / 2);
            topEdge = matchY - (sizeThumbHeight / 2);
            extractedRegion = new Rectangle(leftEdge, topEdge, sizeThumbWidth, sizeThumbHeight);
            thumbBoxArray[xRow, xCol].thisBoundingBoxStart = extractedRegion;
            thumbBoxArray[xRow, xCol].thisBoundingBoxCurrent = extractedRegion;

            return imageFile.ExtractRectangle(extractedRegion);

        }

        public Bitmap NudgeThumbImage(Rectangle xBoundingBox)
        {
            return imageFile.ExtractRectangle(xBoundingBox);
        }

        private void GuessMarkLocation(double fx, double fy, out int x, out int y)
        {
            //use bi-linear interpolation from corner estimates to guess
            //location of marks.

            double x1, x2, y1, y2;

            x1 = (1.0 - fx) * corn1PT.x + fx * corn4PT.x;
            y1 = (1.0 - fx) * corn1PT.y + fx * corn4PT.y;

            x2 = (1.0 - fx) * corn2PT.x + fx * corn3PT.x;
            y2 = (1.0 - fx) * corn2PT.y + fx * corn3PT.y;

            x = (int)Math.Round((1.0 - fy) * x1 + fy * x2);
            y = (int)Math.Round((1.0 - fy) * y1 + fy * y2);

        }

        public string DD2DDM(double dd)
        {
            int degrees, minutes, seconds;
            double f;
            string s;

            degrees = (int)Math.Floor(dd);

            f = 60 * (dd - degrees);
            minutes = (int)Math.Floor(f);

            f = 60 * (f - minutes);
            seconds = (int)Math.Round(f);

            s = String.Format("{0:d2} {1:d2} {2:d2}",
                degrees, minutes, seconds);

            return s;
        }

        public double LSFit(int xRow, int xCol)
        {
            ///least squares fit using all control marks except controlMark[xRow, xCol]
            
            ///if xRow and/or xCol is out of range then fit
            /// using all marks (i.e., use all totalThumbRows*totalThumbColumns of them)

            ///return pixel RMSE for all marks including controlMark[xRow,xCol]

            ///compute and store predicted lon,lat and u,v (.lonHat, .latHat, etc.) 
            ///for all marks

            int nobs;   //number of observations---most likely 15 or 16 control marks being used
            int nparms; //number of parameters (always 3 or 6 for us)


            int nSing;  //number of singular values (should be 3 unless very ill-conditioned)
            int nDigs;  //approximate number of significant digits in parameter estimates

            double rMSE;   //prediction error


            //will find X to minimize |AX-B|
            Mapack.Matrix AlonLat2uv;     //design matrix
            Mapack.Matrix Buv;            //right-hand sides (both u and v)

            Mapack.SingularValueDecomposition svd; //use svd to obtain LS approximation

            if (degreeFit != 2) nparms = 3; //linear, e.g., u = a + b*lon + c*lat or v = d + e*lon + f*lat
            else                nparms = 6; //quadratic u = a + b*lon + c*lat + d*lon^2 + e*lat^2 +f*lat^2 +f*lon*lat


            

            //number of observations
            if (xRow < 0 || xRow >= totalThumbRows ||
                xCol < 0 || xCol >= totalThumbColumns) 
                
                 nobs = totalThumbBoxes;

            //not using mark[row,col]
            else nobs = totalThumbBoxes - 1;


            AlonLat2uv = new Mapack.Matrix(nobs, nparms);
            Buv = new Mapack.Matrix(nobs, 2); //u will go in first column, v in second

            //scale factors to improve conditioning of A
            double u0 = 0;
            double v0 = 0;
            double uScale = 1.0 / imageHeight;
            double vScale = uScale;

            double lon0 = controlMark[0, 0].lonTrue;
            double lat0 = controlMark[0, 0].latTrue;
            double lonScale = 1.0 / Math.Abs(controlMark[0, 0].lonTrue - controlMark[0, totalThumbColumns - 1].lonTrue);
            double latScale = 1.0 / Math.Abs(controlMark[0, 0].latTrue - controlMark[totalThumbRows - 1, 0].latTrue);

            scaleLonLat.Init();
            scaleLonLat.AddTranslation(-lon0, -lat0);
            scaleLonLat.AddScale(lonScale, latScale);

            scaleuv.Init();
            scaleuv.AddTranslation(-u0, -v0);
            scaleuv.AddScale(uScale, vScale);

            int n = 0;
            for (int row = 0; row < totalThumbRows; row++)
            {
                for (int col = 0; col < totalThumbColumns; col++)
                {
                    if ( (row != xRow || col != xCol) && controlMark [row,col].Type != MarkType.None )
                    {
                        double lonscaled, latscaled, uscaled, vscaled;

                        scaleLonLat.Forward(controlMark[row, col].lonTrue,
                            controlMark[row, col].latTrue,
                            out lonscaled, out latscaled);


                        //build left hand side
                        AlonLat2uv[n, 0] = 1.0;
                        AlonLat2uv[n, 1] = lonscaled;
                        AlonLat2uv[n, 2] = latscaled;

                        if (nparms > 3)
                        {
                            AlonLat2uv[n, 3] = lonscaled * lonscaled;
                            AlonLat2uv[n, 4] = latscaled * latscaled;
                            AlonLat2uv[n, 5] = lonscaled * latscaled;
                        }


                        scaleuv.Forward(controlMark[row, col].xImage,
                            controlMark[row, col].yImage,
                            out uscaled, out vscaled);

                        //build right hand side
                        Buv[n, 0] = uscaled;
                        Buv[n, 1] = vscaled;


                        n++;
                    }
                }

            }

            if (n != nobs) throw new ArgumentException("Error building Least Squares matrices");

            //decompose and back-solve
            svd = new Mapack.SingularValueDecomposition(AlonLat2uv);
            polyFit = svd.Solve(Buv, out nSing, out nDigs);

            //use transform to get predicted lon,lat and its inverse for predicted u,v
            rMSE = 0;
            for (int row = 0; row < totalThumbRows; row++)
            {
                for (int col = 0; col < totalThumbColumns; col++)
                {

                    if (controlMark[row, col].Type != MarkType.None)
                    {
                        //u,v for actual control mark lon,lat
                        lonLat2uv(controlMark[row, col].lonTrue, controlMark[row, col].latTrue,
                           out controlMark[row, col].xHat,
                           out controlMark[row, col].yHat);

                        double uverror = controlMark[row, col].PixelError; //from predicted u,v

                        controlMark[row, col].LSFitError = uverror;

                        rMSE += uverror * uverror;
                    }
                    else controlMark[row, col].LSFitError = 0;
                   

                }
            }

            rMSE = Math.Sqrt(rMSE / (totalThumbBoxes));



            return rMSE;
        }

        public void lonLat2uv(double lon, double lat, out double uhat, out double vhat)
        {
            double lonscale, latscale, uscale, vscale;

            scaleLonLat.Forward(lon, lat, out lonscale, out latscale);

            uscale = polyFit[0, 0] + polyFit[1, 0] * lonscale + polyFit[2, 0] * latscale;
            vscale = polyFit[0, 1] + polyFit[1, 1] * lonscale + polyFit[2, 1] * latscale;

            if (polyFit.Rows > 3)
            {
                uscale += polyFit[3, 0] * lonscale * lonscale + polyFit[4, 0] * latscale * latscale + polyFit[5, 0] * lonscale * latscale;
                vscale += polyFit[3, 1] * lonscale * lonscale + polyFit[4, 1] * latscale * latscale + polyFit[5, 1] * lonscale * latscale;
            }
            scaleuv.Inverse(uscale, vscale, out uhat, out vhat);



        }

        public class Affine
        {
            private Mapack.Matrix A, Ainv;

            public Affine()
            {
                A = new Mapack.Matrix(3, 3);
                Ainv = new Mapack.Matrix(3, 3);

                Init();
            }
            public void AddTranslation(double ptrans, double qtrans)
            {
                Mapack.Matrix T = new Mapack.Matrix(3, 3);

                T[0, 0] = 1.0; T[0, 1] = 0.0; T[0, 2] = 0.0;
                T[1, 0] = 0.0; T[1, 1] = 1.0; T[1, 2] = 0.0;
                T[2, 0] = ptrans; T[2, 1] = qtrans; T[2, 2] = 1.0;

                A = A * T;
                Ainv = A.Inverse;
            }
            public void AddScale(double pscale, double qscale)
            {
                Mapack.Matrix S = new Mapack.Matrix(3, 3);

                S[0, 0] = pscale; S[0, 1] = 0.0; S[0, 2] = 0.0;
                S[1, 0] = 0.0; S[1, 1] = qscale; S[1, 2] = 0.0;
                S[2, 0] = 0.0; S[2, 1] = 0.0; S[2, 2] = 1.0;

                A = A * S;
                Ainv = A.Inverse;
            }
            public void AddM(Mapack.Matrix M)
            {
                if (M.Rows != 3 || M.Columns < 2)
                    throw new ArgumentException("Need 3x2 matrix for AddM");

                Mapack.Matrix M2 = new Mapack.Matrix(3, 3);

                M2[0, 0] = M[0, 0]; M2[0, 1] = M[0, 1]; M2[0, 2] = 0.0;
                M2[1, 0] = M[1, 0]; M2[1, 1] = M[1, 1]; M2[1, 2] = 0.0;
                M2[2, 0] = M[2, 0]; M2[2, 1] = M[2, 1]; M2[2, 2] = 1.0;

                A = A * M2;
                Ainv = A.Inverse;
            }
            public void Init()
            {
                A[0, 0] = 1.0; A[0, 1] = 0.0; A[0, 2] = 0.0;
                A[1, 0] = 0.0; A[1, 1] = 1.0; A[1, 2] = 0.0;
                A[2, 0] = 0.0; A[2, 1] = 0.0; A[2, 2] = 1.0;

                Ainv = A;
            }

            public void Forward(double p, double q, out double r, out double s)
            {
                ///transform point (p,q) to (r,s)

                r = p * A[0, 0] + q * A[1, 0] + A[2, 0];
                s = p * A[0, 1] + q * A[1, 1] + A[2, 1];


            }
            public void Inverse(double p, double q, out double r, out double s)
            {
                ///transform point (p,q) to (r,s)

                r = p * Ainv[0, 0] + q * Ainv[1, 0] + Ainv[2, 0];
                s = p * Ainv[0, 1] + q * Ainv[1, 1] + Ainv[2, 1];


            }
        }

        public void TransformImage()
        {
            int    nlats, nlons, um1, vm1, up1, vp1;
            double lat, lon, lonStart, lonEnd, latStart, latEnd;
            double dlon, dlat, lonpixels, latpixels, londif, latdif;
            double u, v;

            string s, cmd, args, ppmFileName, geoTifFileName;

            Pixel[,] swath = new Pixel[1, 1];

            byte[] pixArray;
            Pixel  pLL, pUL, pUR, pLR;
            double f, rLeft, gLeft, bLeft, rRight, gRight, bRight;

            int rectHeight = 1000; //will pull rectangles from the image file of this height

            if (false && DemoOnly)
            {

                TBX_Output.AppendText("Image transform disabled"+newLine);
                TBX_Output.Refresh();

                Application.DoEvents();
                return;
            }

            lonStart = controlMark[0, 0].lonTrue;
            lonEnd   = controlMark[0, totalThumbColumns - 1].lonTrue;

            latStart = controlMark[totalThumbRows - 1, 0].latTrue;
            latEnd   = controlMark[0, 0].latTrue;


            londif   = Math.Abs(lonEnd - lonStart);
            latdif   = Math.Abs(latEnd - latStart);

            lonpixels = Math.Sqrt(
                Math.Pow(controlMark[0, totalThumbColumns - 1].xImage - controlMark[0, 0].xImage, 2) +
                Math.Pow(controlMark[0, totalThumbColumns - 1].yImage - controlMark[0, 0].yImage, 2));

            latpixels = Math.Sqrt(
                Math.Pow(controlMark[totalThumbRows - 1, 0].xImage - controlMark[0, 0].xImage, 2) +
                Math.Pow(controlMark[totalThumbRows - 1, 0].yImage - controlMark[0, 0].yImage, 2));

            //dlon and dlat are degrees per pixel, will be the same here but set up for
            //cases where might be different

            //use whichever resolution is higher for dlon and dlat
            if (lonpixels / londif > latpixels / latdif)
            {
                dlon = londif / lonpixels;
                dlat = dlon;
            }
            else
            {
                dlat = latdif / latpixels;
                dlon = dlat;
            }

            if (RBN_IncludeCollar.Checked || RBN_Both.Checked)
            {
                //user wants collar, have to get extent of map sheet for that
                double lon0 = (lonStart + lonEnd) / 2.0;
                double lat0 = (latStart + latEnd) / 2.0;

                double lastv,lastu;
                int offset;

                lastu = imageWidth;
                offset = 0;
                do
                {
                    lonStart = controlMark[0, 0].lonTrue - offset * dlon;
                    lonLat2uv(lonStart, lat0, out u, out v);
                    offset++;
                    if (u >= lastu)
                    {
                        u = lastu;
                        ReportConcavity("left longitude", (int)Math.Truncate(u), latEnd);
                        break;
                    }
                    else lastu = u;

                } while (u > 10 && offset < imageWidth);

                if (offset >= imageWidth) ReportExtrapolation("left longitude", (int)Math.Truncate(u));



                lastu = offset = 0; 
                do
                {
                    lonEnd = controlMark[0, totalThumbColumns - 1].lonTrue + offset * dlon;
                    lonLat2uv(lonEnd, lat0, out u, out v);
                    offset++;
                    if (u < lastu)
                    {
                        u = lastu;
                        ReportConcavity("right longitude", (int)Math.Truncate(u), latEnd);
                        break;
                    }
                    else lastu = u;
                } while (u < imageWidth - 11 && offset < imageWidth);

                if (offset >= imageWidth) ReportExtrapolation("right longitude", (int)Math.Truncate(u));

                lastv = imageHeight;
                offset = 0;
                do
                {
                    latStart = controlMark[totalThumbRows - 1, 0].latTrue + offset * dlat;
                    lonLat2uv(lon0, latStart, out u, out v);
                    offset++;
                    
                    if (v >= lastv)
                    {
                        v = lastv;
                        ReportConcavity("top latitude", (int)Math.Truncate(v), latEnd);
                        break;
                    }
                    else lastv = v;

                } while (v > 10 && offset < imageHeight);
                if (offset >= imageHeight) ReportExtrapolation("top latitude", (int)Math.Truncate(v));

                lastv = offset = 0;
                do
                {
                    latEnd = controlMark[0, 0].latTrue - offset * dlat;
                    lonLat2uv(lon0, latEnd, out u, out v);

                    offset++;
                    if (v <= lastv)
                    {
                        v = lastv;
                        ReportConcavity("bottom latitude",(int)Math.Truncate(v),latEnd);
                        break;
                    }
                    else lastv = v;


                    } while (v < imageHeight - 11 && offset < imageHeight);
                
            }

            londif = Math.Abs(lonEnd - lonStart);
            latdif = Math.Abs(latEnd - latStart);

            nlons = 1 + (int)Math.Truncate(0.5 + londif / dlon);
            nlats = 1 + (int)Math.Truncate(0.5 + latdif / dlat);


            lonEnd = lonStart + (nlons - 1) * dlon;

            dlat = -dlon;
            latEnd = latStart + (nlats - 1) * dlat;

            if (DemoOnly)
            {
                int len = ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName.LastIndexOf(".");
                if (len < 1) len = ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName.Length;

                geoTifFileName = transformedImageDirectory + "\\" +
                    ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName.Substring(0, len)
                    + "Geo.tif";

                for (int row = 1000; row < 10001; row += 1000)
                {
                    s = String.Format("row {0:d5} done", row);
                    TBX_Output.AppendText(s + newLine);
                    TBX_Output.Refresh();

                    Application.DoEvents();
                    Thread.Sleep(200);
                }
                if(!noQA) DoQAcheck(lonStart, latStart, lonEnd, latEnd, geoTifFileName);
                return;
            }

            ppmFileName = Path.GetTempFileName();

            FileStream   file = File.OpenWrite(ppmFileName);
            BinaryWriter bw   = new BinaryWriter(file);

            char[] header = String.Format("P6\n{0} {1}\n255\n",
                nlons, nlats).ToCharArray();

            bw.Write(header);

            pixArray = new byte[nlons * 3];

            swath[0, 0].red = swath[0, 0].green = swath[0, 0].blue = 0;

            int fileRow = 1000000;  //ensures swath is initialized in loop
            int nSwath = 0;         //swath number
          
            for (int klat = 0; klat < nlats; klat++)
            {
                lat = latStart + klat * dlat;

                for (int klon = 0; klon < nlons; klon++)
                {
                    lon = lonStart + klon * dlon;

                    lonLat2uv(lon, lat, out u, out v);

                    u = Math.Max(Math.Min(u, imageWidth - 10.0), 1.0);
                    v = Math.Max(Math.Min(v, imageHeight - 10.0), 1.0);

                    um1 = (int)Math.Truncate(u);
                    vm1 = (int)Math.Truncate(v);

                    up1 = um1 + 1;
                    vp1 = vm1 + 1;

                    if (vm1 < fileRow || vp1 > fileRow + rectHeight - 1)
                    {
                        Application.DoEvents();
                        nSwath++;

                        fileRow = Math.Max(0, vm1 - 100);

                        int ySize = Math.Min(rectHeight,
                                             imageHeight - fileRow - 1);

                        if (fileRow + ySize > imageHeight - 1 || ySize < 1)
                        {
                            MessageBox.Show("wrong rectangle start or height");
                        }

                        s = string.Format("swath {0:d4}, file row start: {1:d5}, end:    {2:d5}",
                            nSwath, fileRow, fileRow + ySize - 1);

                        TBX_Output.AppendText(s + newLine);
                        TBX_Output.Refresh();

                        //TBX_Output.AppendText(string.Format("stripe: {0:d5} v0:{1} height:{2}    \n", nstripes, v0,ySize));

                        swath = imageFile.ExtractSwath(fileRow, ySize);

                        if (stopProcessing)
                        {
                            file.Close();
                            File.Delete(ppmFileName);
                            if (!runningAutoMode)
                            {
                                ResetButtons(true,true);
                            }
                            return;
                        }
                    }

                    //4 pixels surrounding u,v (lower left, upper left, ...)
                    pLL = swath[vp1 - fileRow, um1];
                    pUL = swath[vm1 - fileRow, um1];
                    pUR = swath[vm1 - fileRow, up1];
                    pLR = swath[vp1 - fileRow, up1];

                    //linear interp in v-direction along left edge
                    f = v - vm1;
                    rLeft = (1.0 - f) * pUL.red + f * pLL.red;
                    gLeft = (1.0 - f) * pUL.green + f * pLL.green;
                    bLeft = (1.0 - f) * pUL.blue + f * pLL.blue;

                    //and right edge
                    rRight = (1.0 - f) * pUR.red + f * pLR.red;
                    gRight = (1.0 - f) * pUR.green + f * pLR.green;
                    bRight = (1.0 - f) * pUR.blue + f * pLR.blue;


                    //linear interp in u-direction
                    f = u - um1;
                    pixArray[klon * 3] = (byte)Math.Round((1.0 - f) * rLeft + f * rRight);
                    pixArray[klon * 3 + 1] = (byte)Math.Round((1.0 - f) * gLeft + f * gRight);
                    pixArray[klon * 3 + 2] = (byte)Math.Round((1.0 - f) * bLeft + f * bRight);

                }

                bw.Write(pixArray);
                if (klat % 100 == 0)
                {
                    s = String.Format("row {0:d5} done", klat);
                    TBX_Output.AppendText(s + newLine);
                    TBX_Output.Refresh();

                    Application.DoEvents();
                }

            }



            TBX_Output.AppendText(String.Format("all done, {0:d5} rows (inc. margins), {1} swaths",
                nlats,nSwath) + newLine);

            TBX_Output.Refresh();
            file.Close();

            if (((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).outputGeographicFileName != null
                &&
                ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).outputGeographicFileName != "")
            {
                geoTifFileName = transformedImageDirectory + "\\" + 
                    ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).outputGeographicFileName;
            }
            else
            {
                int len = ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName.LastIndexOf(".");
                if (len < 1) len = ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName.Length;

                geoTifFileName = transformedImageDirectory + "\\" + 
                    ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName.Substring(0,len) 
                    + "Geo.tif";
                
            }


            cmd = converterProc;

            if (((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).datum != null)
            {
                datum = ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).datum.ToString();
            }
            else datum = defaultDatum;

            args = "-of gtiff -a_srs \"" + datum + "\" -a_ullr ";

            s = String.Format("{0:f15} {1:f15} {2:f15} {3:f15} ",
                  lonStart - dlon / 2,
                  latStart - dlat / 2,
                  lonEnd + dlon / 2,
                  latEnd + dlat / 2);
            args += s;

            if (!File.Exists(ppmFileName))
            {
                    MessageBox.Show("Can't find temporary file, looking for:"+ newLine + ppmFileName);
            }

            args += " \"" + ppmFileName + "\"  \"" + geoTifFileName + "\"";

            TBX_Output.AppendText(newLine + String.Format("Converting temporary image file: {0}", ppmFileName)
                   + newLine);

            TBX_Output.AppendText(String.Format("         to geotiff output file: {0}", geoTifFileName)
                   + newLine + newLine);

            TBX_Output.AppendText(String.Format("Command:{0}", cmd) + newLine + newLine);
            TBX_Output.AppendText(String.Format("Arguments:{0}", args) + newLine);

            TBX_Output.Refresh();

            System.Diagnostics.ProcessStartInfo startInfo = new ProcessStartInfo(cmd, args);

            //startInfo.RedirectStandardOutput = true;
            //startInfo.RedirectStandardError = true;
            //startInfo.UseShellExecute = false;
            //startInfo.CreateNoWindow = true;
            using (Process proc = Process.Start(startInfo))
            {
                
                while (!proc.HasExited)
                {
                    Thread.Sleep(2000);
                }
                Cursor.Current = Cursors.Arrow;
                if (proc.ExitCode != 0)
                {
                    s = "Non-zero error code from GDAL. Command issued" + newLine;
                    s += cmd + " " + args;
                    //s += newLine;
                    //s += proc.StandardError.ReadToEnd();
                    MessageBox.Show(s, "GDAL Error");
                    //TBX_Output.AppendText(newLine + s + newLine);
                }
            }


            if (RBN_Both.Checked && !s.Contains("GDAL Error")) 
            {
               string geoTif2;
               int len = geoTifFileName.LastIndexOf(".");
               if (len > 0) geoTif2 = geoTifFileName.Substring(0, len) + "NoCollar" + geoTifFileName.Substring(len);
               else geoTif2 = geoTifFileName + "NoCollar";
            
               args = "-of gtiff -projwin ";

               s = String.Format("{0:f15} {1:f15} {2:f15} {3:f15} ",
                  controlMark[0, 0].lonTrue - dlon / 2,
                  controlMark[totalThumbRows - 1, 0].latTrue - dlat / 2,
                  controlMark[0, totalThumbColumns - 1].lonTrue + dlon / 2,
                  controlMark[0, 0].latTrue + dlat / 2);

               args += s;
               
               args += " \"" + geoTifFileName + "\"  \"" + geoTif2 + "\"";

               TBX_Output.AppendText(String.Format("Creating no-collar version: {0}", geoTif2)
                   + newLine + newLine);

               TBX_Output.AppendText(String.Format("Command:{0}", cmd) + newLine + newLine);
               TBX_Output.AppendText(String.Format("Arguments:{0}", args) + newLine);

               TBX_Output.Refresh();

               using (Process proc = Process.Start(cmd, args))
               {
                  while (!proc.HasExited)
                  {
                    Thread.Sleep(2000);
                  }
                  Cursor.Current = Cursors.Arrow;
                  if (proc.ExitCode != 0)
                  {
                    s = "Non-zero error code from GDAL. Command issued:"+ newLine;
                    s += cmd + " " + args;
                    MessageBox.Show(s, "GDAL Error");
                  }
               }

            }

            File.Delete(ppmFileName);
            if (!s.Contains("error"))
            {
                s = ((XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex]).fileName;
                 if (noQA)
                {
                    TBX_Output.AppendText(newLine + "All done with "+s+".");
                }
                else 
                {
                    TBX_Output.AppendText(newLine + "Starting QA. ");
                    DoQAcheck(lonStart, latStart, lonEnd, latEnd, geoTifFileName);
                    TBX_Output.AppendText(newLine + "All done with "+s + ".");
                }
            }
        }
        private void ReportConcavity(string msg, int pixLimit, double geogLimit)
        {
            string s1 = String.Format("Polynomial not reliable for extrapolating to {0}. ",msg);

            if (msg.ToUpper().Contains("LAT") )
            {
                s1 += "(Latitude not \nmonotonic in v.)";
                s1 += String.Format(" Can extend only to pixel {0}, latitude {1}.",pixLimit,geogLimit);
            }
            else    
            {
                s1 += "(Longitude not \nmonotonic in u.)";
                s1 += String.Format(" Can extend only to pixel {0}, longitude {1}.", pixLimit, geogLimit);
            }
                           
            s1 += "\nCollar pixels will likely be distorted. Consider linear fit or excluding collar.";
            
            if(RBN_Manual.Checked) MessageBox.Show(s1, "Extrapolation Error");

            TBX_Output.AppendText(s1);


        }
        private void ReportExtrapolation(string Msg,int pixLimit)
        {
            string s1 = String.Format(
                "Problem looking for {0} \n Will use pixel {1} \nAre you extrapolating using a 2nd degree fit?",
                Msg,pixLimit);

            if (RBN_Manual.Checked) MessageBox.Show(s1, "Extrapolation Error");

            TBX_Output.AppendText(s1);
        }
        private void DoQAcheck(double LonStart, double LatStart, double LonEnd, double LatEnd, 
            string FileName)
        {
            int    n,rowStart,rowEnd,colStart,colEnd;
   
            double dLon, dLat;
            double rmsError, minError, maxError,meanError;
            double degLat, degLon;

            string s, fullVerifyFileName;

            

            imageFile.CloseFiles();
            imageFile = new ImageFile(this);

            if (!File.Exists(FileName) || !imageFile.LoadFile(FileName,converterProc)) return;

            TBX_Output.AppendText("Verifying "+ Path.GetFileName(FileName) + newLine);
            TBX_Output.Refresh();
            Application.DoEvents();


            imageWidth  = imageFile.Width;
            imageHeight = imageFile.Height;

            dLon = (LonEnd - LonStart) / (imageWidth-1);
            dLat = (LatEnd - LatStart) / (imageHeight-1);

            bitmapWidthPixels  = imageWidth;
            bitmapHeightPixels = imageHeight;

            bitmapWidthBytes  = imageWidth * 3;
            bitmapHeightBytes = imageHeight * 3;


            if (mapScale > 1) metersPerPixel = mapScale/ (imageDPI * 39.37);
            else
            {
                DegreeLength( (controlMark[0,0].latTrue+controlMark[totalThumbRows-1,0].latTrue)/2.0,
                    out degLon, out degLat);

                metersPerPixel = degLat*(controlMark[totalThumbRows-1,0].latTrue - controlMark[0,0].latTrue)/
                                        (controlMark[0,0].yImage                 - controlMark[totalThumbRows-1,0].yImage);
            }
            
            CreateThumbBoxes();

            if (RBN_LoseCollar.Checked)
            {
                rowStart = colStart = 1;
                rowEnd   = totalThumbRows - 2; 
                colEnd   = totalThumbColumns - 2;
            }
            else
            {
                rowStart = colStart = 0;
                rowEnd   = totalThumbRows - 1;
                colEnd   = totalThumbColumns - 1;
            }
            
            fullVerifyFileName = Path.GetDirectoryName(FileName) + "\\"+ verifyCSVFilename;
            
            //use this to write error measures for all scans to CSV file (note commented block below)
            using (StreamWriter sw = File.AppendText(fullVerifyFileName))
            {
              n        = 0;
              rmsError = meanError = 0;
              minError = 1.0e100;
              maxError = -minError;

              for (int row = rowStart; row <= rowEnd; row++)
              {
                 for (int col = colStart; col <= colEnd; col++)
                 {

                    
                    double xd = (imageFile.Width-1)  * (controlMark[row, col].lonTrue - LonStart) / (LonEnd - LonStart);
                    double yd = (imageFile.Height-1) * (controlMark[row, col].latTrue - LatStart) / (LatEnd - LatStart);

                    int x = (int)Math.Round(xd); 
                    int y = (int)Math.Round(yd); 

                    Rectangle r = new Rectangle(x - (searchActualSize)/2, 
                                                y - (searchActualSize)/2, searchActualSize,searchActualSize);

                    Bitmap bmp = imageFile.ExtractRectangle(r);
                    
                     //string bmpName = FileName+"Verify"+Convert.ToString(row)+"-"+Convert.ToString(col)+".png";
                    //bmp.Save(bmpName,ImageFormat.Png);

                    int xSave = 0;
                    int ySave = 0;

                    if (controlMark[row, col].Type != MarkType.None)
                    {
                        double err = controlMark[row, col].MatchPattern(bmp, "");

                        double xFound = x + controlMark[row, col].xSW - searchActualSize / 2;
                        double yFound = y + controlMark[row, col].ySW - searchActualSize / 2;

                        double lonPixelError = (xFound - xd);
                        double latPixelError = (yFound - yd);

                        DegreeLength(controlMark[row, col].latTrue, out degLon, out degLat);

                        double lonError = dLon * lonPixelError * degLon;
                        double latError = dLat * latPixelError * degLat;

                        double pixError = Math.Sqrt(Math.Pow(xFound - xd, 2) +
                                                  Math.Pow(yFound - yd, 2));

                        double earthError = Math.Sqrt(Math.Pow(lonError, 2) +
                                                    Math.Pow(latError, 2));

                        rmsError += earthError * earthError;
                        meanError += earthError;
                        minError = Math.Min(earthError, minError);
                        maxError = Math.Max(earthError, maxError);

                        n++;

                        xSave = controlMark[row, col].xImage;
                        ySave = controlMark[row, col].yImage;


                        //needed to centerThumbImage
                        controlMark[row, col].xImage = x;
                        controlMark[row, col].yImage = y;


                        controlMark[row, col].xGeoRef = xFound;
                        controlMark[row, col].yGeoRef = yFound;

                        if (err < 0.5) controlMark[row, col].GeoTiffError = earthError;
                        else controlMark[row, col].GeoTiffError = 9999.99;

                    }
                    thumbBoxArray[row, col].bitmapLocalFile = CenterThumbImage(row, col);

                    thumbBoxArray[row, col].CaptionType = ThumbCaptionType.GeoTiffError;
                    thumbBoxArray[row, col].Redraw();

                    Console.WriteLine("x,y: {0},{1}", controlMark[row,col].xSW, controlMark[row,col].ySW);

                    if (controlMark[row, col].Type != MarkType.None)
                    {
                        controlMark[row, col].xImage = xSave;
                        controlMark[row, col].yImage = ySave;
                    }

                 }
              }
              rmsError = Math.Sqrt(rmsError / n);
              meanError /= n;

              
              //append error measures to csvdata
              if (new FileInfo(fullVerifyFileName).Length < 10) //new file, need header for csv file
              {
                  s = "RMS Error (m), Max Error (m), Min Error (m), Mean Error (m), n, Georeferenced File";
                  sw.WriteLine(s);
              }
              s = string.Format("{0:f3}, {1:f3}, {2:f3}, {3:f3}, {4:d3}, ",
                                rmsError, maxError, minError, meanError, n);
              s += Path.GetFileName(FileName);
            
              sw.WriteLine(s);
              

              SaveAsPng(PNL_ThumbArea, Path.GetDirectoryName(FileName)+ "\\"+ 
                                       Path.GetFileNameWithoutExtension(FileName) + "Verify.png");


            }
        }
        private void BTN_ExportImage_Click(object sender, EventArgs e)
        {
            BTN_Process.Enabled = false;
            BTN_FindWindows.Enabled = false;
            BTN_Fit.Enabled = false;
            BTN_ExportImage.Enabled = false;
            BTN_Stop.Enabled = true;
            LBX_ListOfFiles.Enabled = false;
            GBX_Mode.Enabled = false;
            GBX_Extent.Enabled = false;
            DisableButtons();



            if (CheckBelowThreshold() && converterProc.Length > 10 && !noGeoref || DemoOnly)
            {
                TransformImage();
            }
            WriteOutputFile();


            if (noQA) ResetButtons(false, true);
            else      ResetButtons(false, false);

        }

        private void BTN_ExportXML_Click(object sender, EventArgs e)
        {
            BTN_FindWindows.Enabled = false;
            DisableButtons();
            WriteOutputFile();
        }

        private void RBN_Automatic_CheckedChanged(object sender, EventArgs e)
        {
            if (RBN_Automatic.Checked)
            {
                BTN_Process.Text = "Run Batch";
            }
            else
            {
                BTN_Process.Text = "Find Corners";
            }
            DisableButtons();
        }

        private void RBN_Grayscale_CheckedChanged(object sender, EventArgs e)
        {
            if (RBN_Grayscale.Checked)
            {
                thumbBoxesAreColor = false;
            }
            else
            {
                thumbBoxesAreColor = true;
            }
            RedrawThumbBoxes();
        }

        private void RBN_IncludeCollar_CheckedChanged(object sender, EventArgs e)
        {
        }

        private void RBN_LoseCollar_CheckedChanged(object sender, EventArgs e)
        {
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void BTN_FindWindows_Click(object sender, EventArgs e)
        {

            for (int row = 0; row < totalThumbRows; row++)
            {
                for (int col = 0; col < totalThumbColumns; col++)
                {
                    thumbBoxArray[row, col].CaptionType = ThumbCaptionType.Nothing;
                    thumbBoxArray[row, col].EmptyThumbBoxes(bitmapGray);
                    controlMark[row, col].ResetError();
                }
            }

            UpdateCorners();
            
            for (int row = 0; row < totalThumbRows; row++)
            {
               for (int col = 0; col < totalThumbColumns; col++)
               {
                        FindWindow(row, col);
               }
            }

            BTN_FindWindows.Enabled         = false;
            BTN_SearchWindows.Enabled = true;
            

        }

        public bool UpdateCorners()
        {
            //use thumbox array because corners might have been adjusted by user
            //this will update corner points accorndingly 

            Rectangle tbc;

            tbc = thumbBoxArray[0, 0].thisBoundingBoxCurrent;
            corn1PT.x = tbc.X + (sizeThumbWidth / 2);
            corn1PT.y = tbc.Y + (sizeThumbHeight / 2);

            tbc = thumbBoxArray[totalThumbRows - 1, 0].thisBoundingBoxCurrent;
            corn2PT.x = tbc.X + (sizeThumbWidth / 2);
            corn2PT.y = tbc.Y + (sizeThumbHeight / 2);

            tbc = thumbBoxArray[totalThumbRows - 1, totalThumbColumns - 1].thisBoundingBoxCurrent;
            corn3PT.x = tbc.X + (sizeThumbWidth / 2);
            corn3PT.y = tbc.Y + (sizeThumbHeight / 2);

            tbc = thumbBoxArray[0, totalThumbColumns - 1].thisBoundingBoxCurrent;
            corn4PT.x = tbc.X + (sizeThumbWidth / 2);
            corn4PT.y = tbc.Y + (sizeThumbHeight / 2);

            return true;
        }

        private void BTN_Stop_Click(object sender, EventArgs e)
        {
            stopProcessing = true;
            runningAutoMode = false;
        }

        private void ResetButtons(bool StopClicked,bool ClearBoxes)
        {
            BTN_Process.Enabled = true;
            BTN_FindWindows.Enabled = false;
            BTN_Fit.Enabled = false;
            BTN_ExportImage.Enabled = false;
            BTN_Stop.Enabled = false;
            LBX_ListOfFiles.Enabled = true;
            GBX_Mode.Enabled = true;
            GBX_Extent.Enabled = true;
            stopProcessing = false;

            if (StopClicked || ClearBoxes)
            {
                for (int row = 0; row < totalThumbRows; row++)
                {
                    for (int col = 0; col < totalThumbColumns; col++)
                    {
                        thumbBoxArray[row, col].EmptyThumbBoxes(bitmapGray);
                    }
                }
            }

            if (StopClicked) TBX_Output.AppendText(newLine+ "Operation stopped by user." + newLine);
            PBR_ProgressTransform.Value = PBR_ProgressTransform.Minimum;
        }

        public void CreateThumbBoxes()
        {
            PNL_ThumbArea.Controls.Clear();
            thumbBoxArray = new ThumbBox[totalThumbRows, totalThumbColumns];
            int thumbBoxNumber = 1;


            for (int row = 0; row < totalThumbRows; row++)
            {
                for (int col = 0; col < totalThumbColumns; col++)
                {


                    thumbBoxArray[row, col] = new ThumbBox(this, CenterThumbBoxesVertical(col), CenterThumbBoxesHorizontal(row), sizeThumbWidth, sizeThumbHeight, "ThumbBox_" + row.ToString() + "_" + col.ToString(), bitmapGray, thumbBoxNumber, row, col);

                    this.PNL_ThumbArea.Controls.Add(this.thumbBoxArray[row, col]);
                    thumbBoxArray[row, col].EmptyThumbBoxes(bitmapGray);
                    thumbBoxArray[row, col].Visible = true;

                    thumbBoxNumber++;
                }
            }
        }
        public void SetThumbsVisible(bool Vis)
        {

            for (int row = 0; row < totalThumbRows; row++)
            {
                for (int col = 0; col < totalThumbColumns; col++)
                {
                    if (Vis) thumbBoxArray[row, col].Visible = true;
                    else thumbBoxArray[row, col].Visible = false;

                }
            }
        }

        private void PNL_ThumbArea_Paint(object sender, PaintEventArgs e)
        {

        }

        private void ModeButtonClick(object sender, EventArgs e)
        {
            if (RBN_Manual.Checked) BTN_Process.Text = "Find Corners";
            else                    BTN_Process.Text = "Run Batch";
            
        }

        private void BTN_SearchWindows_Click(object sender, EventArgs e)
        {
            SearchAllWindows();
            BTN_FindWindows.Enabled = false;
            BTN_Fit.Enabled         = true;
            
        }

        private void manualInputToolStripMenuItem_Click(object sender, EventArgs e)
        {

            if (LBX_ListOfFiles.SelectedIndex >= 0)
                xmlScanNode = (XmlScanNode)scanListArray[LBX_ListOfFiles.SelectedIndex];

            else xmlScanNode = new XmlScanNode();

            manualInputDialog = new ManualInput(xmlScanNode, sourceImageDirectory);
            
            
            if (manualInputDialog.ShowDialog(this) == DialogResult.OK)
            {
                xmlScanNode = manualInputDialog.node;

                if (LBX_ListOfFiles.SelectedIndex >= 0)
                {
                    scanListArray[LBX_ListOfFiles.SelectedIndex] = xmlScanNode;

                    LBX_ListOfFiles_SelectedIndexChanged(this, e);
                }
                else
                {
                    DisableButtons();
                    LBX_ListOfFiles.Items.Clear();
                    
                    scanListArray.Clear();
                    scanListArray.Add(xmlScanNode);

                    totalFilesInList = 1;

                    fileList = new String[totalFilesInList, 3];

                    FillFileListBox();
                }
            }
        }

        private void CB_CVerrors_CheckedChanged(object sender, EventArgs e)
        {
            if (CB_CVerrors.Checked) wantCVErrors = true;
            else                     wantCVErrors = false;

            RedrawThumbBoxes();

        }
       
    }
}
