﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace QUADG
{
    public partial class ManualInput : Form
    {
        public bool canceled;
        public XmlScanNode node;
        public String sourceImageDirectory;

        public ManualInput(XmlScanNode Node, String imageDirectory)
        {
            node = Node;
            sourceImageDirectory = imageDirectory;
            InitializeComponent();

            if (node.fileName != null)
            {
                TBX_ScanFile.Text = node.fileName;
                TBX_Size.Text = node.controlMarkSize;
                TBX_LatSpacing.Text = node.controMarkLatSpacing;
                TBX_LonSpacing.Text = node.controMarkLonSpacing;

                CB_PerimeterMarksOnly.Checked = node.perimeterMarksOnly.ToUpper().Contains("T");

                string[] findLayout = node.controlMarkLayout.ToLower().Split('x');
                TBX_Rows.Text = findLayout[0];
                TBX_Columns.Text = findLayout[1];


                TBX_Datum.Text = node.datum;
                TBX_Latitude.Text = node.latitude;
                TBX_Longitude.Text = node.longitude;
                TBX_QuadName.Text = node.mapName;
                TBX_MapScale.Text = node.mapScale;
                TBX_Resolution.Text = node.resolution;
                TBX_OutputFile.Text = node.outputGeographicFileName;
            }
        }

        private void BTN_Cancel_Click(object sender, EventArgs e)
        {
            this.DialogResult = DialogResult.Cancel;
        }

        private void BTN_OK_Click(object sender, EventArgs e)
        {
            if (File.Exists(sourceImageDirectory + TBX_ScanFile.Text) || File.Exists(sourceImageDirectory + "\\" + TBX_ScanFile.Text))
            {
                node.fileName = TBX_ScanFile.Text;
            }
            else
            {
                MessageBox.Show("Scan file does not exist.");
                return;
            }
            if (CheckControlMarkSize(TBX_Size.Text))
            {
                node.controlMarkSize = TBX_Size.Text;
            }
            else
            {
                MessageBox.Show("Invalid control mark size.");
                return;
            }
            if (CheckDegreeString(TBX_LatSpacing.Text))
            {
                node.controMarkLatSpacing = TBX_LatSpacing.Text;
            }
            else
            {
                MessageBox.Show("Invalid lat control mark spacing.");
                return;
            }
            if (CheckDegreeString(TBX_LonSpacing.Text))
            {
                node.controMarkLonSpacing = TBX_LonSpacing.Text;
            }
            else
            {
                MessageBox.Show("Invalid lon control mark spacing.");
                return;
            }

            if (CheckForInt(TBX_Rows.Text) && CheckForInt(TBX_Columns.Text))
            {
                node.controlMarkLayout = TBX_Rows.Text + "x" + TBX_Columns.Text;
            }
            else
            {
                MessageBox.Show("Invalid row and/or column number.");
                return;
            }
            if (TBX_Datum.Text != "")
            {
                node.datum = TBX_Datum.Text;
            }
            else
            {
                MessageBox.Show("Datum is missing.");
                return;
            }
            if (CheckDegreeString(TBX_Latitude.Text))
            {
                node.latitude = TBX_Latitude.Text;
            }
            else
            {
                MessageBox.Show("Invalid latitude.");
                return;
            }
            if (CheckDegreeString(TBX_Longitude.Text))
            {
                node.longitude = TBX_Longitude.Text;
            }
            else
            {
                MessageBox.Show("Invalid longitude.");
                return;
            }
            if (CheckForInt(TBX_Resolution.Text))
            {
                node.resolution = TBX_Resolution.Text;
            }
            else
            {
                MessageBox.Show("Invalid resolution.");
                return;
            }
            if (TBX_MapScale.TextLength == 0)
                node.mapScale = "0";

            else if (CheckForInt(TBX_MapScale.Text))
            {
                node.mapScale = TBX_MapScale.Text;
            }
            else
            {
                MessageBox.Show("Invalid map scale.");
                return;
            }

            if (CB_PerimeterMarksOnly.Checked) node.perimeterMarksOnly = "True";
            else                          node.perimeterMarksOnly = "False";


            node.mapName = TBX_QuadName.Text;
            node.outputGeographicFileName = TBX_OutputFile.Text;
                        
            this.DialogResult = DialogResult.OK;
            
        }

        private void TBX_Latitude_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (e.KeyChar.ToString() != "." && e.KeyChar.ToString() != "-" && e.KeyChar.ToString() != " ")
                {
                    if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                        e.Handled = true;
                }
                
            }
        }

        private void TBX_Longitude_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (e.KeyChar.ToString() != "." && e.KeyChar.ToString() != "-" && e.KeyChar.ToString() != " ")
                {
                    if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                        e.Handled = true;
                }
               
            }
        }

        private void TBX_Rows_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                    e.Handled = true;
            }
        }

        private void TBX_Columns_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                    e.Handled = true;
            }
        }

        private void TBX_Spacing_KeyPress(object sender, KeyPressEventArgs e)
        {
          if (!Char.IsControl(e.KeyChar))
            {
                if (e.KeyChar.ToString() != "." && e.KeyChar.ToString() != " ")
                {
                    if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                        e.Handled = true;
                }
                
            }
        }

        private void TBX_Size_KeyPress(object sender, KeyPressEventArgs e)
        { 
            //check on dialog exit
        }

        private void TBX_Resolution_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                    e.Handled = true;
            }
        }

        private void TBX_MapScale_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                    e.Handled = true;
            }
        }
        private bool CheckControlMarkSize(string SizeString)
        {
            if (SizeString.Length < 1) return false;

            int len = SizeString.ToLower().IndexOf("i");

            if (len < 1) len = SizeString.ToLower().IndexOf("p");
            if (len < 1) len = SizeString.ToLower().IndexOf("f");
            if (len < 1) len = SizeString.Length;

            try
            {
                Convert.ToDouble(SizeString.Substring(0, len));
                return true;
            }
            catch {return false;}

        }
        public bool CheckDegreeString(string Dstring)
        {
            if (Dstring.Length < 1) return false;
            try
            {

                double value = 0, one = 1;

                string[] fields = Dstring.Split(new char[] { ' ' });

                if (Dstring.Contains('-')) one = -1;


                if (fields.Length > 0) value = one * Math.Abs(Convert.ToDouble(fields[0]));
                if (fields.Length > 1) value += one * Math.Abs(Convert.ToDouble(fields[1])) / 60.0;
                if (fields.Length > 2) value += one * Math.Abs(Convert.ToDouble(fields[2])) / 3600.0;

                return true;
            }
            catch { return false; }

        }
        public bool CheckForInt(string S)
        {
            if (S.Length < 1) return false;
            try
            {
                Convert.ToInt32(S);
                return true;
            }
            catch { return false; }
        }



    }
}
