﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.IO;

namespace QUADG
{
    public partial class FRM_Preferences : Form
    {
        private FRM_MainForm thisParent;
        public int    sizeThumbWidth;
        public int    sizeThumbHeight;
        public int    sizeSearchWindow;
        public Color  threshold1Color;
        public Color  threshold2Color;
        public Color  threshold3Color;
        public Color  threshold4Color;
        public Color  crosshairColor;
        public Color  predictionColor;
        public double threshold1;
        public double threshold2;
        public double threshold3;
        public String sourceImageDirectory = "";
        public String transformationXMLDirectory = "";
        public String transformedImageDirectory = "";
        public String errorDirectory = "";

        public bool noRotation, noGeoref, noQA, linearOnly;

        public FRM_Preferences(FRM_MainForm xThisParent)
        {
            thisParent = xThisParent;
            InitializeComponent();
            UpdateValues();
        }

        private void UpdateValues()
        {
            sizeThumbWidth = thisParent.sizeThumbWidth;
            sizeThumbHeight = thisParent.sizeThumbHeight;
            sizeSearchWindow = thisParent.searchActualSize;

            threshold1 = thisParent.threshold1;
            threshold2 = thisParent.threshold2;
            threshold3 = thisParent.threshold3;
            threshold1Color = thisParent.threshold1Color;
            threshold2Color = thisParent.threshold2Color;
            threshold3Color = thisParent.threshold3Color;
            threshold4Color = thisParent.threshold4Color;
            crosshairColor = thisParent.crosshairColor;
            predictionColor = thisParent.predictionColor;

            transformationXMLDirectory = thisParent.transformationXMLDirectory;
            transformedImageDirectory = thisParent.transformedImageDirectory;
            errorDirectory = thisParent.errorDirectory;
            sourceImageDirectory = thisParent.sourceImageDirectory;

            TBX_SizeThumbWidth.Text = Convert.ToString(sizeThumbWidth);
            TBX_SizeThumbHeight.Text = Convert.ToString(sizeThumbHeight);
            TBX_SizeSearchWindow.Text = Convert.ToString(sizeSearchWindow);

            BTN_Threshold1.BackColor = threshold1Color;
            BTN_Threshold2.BackColor = threshold2Color;
            BTN_Threshold3.BackColor = threshold3Color;
            BTN_Threshold4.BackColor = threshold4Color;
            BTN_CrosshairColor.BackColor = crosshairColor;
            BTN_PredictionColor.BackColor = predictionColor;

            LBL_Level1.Text = "0 - " + threshold1;
            LBL_Level2.Text = threshold1 + " - " + threshold2;
            LBL_Level3.Text = threshold2 + " - " + threshold3;
            LBL_Level4.Text = "> " + threshold3;

            TBX_SourceImagesDirectory.Text = sourceImageDirectory;
            TBX_TransformationDirectory.Text = transformationXMLDirectory;
            TBX_TransformedImageDirectory.Text = transformedImageDirectory;
            TBX_ErrorDirectory.Text = errorDirectory;

            TBX_Threshold1.Text = Convert.ToString(threshold1);
            TBX_Threshold2.Text = Convert.ToString(threshold2);
            TBX_Threshold3.Text = Convert.ToString(threshold3);

            noGeoref   = thisParent.noGeoref;
            noQA       = thisParent.noQA;
            noRotation = thisParent.noRotation;
            linearOnly = thisParent.linearOnly;


            CB_LinearOnly.Checked = linearOnly;
            CB_NoRotation.Checked = noRotation;
            CB_NoQA.Checked       = noQA;
            CB_NoGeoref.Checked   = noGeoref;

            if (noGeoref) CB_NoQA.Enabled = false;
            else          CB_NoQA.Enabled = true;

            //this.Text = thisParent.preferencesXMLFile;
        }

        private void BTN_Threshold1_Click(object sender, EventArgs e)
        {
            CDG_SelectColor.Color = BTN_Threshold1.BackColor;
            DialogResult dialogResult = CDG_SelectColor.ShowDialog();

            if (dialogResult == DialogResult.OK)
            {
                BTN_Threshold1.BackColor = CDG_SelectColor.Color;
            }
            threshold1Color = BTN_Threshold1.BackColor;
        }

        private void BTN_CrosshairColor_Click(object sender, EventArgs e)
        {
            CDG_SelectColor.Color = BTN_CrosshairColor.BackColor;
            DialogResult dialogResult = CDG_SelectColor.ShowDialog();

            if (dialogResult == DialogResult.OK)
            {
                BTN_CrosshairColor.BackColor = CDG_SelectColor.Color;
            }
            crosshairColor = BTN_CrosshairColor.BackColor;
        }

        
        private void BTN_Threshold2_Click(object sender, EventArgs e)
        {
            CDG_SelectColor.Color = BTN_Threshold2.BackColor;
            DialogResult dialogResult = CDG_SelectColor.ShowDialog();

            if (dialogResult == DialogResult.OK)
            {
                BTN_Threshold2.BackColor = CDG_SelectColor.Color;
            }
            threshold2Color = BTN_Threshold2.BackColor;
        }

        private void BTN_Threshold3_Click(object sender, EventArgs e)
        {
            CDG_SelectColor.Color = BTN_Threshold3.BackColor;
            DialogResult dialogResult = CDG_SelectColor.ShowDialog();

            if (dialogResult == DialogResult.OK)
            {
                BTN_Threshold3.BackColor = CDG_SelectColor.Color;
            }
            threshold3Color = BTN_Threshold3.BackColor;
        }

        private void BTN_Threshold4_Click(object sender, EventArgs e)
        {
            CDG_SelectColor.Color = BTN_Threshold4.BackColor;
            DialogResult dialogResult = CDG_SelectColor.ShowDialog();

            if (dialogResult == DialogResult.OK)
            {
                BTN_Threshold4.BackColor = CDG_SelectColor.Color;
            }
            threshold4Color = BTN_Threshold4.BackColor;
        }

        private void BTN_BrowseSourceImagesDirectory_Click(object sender, EventArgs e)
        {
 
            FBD_SelectFolder.Description = "Please specify input image directory";

            FBD_SelectFolder.SelectedPath = TBX_SourceImagesDirectory.Text;
            if (FBD_SelectFolder.SelectedPath == "") FBD_SelectFolder.SelectedPath = Directory.GetCurrentDirectory();


            if (FBD_SelectFolder.ShowDialog() == DialogResult.OK)
            {
                TBX_SourceImagesDirectory.Text = FBD_SelectFolder.SelectedPath;
                sourceImageDirectory = TBX_SourceImagesDirectory.Text;
            }
            
        }
        private void BTN_Apply_Click(object sender, EventArgs e)
        {
            if (CheckDirectories() && CheckTextFields()) 
            {
                BTN_Save.Enabled = true;
                BTN_Close.Enabled = true;
                MakeUpdates();
            }
            else                                      
            {
                BTN_Save.Enabled = false;
                BTN_Close.Enabled = false;
            
            }

        }

        private void BTN_OK_Click(object sender, EventArgs e)
        {
            if (CheckDirectories() && CheckTextFields())
            {
                BTN_Save.Enabled = true;
                BTN_Close.Enabled = true;
                MakeUpdates();
            }
            else
            {
                BTN_Save.Enabled = false;
                BTN_Close.Enabled = false;

            }
            
        }

        private bool CheckTextFields()
        {
            if (TBX_SizeThumbWidth.Text != "" &&
                TBX_SizeThumbHeight.Text != "" &&
                TBX_SizeSearchWindow.Text != "" &&
                TBX_Threshold1.Text != "" &&
                TBX_Threshold2.Text != "" &&
                TBX_Threshold3.Text != "" &&
                TBX_SourceImagesDirectory.Text != "" &&
                TBX_TransformationDirectory.Text != "" &&
                TBX_TransformedImageDirectory.Text != "" &&
                TBX_ErrorDirectory.Text != "")
            {
                return true;
            }
            else
            {
                MessageBox.Show("Preferences contain empty fields.");
                return false;
            }
        }
        private bool CheckDirectories()
        {
            string s = "";

            if (!Directory.Exists(TBX_SourceImagesDirectory.Text))    
                s += "Source directory doesn't exist" + Environment.NewLine;

            if (!Directory.Exists(TBX_TransformationDirectory.Text))  
                s += "Transformation directory doesn't exist" + Environment.NewLine;
         
            if(!Directory.Exists(TBX_TransformedImageDirectory.Text)) 
                s += "Transformed image directory doesn't exist" + Environment.NewLine;

            if(!Directory.Exists(TBX_ErrorDirectory.Text))            
                s += "Error directory doesn't exist" + Environment.NewLine;
         
            if (s.Length > 0)
            {
                MessageBox.Show(s);
                return false;
            }
            else return true;


        }
        public void WritePreferences()
        {
            try
            {
                XmlTextWriter preferencesWriter = new XmlTextWriter(@"" + thisParent.preferencesXMLFile, null);
                preferencesWriter.Formatting = Formatting.Indented;
                preferencesWriter.WriteStartDocument();
                preferencesWriter.WriteStartElement("PreferencesList");
                preferencesWriter.WriteStartElement("Preferences");

                preferencesWriter.WriteStartElement("ThumbWidth", "");
                preferencesWriter.WriteString(Convert.ToString(sizeThumbWidth));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("ThumbHeight", "");
                preferencesWriter.WriteString(Convert.ToString(sizeThumbHeight));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("SearchWindowSize", "");
                string s = Convert.ToString(sizeSearchWindow);
                preferencesWriter.WriteString(Convert.ToString(sizeSearchWindow));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("CrosshairColor", "");
                preferencesWriter.WriteString(Convert.ToString(crosshairColor.R + "," + crosshairColor.G + "," + crosshairColor.B));
                preferencesWriter.WriteEndElement();

                preferencesWriter.WriteStartElement("PredictionColor", "");
                preferencesWriter.WriteString(Convert.ToString(predictionColor.R + "," + predictionColor.G + "," + predictionColor.B));
                preferencesWriter.WriteEndElement();
                
                
                preferencesWriter.WriteStartElement("ThresholdLevel1", "");
                preferencesWriter.WriteString(Convert.ToString(threshold1));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("ThresholdLevel2", "");
                preferencesWriter.WriteString(Convert.ToString(threshold2));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("ThresholdLevel3", "");
                preferencesWriter.WriteString(Convert.ToString(threshold3));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("ThresholdColor1", "");
                preferencesWriter.WriteString(Convert.ToString(threshold1Color.R + "," + threshold1Color.G + "," + threshold1Color.B));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("ThresholdColor2", "");
                preferencesWriter.WriteString(Convert.ToString(threshold2Color.R + "," + threshold2Color.G + "," + threshold2Color.B));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("ThresholdColor3", "");
                preferencesWriter.WriteString(Convert.ToString(threshold3Color.R + "," + threshold3Color.G + "," + threshold3Color.B));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("ThresholdColor4", "");
                preferencesWriter.WriteString(Convert.ToString(threshold4Color.R + "," + threshold4Color.G + "," + threshold4Color.B));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("XMLDirectory", "");
                preferencesWriter.WriteString(Convert.ToString(transformationXMLDirectory));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("ImageDirectory", "");


                preferencesWriter.WriteString(Convert.ToString(transformedImageDirectory));

                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("ErrorDirectory", "");
                preferencesWriter.WriteString(Convert.ToString(errorDirectory));
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteStartElement("SourceImageDirectory", "");
                preferencesWriter.WriteString(Convert.ToString(sourceImageDirectory));
                preferencesWriter.WriteEndElement();


                preferencesWriter.WriteStartElement("NoRotation", "");
                preferencesWriter.WriteString(Convert.ToString(noRotation));
                preferencesWriter.WriteEndElement();

                preferencesWriter.WriteStartElement("LinearOnly", "");
                preferencesWriter.WriteString(Convert.ToString(linearOnly));
                preferencesWriter.WriteEndElement();

                preferencesWriter.WriteStartElement("NoGeoref", "");
                preferencesWriter.WriteString(Convert.ToString(noGeoref));
                preferencesWriter.WriteEndElement();

                preferencesWriter.WriteStartElement("NoQA", "");
                preferencesWriter.WriteString(Convert.ToString(noQA));
                preferencesWriter.WriteEndElement();

                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteEndElement();
                preferencesWriter.WriteEndDocument();
                preferencesWriter.Close();
                MessageBox.Show("Preferences saved to:\n" + thisParent.preferencesXMLFile);
            }
            catch
            {
                MessageBox.Show("Could not write preferences file.");
            }
        }

        public void MakeUpdates()
        {
            sizeThumbWidth = Convert.ToInt32(TBX_SizeThumbWidth.Text);
            sizeThumbHeight = Convert.ToInt32(TBX_SizeThumbHeight.Text);
            sizeSearchWindow = Convert.ToInt32(TBX_SizeSearchWindow.Text);
            threshold1 = Convert.ToDouble(TBX_Threshold1.Text);
            threshold2 = Convert.ToDouble(TBX_Threshold2.Text);
            threshold3 = Convert.ToDouble(TBX_Threshold3.Text);
            sourceImageDirectory = TBX_SourceImagesDirectory.Text;
            transformationXMLDirectory = TBX_TransformationDirectory.Text;
            transformedImageDirectory = TBX_TransformedImageDirectory.Text;
            errorDirectory = TBX_ErrorDirectory.Text;
            thisParent.UpdatePreferences();
            thisParent.CreateThumbBoxes();
        }

        private void BTN_BrowseErrorDirectory_Click(object sender, EventArgs e)
        {

            FBD_SelectFolder.Description = "Please specify output directory for error reports";
            FBD_SelectFolder.SelectedPath = TBX_ErrorDirectory.Text;
            if (FBD_SelectFolder.SelectedPath == "") FBD_SelectFolder.SelectedPath = Directory.GetCurrentDirectory();

            if (FBD_SelectFolder.ShowDialog() == DialogResult.OK)
            {
                TBX_ErrorDirectory.Text = FBD_SelectFolder.SelectedPath;
                errorDirectory = TBX_ErrorDirectory.Text;
            }
        }

        private void BTN_BrowseTransformedImageDirectory_Click(object sender, EventArgs e)
        {

            FBD_SelectFolder.Description = "Please specify output directory for georeferenced images";
            FBD_SelectFolder.SelectedPath = TBX_TransformedImageDirectory.Text;

            if (FBD_SelectFolder.SelectedPath == "") FBD_SelectFolder.SelectedPath = Directory.GetCurrentDirectory();

            if (FBD_SelectFolder.ShowDialog() == DialogResult.OK)
            {
                TBX_TransformedImageDirectory.Text = FBD_SelectFolder.SelectedPath;
                transformedImageDirectory = TBX_TransformedImageDirectory.Text;
            }
        }

        private void BTN_BrowseTransformationDirectory_Click(object sender, EventArgs e)
        {

            FBD_SelectFolder.Description = "Please specify output directory for transform statistics";
            FBD_SelectFolder.SelectedPath = TBX_TransformationDirectory.Text;

            if (FBD_SelectFolder.SelectedPath == "") FBD_SelectFolder.SelectedPath = Directory.GetCurrentDirectory();

            if (FBD_SelectFolder.ShowDialog() == DialogResult.OK)
            {
                TBX_TransformationDirectory.Text = FBD_SelectFolder.SelectedPath;
                transformationXMLDirectory = TBX_TransformationDirectory.Text;
            }
        }

        private void TBX_Threshold1_TextChanged(object sender, EventArgs e)
        {
            LBL_Level1.Text = "0 - " + TBX_Threshold1.Text;
            LBL_Level2.Text = TBX_Threshold1.Text + " - " + TBX_Threshold2.Text;
        }

        private void TBX_Threshold2_TextChanged(object sender, EventArgs e)
        {
            LBL_Level2.Text = TBX_Threshold1.Text + " - " + TBX_Threshold2.Text;
            LBL_Level3.Text = TBX_Threshold2.Text + " - " + TBX_Threshold3.Text;
        }

        private void TBX_Threshold3_TextChanged(object sender, EventArgs e)
        {
            LBL_Level3.Text = TBX_Threshold2.Text + " - " + TBX_Threshold3.Text;
            LBL_Level4.Text = "> " + TBX_Threshold3.Text;
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MakeUpdates();
            WritePreferences();
        }

        private void openFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog();
            openFileDialog.Filter = "XML Files|*.xml";
            openFileDialog.Title = "Select an XML File";
            openFileDialog.FileName = thisParent.preferencesXMLFile;

            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                XmlDocument xmlDoc = new XmlDocument();
                xmlDoc.Load(openFileDialog.FileName);

                if (xmlDoc.GetElementsByTagName("ThumbWidth")[0] == null)
                {
                    MessageBox.Show("Invalid preferences XML file.");
                }
                else
                {
                    thisParent.preferencesXMLFile = openFileDialog.FileName;
                    this.Text = "Preferences File: " + Path.GetFileName(thisParent.preferencesXMLFile);
                    //this.Text = thisParent.preferencesXMLFile;
                    thisParent.ReadPreferences();
                    UpdateValues();
                }
            }
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveFileDialog = new SaveFileDialog();
            saveFileDialog.DefaultExt = "xml";
            saveFileDialog.Filter = "XML Files|*.xml";
            saveFileDialog.FileName = thisParent.preferencesXMLFile;
            saveFileDialog.AddExtension = true;
            saveFileDialog.RestoreDirectory = true;
            saveFileDialog.Title = "Save preferences XML file";
            saveFileDialog.InitialDirectory = @"";
            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                thisParent.preferencesXMLFile = saveFileDialog.FileName;
                this.Text = "Preferences File: " + Path.GetFileName(thisParent.preferencesXMLFile);
                //this.Text = thisParent.preferencesXMLFile;
                WritePreferences();
            }

            saveFileDialog.Dispose();
            saveFileDialog = null;
        }

        private void TBX_SizeThumbWidth_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                    e.Handled = true;
            }
        }

        private void TBX_SizeThumbHeight_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                    e.Handled = true;
            }
        }

        private void TBX_SizeSearchWindow_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                    e.Handled = true;
            }
        }
        private void TBX_Threshold1_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (e.KeyChar.ToString() != ".")
                {
                    if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                        e.Handled = true;
                }
                else
                {
                    string[] checkDecimals = TBX_Threshold1.Text.Split('.');
                    if (checkDecimals.Length > 1)
                    {
                        e.Handled = true;
                    }
                }
            }
        }

        private void TBX_Threshold2_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (e.KeyChar.ToString() != ".")
                {
                    if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                        e.Handled = true;
                }
                else
                {
                    string[] checkDecimals = TBX_Threshold2.Text.Split('.');
                    if (checkDecimals.Length > 1)
                    {
                        e.Handled = true;
                    }
                }
            }
        }

        private void TBX_Threshold3_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (!Char.IsControl(e.KeyChar))
            {
                if (e.KeyChar.ToString() != ".")
                {
                    if (!System.Text.RegularExpressions.Regex.IsMatch(e.KeyChar.ToString(), "\\d+"))
                        e.Handled = true;
                }
                else
                {
                    string[] checkDecimals = TBX_Threshold3.Text.Split('.');
                    if (checkDecimals.Length > 1)
                    {
                        e.Handled = true;
                    }
                }
            }
        }

        private void BTN_Save_Click(object sender, EventArgs e)
        {
            if(!CheckDirectories())
            {
                MessageBox.Show("One or more directories does not exist. Can't save changes.");
            }
            else if (CheckTextFields())
            {
                MakeUpdates();
                WritePreferences();
            }
        }

        private void BTN_PredictionColor_Click(object sender, EventArgs e)
        {
            CDG_SelectColor.Color = BTN_PredictionColor.BackColor;
            DialogResult dialogResult = CDG_SelectColor.ShowDialog();

            if (dialogResult == DialogResult.OK)
            {
                BTN_PredictionColor.BackColor = CDG_SelectColor.Color;
            }
            predictionColor = BTN_PredictionColor.BackColor;
        }

        private void CB_NoRotation_CheckedChanged(object sender, EventArgs e)
        {
            noRotation = CB_NoRotation.Checked;
        }

        private void CB_LinearOnly_CheckedChanged(object sender, EventArgs e)
        {
            linearOnly = CB_LinearOnly.Checked;
        }
        private void CB_NoGeoref_CheckedChanged(object sender, EventArgs e)
        {
            noGeoref = CB_NoGeoref.Checked;

            if (noGeoref)
            {
                CB_NoQA.Enabled = false;
                CB_NoQA.Checked = true;
            }
            else
            {
                CB_NoQA.Enabled = true;
            }
        }

        private void CB_NoQA_CheckedChanged(object sender, EventArgs e)
        {
            if (noGeoref) CB_NoQA.Checked = true;

            noQA = CB_NoQA.Checked;
        }



    }
}
