﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;

namespace QUADG
{

    public unsafe class ThumbBox : System.Windows.Forms.PictureBox
    {
        private FRM_MainForm thisParent;
        private string thisName;
        public int thisXPos;
        public int thisYPos;
        public int thisWidth;
        public int thisHeight;
        private int lastMousePositionX = 0;
        private int lastMousePositionY = 0;
        private bool mousePressed = false;
        public int cornerXSquare = 0;
        public int cornerYSquare = 0;
        public Bitmap bitmapWindowSized = new Bitmap(200, 200);
        public Bitmap bitmapLocalFile = new Bitmap(200, 200);
        public int bitmapWidth;
        public BitmapData bitmapData = null;
        public Byte* pBase = null;
        public Boolean thisHasFocus = false;
        private int thisThumbBoxNumber;
        public Rectangle thisBoundingBoxStart;
        public Rectangle thisBoundingBoxCurrent;
        private int thisRow;
        private int thisColumn;
        private ThumbCaptionType captionType;


        public ThumbBox(FRM_MainForm xThisParent, int xThisXPos, int xThisYPos, int xThisWidth, int xThisHeight, string xThisName, Bitmap xBitmapLocalFile, int xThumbBoxNumber, int xThisRow, int xThisColumn)
        {
            thisParent = xThisParent;
            thisXPos = xThisXPos;
            thisYPos = xThisYPos;
            thisWidth = xThisWidth;
            thisHeight = xThisHeight;
            thisName = xThisName;
            thisRow = xThisRow;
            thisColumn = xThisColumn;
            thisThumbBoxNumber = xThumbBoxNumber;
            bitmapLocalFile = xBitmapLocalFile;
            MakeGrayscale();
            EstablishProperties();
            captionType = ThumbCaptionType.Nothing;

            bool canFocus = this.CanFocus;
            Redraw();
        }

        public ThumbCaptionType CaptionType
        {
            set { captionType = value; }
            get { return captionType; }
        }

        public void EstablishProperties()
        {
            ((System.ComponentModel.ISupportInitialize)(this)).BeginInit();
            this.BorderStyle = System.Windows.Forms.BorderStyle.FixedSingle;
            this.Location = new System.Drawing.Point(thisXPos, thisYPos);
            this.Name = thisName;
            this.Size = new System.Drawing.Size(thisWidth, thisHeight);
            this.TabIndex = 0;
            this.TabStop = false;
            this.MouseDown += new System.Windows.Forms.MouseEventHandler(ThumbBox_MouseDown);
            this.MouseMove += new System.Windows.Forms.MouseEventHandler(ThumbBox_MouseMove);
            this.MouseUp += new System.Windows.Forms.MouseEventHandler(ThumbBox_MouseUp);
            this.KeyDown += new System.Windows.Forms.KeyEventHandler(ThumbBox_KeyDown);
            this.KeyPress += new System.Windows.Forms.KeyPressEventHandler(ThumbBox_KeyPress);
            this.KeyUp += new System.Windows.Forms.KeyEventHandler(ThumbBox_KeyUp);
            this.BackgroundImage = bitmapLocalFile;
            this.Refresh();
            ((System.ComponentModel.ISupportInitialize)(this)).EndInit();

        }

        private void ThumbBox_MouseDown(object sender, MouseEventArgs e)
        {

            if (!thisBoundingBoxCurrent.IsEmpty)
            {
                captionType = ThumbCaptionType.SearchError;
    
                mousePressed = true;
                lastMousePositionX = e.X;
                lastMousePositionY = e.Y;
                thisHasFocus = true;
                Redraw();
                thisParent.DeselectLastThumbBox(this);
                this.Focus();
                
            }
        }

        private void ThumbBox_MouseMove(object sender, MouseEventArgs e)
        {
            if (mousePressed)
            {
                AdjustBoundingBox(lastMousePositionX - e.X, lastMousePositionY - e.Y);
                PanMapSquare();
                lastMousePositionX = e.X;
                lastMousePositionY = e.Y;
            }
        }

        private void ThumbBox_MouseUp(object sender, MouseEventArgs e)
        {

            mousePressed = false;
            //needTemplateMatchValue = false;

        }
        private void AdjustBoundingBox(int offsetX, int offsetY)
        {
            thisBoundingBoxCurrent.X += offsetX;
            thisBoundingBoxCurrent.Y += offsetY;
            thisParent.controlMark[thisRow, thisColumn].xImage += offsetX;
            thisParent.controlMark[thisRow, thisColumn].yImage += offsetY;
            /*
            string s = String.Format(
                                    "{0:d5}\t{1:d5}\t{2:d5}\t{3:f3}\t {4:f3}     \n",
                                    thisThumbBoxNumber,
                                    thisParent.controlMark[thisRow, thisColumn].xImage,
                                    thisParent.controlMark[thisRow, thisColumn].yImage,
                                    thisParent.controlMark[thisRow, thisColumn].lonTrue,
                                    thisParent.controlMark[thisRow, thisColumn].latTrue);
            thisParent.TBX_Output.AppendText(s);
            thisParent.TBX_Output.Update();
            */


        }

        public void EmptyThumbBoxes(Bitmap solidColor)
        {
            this.BackgroundImage = solidColor;
            this.Refresh();
        }

        private void RestrictBoundingBox()
        {
            thisBoundingBoxCurrent.X = thisBoundingBoxCurrent.X < 0 ? 0 : thisBoundingBoxCurrent.X;
            thisBoundingBoxCurrent.X = thisBoundingBoxCurrent.X > thisParent.bitmapWidthPixels - thisWidth ? thisParent.bitmapWidthPixels - thisWidth : thisBoundingBoxCurrent.X;
            thisBoundingBoxCurrent.Y = thisBoundingBoxCurrent.Y < 0 ? 0 : thisBoundingBoxCurrent.Y;
            thisBoundingBoxCurrent.Y = thisBoundingBoxCurrent.Y > thisParent.bitmapHeightPixels - thisHeight ? thisParent.bitmapHeightPixels - thisHeight : thisBoundingBoxCurrent.Y;
        }

        private void PanMapSquare()
        {
            //thisParent.DisableButtons();
            string s;



            Pen borderPen = new Pen(Color.White, 2);


            CMark mark = thisParent.controlMark[thisRow, thisColumn];

            if (mark.Type == MarkType.None)
            {
                bitmapWindowSized = new Bitmap(thisWidth, thisHeight);
            }
            else if (thisParent.thumbBoxesAreColor)
            {
                if (thisBoundingBoxCurrent.IsEmpty)
                {
                    bitmapWindowSized = bitmapLocalFile.Clone(new Rectangle(0, 0, thisWidth, thisHeight), PixelFormat.Format24bppRgb);
                }
                else
                {
                    RestrictBoundingBox();
                    bitmapWindowSized = thisParent.NudgeThumbImage(thisBoundingBoxCurrent);
                }
            }
            
            else
            {
                RestrictBoundingBox();
                bitmapWindowSized = thisParent.NudgeThumbImage(thisBoundingBoxCurrent);
                MakeGrayscale();
            }

            Graphics gBmp = Graphics.FromImage(bitmapWindowSized);


            if (!thisBoundingBoxCurrent.IsEmpty && mark.Type != MarkType.None )
            {
                SolidBrush captionBrush = new SolidBrush(Color.White);

                Pen crossPen = new Pen(thisParent.crosshairColor,1);
                Pen otherPen = new Pen(thisParent.predictionColor,1);

                double err, xOther, yOther;

                double thresh1 = thisParent.threshold1;
                double thresh2 = thisParent.threshold2;
                double thresh3 = thisParent.threshold3;

                err = xOther = yOther = 0;
                if (captionType == ThumbCaptionType.TransformError)
                {
                    if (!thisParent.wantCVErrors)
                    {
                        err = mark.PixelError;
                        xOther = mark.xHat;
                        yOther = mark.yHat;
                    }
                    else
                    {
                        err = mark.PixelErrorCV;
                        xOther = mark.xHatCV;
                        yOther = mark.yHatCV;
                    }
                }
                else if (captionType == ThumbCaptionType.GeoTiffError)
                {
                    thresh1  *= thisParent.metersPerPixel;
                    thresh2  *= thisParent.metersPerPixel;
                    thresh3  *= thisParent.metersPerPixel;
                    xOther   = mark.xGeoRef;
                    yOther   = mark.yGeoRef;
                    err      = mark.GeoTiffError;
                    crossPen = new Pen(Brushes.Magenta);
                    otherPen = new Pen(Brushes.Blue);
                }

                if (captionType == ThumbCaptionType.SearchError)
                    captionBrush.Color = Color.White;

                else if(err <= thresh1)
                {
                    captionBrush.Color = thisParent.threshold1Color;
                }
                else if (err > thresh1 && err <= thresh2)
                {
                    captionBrush.Color = thisParent.threshold2Color;
                }
                else if (err > thresh2 && err <= thresh3)
                {
                    captionBrush.Color = thisParent.threshold3Color;
                }
                else if (err > thresh3)
                {
                    captionBrush.Color = thisParent.threshold4Color;
                }


                
                Rectangle captionRect = new Rectangle(0, thisHeight - 18, thisWidth, 18);

                StringFormat sf = new StringFormat();
                sf.Alignment = StringAlignment.Center;


                if (mark.Type == MarkType.None)
                {
                    gBmp.FillRectangle(captionBrush, captionRect);
                    gBmp.DrawString("n/a: Perimeter Only", new Font("Arial", 9), Brushes.Black, captionRect, sf);
                }
                else if (captionType == ThumbCaptionType.TransformError)
                {
                    gBmp.FillRectangle(captionBrush, captionRect);
                    if(thisParent.wantCVErrors) s = String.Format("CV Error: {0:f2} pixels",err);
                    else                        s = String.Format("Residual: {0:f2} pixels", err);

                    gBmp.DrawString(s, new Font("Arial", 9), Brushes.Black, captionRect,sf);
                    
                    int x = (int)Math.Round(xOther - thisBoundingBoxCurrent.Left);
                    int y = (int)Math.Round(yOther - thisBoundingBoxCurrent.Top);


                    gBmp.DrawRectangle(otherPen, x - 3, y - 3, 7, 7);
                    gBmp.DrawRectangle(otherPen, x, y, 1, 1);
                    gBmp.DrawLine(otherPen, x - 20, y, x - 3, y);
                    gBmp.DrawLine(otherPen, x + 4, y, x + 20, y);
                    gBmp.DrawLine(otherPen, x, y - 20, x, y - 3);
                    gBmp.DrawLine(otherPen, x, y + 4, x, y + 20);
                    
           
                }
                else if (captionType == ThumbCaptionType.SearchError)
                {
                    int offsetX = thisBoundingBoxStart.X - thisBoundingBoxCurrent.X;
                    int offsetY = thisBoundingBoxStart.Y - thisBoundingBoxCurrent.Y;


                    gBmp.FillRectangle(captionBrush, captionRect);
                    Rectangle rFull = new Rectangle(
                               mark.xImage - thisParent.searchActualSize / 2,
                               mark.yImage - thisParent.searchActualSize / 2,
                               thisParent.searchActualSize, thisParent.searchActualSize);
                     
                    
                    s = String.Format("u,v=({0:d},{1:d}) e={2:f3}",
                             offsetX,offsetY,
                             mark.CheckLocation(
                                  thisParent.imageFile.ExtractRectangle(rFull),
                                  rFull.Width / 2, rFull.Height/2 ));

                    gBmp.DrawString(s, new Font("Arial", 9), Brushes.Black, captionRect,sf);
                }

                else if (captionType == ThumbCaptionType.GeoTiffError)
                {
                    int offsetX = thisBoundingBoxStart.X - thisBoundingBoxCurrent.X;
                    int offsetY = thisBoundingBoxStart.Y - thisBoundingBoxCurrent.Y;


                    gBmp.FillRectangle(captionBrush, captionRect);
                    Rectangle rFull = new Rectangle(
                               mark.xImage - thisParent.searchActualSize / 2,
                               mark.yImage - thisParent.searchActualSize / 2,
                               thisParent.searchActualSize, thisParent.searchActualSize);


                    {

                        int x = (int)Math.Round(xOther - thisBoundingBoxCurrent.Left);
                        int y = (int)Math.Round(yOther - thisBoundingBoxCurrent.Top);


                        gBmp.DrawRectangle(otherPen, x - 3, y - 3, 7, 7);
                        gBmp.DrawRectangle(otherPen, x, y, 1, 1);
                        gBmp.DrawLine(otherPen, x - 20, y, x - 3, y);
                        gBmp.DrawLine(otherPen, x + 4, y, x + 20, y);
                        gBmp.DrawLine(otherPen, x, y - 20, x, y - 3);
                        gBmp.DrawLine(otherPen, x, y + 4, x, y + 20);
                    }

                    if (err < 999) s = String.Format("Poss. Error={0:f2}m", err);
                    else s = "Mark Not Found";

                    gBmp.DrawString(s, new Font("Arial", 9), Brushes.Black, captionRect, sf);
                }

                int centerThumbX = thisWidth / 2;
                int centerThumbY = thisHeight / 2;

                gBmp.DrawRectangle(crossPen, centerThumbX - 3, centerThumbY - 3, 7, 7);
                gBmp.DrawRectangle(crossPen, centerThumbX, centerThumbY, 1, 1);
                gBmp.DrawLine(crossPen, centerThumbX - 20, centerThumbY, centerThumbX - 3, centerThumbY);
                gBmp.DrawLine(crossPen, centerThumbX + 4, centerThumbY, centerThumbX + 20, centerThumbY);
                gBmp.DrawLine(crossPen, centerThumbX, centerThumbY - 20, centerThumbX, centerThumbY - 3);
                gBmp.DrawLine(crossPen, centerThumbX, centerThumbY + 4, centerThumbX, centerThumbY + 20);

                captionBrush.Dispose();
                crossPen.Dispose();
                otherPen.Dispose();
            }

            if (thisHasFocus)
            {
                gBmp.DrawRectangle(borderPen, 1, 1, thisWidth - 5, thisHeight - 5);
            }

            borderPen.Dispose();
            gBmp.Dispose();

            this.BackgroundImage = bitmapWindowSized;
            
        }

        public void Redraw()
        {
            if (!thisBoundingBoxCurrent.IsEmpty)
            {
                PanMapSquare();
                this.Refresh();
                Application.DoEvents();
            }
        }

        public struct Pixel
        {
            public byte blue;
            public byte green;
            public byte red;
        }

        public Pixel* PixelAt(int x, int y)
        {
            return (Pixel*)(pBase + y * bitmapWidth + x * sizeof(Pixel));
        }

        public void LockBitmap()
        {
            GraphicsUnit unit = GraphicsUnit.Pixel;
            RectangleF boundsF = bitmapWindowSized.GetBounds(ref unit);
            Rectangle bounds = new Rectangle((int)boundsF.X, (int)boundsF.Y, (int)boundsF.Width, (int)boundsF.Height);
            bitmapWidth = (int)boundsF.Width * sizeof(Pixel);

            if (bitmapWidth % 4 != 0)
            {
                bitmapWidth = 4 * (bitmapWidth / 4 + 1);
            }

            bitmapData = bitmapWindowSized.LockBits(bounds, ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb);
            pBase = (Byte*)bitmapData.Scan0.ToPointer();
        }

        public void UnlockBitmap()
        {
            bitmapWindowSized.UnlockBits(bitmapData);
            bitmapData = null;
            pBase = null;

        }

        public Point PixelSize
        {
            get
            {
                GraphicsUnit unit = GraphicsUnit.Pixel;
                RectangleF bounds = bitmapWindowSized.GetBounds(ref unit);

                return new Point((int)bounds.Width, (int)bounds.Height);
            }
        }

        public void MakeGrayscale()
        {
            LockBitmap();

            for (int y = 0; y < thisHeight; y++)
            {
                for (int x = 0; x < thisWidth; x++)
                {
                    Pixel* pPixel = PixelAt(x, y);
                    byte makeGrey = (byte)((pPixel->red + pPixel->green + pPixel->blue) / 3);
                    pPixel->red = makeGrey;
                    pPixel->green = makeGrey;
                    pPixel->blue = makeGrey;
                }
            }
            UnlockBitmap();
        }

        private void ThumbBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode.ToString() == "Up")
            {
                if ((e.Modifiers & Keys.Shift) == Keys.Shift)
                {
                    AdjustBoundingBox(0, 10);
                    PanMapSquare();
                }
                else
                {
                    AdjustBoundingBox(0, 1);
                    PanMapSquare();
                }
            }

            if (e.KeyCode.ToString() == "Down")
            {
                if ((e.Modifiers & Keys.Shift) == Keys.Shift)
                {
                    AdjustBoundingBox(0, -10);
                    PanMapSquare();
                }
                else
                {
                    AdjustBoundingBox(0, -1);
                    PanMapSquare();
                }
            }

            if (e.KeyCode.ToString() == "Left")
            {
                if ((e.Modifiers & Keys.Shift) == Keys.Shift)
                {
                    AdjustBoundingBox(10, 0);
                    PanMapSquare();
                }
                else
                {
                    AdjustBoundingBox(1, 0);
                    PanMapSquare();
                }
            }

            if (e.KeyCode.ToString() == "Right")
            {
                if ((e.Modifiers & Keys.Shift) == Keys.Shift)
                {
                    AdjustBoundingBox(-10, 0);
                    PanMapSquare();
                }
                else
                {
                    AdjustBoundingBox(-1, 0);
                    PanMapSquare();
                }
            }
        }

        private void ThumbBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            e.Handled = true;
        }

        private void ThumbBox_KeyUp(object sender, KeyEventArgs e)
        {
            
            mousePressed = false;
            
            e.Handled = true;
        }
        protected override void OnLostFocus(EventArgs e)
        {
            thisHasFocus = false;
            Redraw();
        }
    }
}
